unit mmobd;

interface

uses SysUtils,Classes,mgscript,worldcontent;

type
 TBag = record
  slot: array [0..5] of integer;
 end;
 TSpell = record
  rawcode: string;
  lvl: integer;
 end;
 TSpells = array [0..5] of TSpell;
 TChar = class(TUnit)
  charid: integer;
  status: boolean;
  utype: string;
  items: TBag;
  constructor Create(charinfo: string);
  destructor Destroy;
  public
   function GetInfStr: string;
   procedure UpdateInfo(upinfo: string);
   procedure UpdateStateInfo(stateinfo: string);
  private
   _index: integer;
end;

type
 TCharDataBase = class
  charcount: integer;
  lines: TStrings;
  savepath: string;
  Chars: array of TChar;
  constructor Load(path: string);
  destructor Destroy;
  public
   procedure Save;
   function FindChar(name: string):integer;
   function FindCharById(id: integer):integer;
   function GetCharStatus(charid: integer):boolean;
   procedure UpdateCharInfo(update: string);
   procedure UpdateCharStateInfo(stateupdate: string);
end;

implementation

uses mgscriptnatives;

function B2S(b: boolean): string;
begin
 if b then
  Result:= '1'
 else
  Result:= '0';
end;

//=====================================
// TChar
//

constructor TChar.Create(charinfo: string);
begin
 inherited Create(false);
 charid:= StrToInt(GetComParam(charinfo,1));
 status:= false;
 name:= GetComParam(charinfo,2);
 utype:= GetComParam(charinfo,3);
 x:= StrToInt(GetComParam(charinfo,4));
 y:= StrToInt(GetComParam(charinfo,5));
 yaw:= StrToInt(GetComParam(charinfo,6));
 health:= StrToInt(GetComParam(charinfo,7));
 mana:= StrToInt(GetComParam(charinfo,8));
end;

function TChar.GetInfStr: string;
begin
 Result:= IntToStr(charid)+','+name+','+utype+',';
 Result:= Result + IntToStr(x)+','+IntToStr(y)+','+IntToStr(yaw)+',';
 Result:= Result + IntToStr(health)+','+IntToStr(mana);
end;

procedure TChar.UpdateInfo(upinfo: string);
begin
 x:= StrToInt(GetComParam(upinfo,2));
 y:= StrToInt(GetComParam(upinfo,3));
 yaw:= StrToInt(GetComParam(upinfo,4));
end;

procedure TChar.UpdateStateInfo(stateinfo: string);
begin
 mana:= StrToInt(GetComParam(stateinfo,2));
 health:= StrToInt(GetComParam(stateinfo,3));
// if health < 1 then
//  Compile('scripts\HeroEvent.mgscript','Death',IntToStr(hunit));
end;

destructor TChar.Destroy;
begin
 inherited Destroy();
end;

//=====================================
// TCharDataBase
//

constructor TCharDataBase.Load(path: string);
 var i: integer;
begin
 inherited Create();
 lines:= TStringList.Create;
 lines.LoadFromFile(path);
 charcount:= 0;
 savepath:= path;
 for i:= 0 to lines.Count-1 do
 begin
  inc(charcount);
  SetLength(Chars,charcount);
  Chars[charcount-1]:= TChar.Create(lines[i]);
 end;
end;

destructor TCharDataBase.Destroy;
begin
 inherited Destroy;
 lines.Destroy;
end;


procedure TCharDataBase.Save;
 var i: integer;
     locstr: TStrings;
begin
 locstr:= TStringList.Create;
 for i:= 0 to charcount-1 do
  locstr.Add(Chars[i].GetInfStr);
 locstr.SaveToFile(savepath);
 locstr.Destroy;
 //lines.SaveToFile(savepath);
end;

function TCharDataBase.FindChar(name: string):integer;
 var i: integer;
begin
 for i:= 0 to charcount-1 do
  if LowerCase(Chars[i].name) = LowerCase(name) then
  begin
   result:= i;
   exit;
 end;
 Result:= -1;
end;

function TCharDataBase.FindCharById(id: integer):integer;
 var i: integer;
begin
 for i:= 0 to charcount-1 do
  if Chars[i].charid = id then
  begin
   result:= i;
   exit;
 end;
 Result:= -1;
end;

procedure TCharDataBase.UpdateCharInfo(update: string);
 var id: integer;
begin
 id:= FindCharById(StrToInt(GetComParam(update,1)));
 if id < 0 then exit;
 Chars[id].UpdateInfo(update);
end;

procedure TCharDataBase.UpdateCharStateInfo(stateupdate: string);
 var id: integer;
begin
 id:= FindCharById(strToInt(GetComParam(stateupdate,1)));
 if id < 0 then exit;
 Chars[id].UpdateStateInfo(stateupdate);
end;

function TCharDataBase.GetCharStatus(charid: integer):boolean;
 var id: integer;
begin
 id:= FindCharById(charid);
 if id < 0 then exit;
 Result:= Chars[id].status;
end;

end.
