unit worldcontent;

interface

uses SysUtils;

type
 TDestState = (ds_open,ds_close,ds_destroy,ds_show,ds_hide);
 TDestStates = set of TDestState;
 TDest = record
  id: integer;
  state: TDestState;
 end;
 TTrig = record
  id: integer;
  enable: boolean;
 end;
 TOrderType = (uotXY, uotUnit, uotSimple);
 TOrder = record
  _type: TOrderType;
  order: string;
 end;
 TUnit = class
  x,y,yaw,health,mana: integer;
  hunit: integer;
  id: integer;
  name: string;
  currentOrder: TOrder;
  constructor Create(getHUNIT: boolean);
  destructor Destroy;
 end;

function DestStateToOperateID(DS: TDestState): integer;
procedure DefineDest(id: integer; state: TDestState);
function FindDestByID(id: integer): integer;
function OperateIDToDestState(id: integer): TDestState;

procedure DefineTrig(id: integer; enable: boolean);
function FindTrigByID(id: integer): integer;
function TrigEnableToStr(e: boolean): string;
function StrToTrigEnable(s: string): boolean;

function GetFreeHUNIT: integer;
procedure FreeHUNIT (hunit: integer);

function GetUnitByHUnit(hunit: integer): integer;

var
 GameUnits: array of TUnit;
 GameUnitCount: integer = 0;
 dests: array [0..8191] of TDest;
 trigs: array [0..8191] of TTrig;
 npccount: integer = 0;
 destcount: integer = 0;
 trigcount: integer = 0;

implementation

var
 unithandles: array of boolean;
 unitcount: integer = 0;

function GetFreeHUNIT: integer;
 var i: integer;
begin
 for i:= 0 to unitcount-1 do
  if not unithandles[i] then
  begin
   unithandles[i]:= true;
   Result:= i;
   exit;
 end;
 Result:= unitcount;
 inc(unitcount);
 SetLength(unithandles,unitcount);
 unithandles[result]:= true;
end;

procedure FreeHUNIT (hunit: integer);
begin
 unithandles[hunit]:= false;
end;

//==========================================

function FindDestByID(id: integer): integer;
 var i: integer;
begin
 Result:= -1;
 for i:= 0 to destcount-1 do
  if dests[i].id = id then
  begin
   result:= i;
   exit;
 end;
end;

procedure DefineDest(id: integer; state: TDestState);
begin
 dests[destcount].id:= id;
 dests[destcount].state:= state;
 inc(destcount);
end;

function DestStateToOperateID(DS: TDestState): integer;
begin
 case DS of
  ds_open: Result:= 0;
  ds_close: Result:= 1;
  ds_destroy: Result:= 2;
  ds_show: Result:= 3;
  ds_hide: result:= 4;
 end;
end;

function OperateIDToDestState(id: integer): TDestState;
begin
 case id of
  0: Result:= ds_open;
  1: Result:= ds_close;
  2: Result:= ds_destroy;
  3: Result:= ds_show;
  4: Result:= ds_hide;
 end;
end;

//==============================================================================

function FindTrigByID(id: integer): integer;
 var i: integer;
begin
 Result:= -1;
 for i:= 0 to trigcount - 1 do
  if trigs[i].id = id then
  begin
   Result:= i;
   exit;
 end;
end;

procedure DefineTrig(id: integer; enable: boolean);
begin
 trigs[trigcount].id:= id;
 trigs[trigcount].enable:= enable;
 inc(trigcount);
end;

function StrToTrigEnable(s: string): boolean;
begin
 Result:= false;
 if (s = '0') or (s = '2') then Result:= true;
end;

function TrigEnableToStr(e: boolean): string;
begin
 Result:= '0';
 if e = false then Result:= '1';
end;

//==============================================================================
//==============================================================================
//==============================================================================
//            TUnit 
//==============================================================================
//==============================================================================
//==============================================================================

function GetUnitByHUnit(hunit: integer): integer;
 var i: integer;
begin
 for i:= 0 to GameUnitCount-1 do
  if GameUnits[i].hunit = hunit then
  begin
   Result:= i;
   exit;
 end;
 Result:= -1;
end;

constructor TUnit.Create(getHUNIT: boolean);
begin
 inherited Create;
 if getHUNIT then
  HUNIT:= GetFreeHUnit
 else
  HUnit:= -1;
 id:= GameUnitCount;
 inc(GameUnitCount);
 SetLength(GameUnits,GameUnitCount);
 GameUnits[id]:= Self;
end;

destructor TUnit.Destroy;
begin
 if HUnit >= 0 then
  FreeHUnit(hunit);
 dec(GameUnitCount);
 GameUnits[id]:= GameUnits[GameUnitCount];
 GameUnits[id].id:= id;
end;

end.
