unit mgscriptnatives;

interface

uses ScktComp,mgscript,Classes,SysUtils,Dialogs, worldcontent;

type
 TVarType = (vtString,vtInteger,vtReal);
 TVar = class
  name,value: string;
  id: integer;
  t: TVarType;
  constructor New(nname: string; nt: TVarType; nvalue: string);
  destructor Destroy;
 end;
 TRunPoint = record
  name: string;
  pos: integer;
 end;
 TRunPoints = array of TRunPoint;


procedure InitNatives(Socket: TServerSocket);
procedure Compile(scriptlist: TStrings; runpoint: string; scriptparams: string);
//natives
procedure Msg(msg: string);

function GetValue(parm: string):string;


var
 ServerSocket: TServerSocket;
 vars: array of TVar;
 varcount: integer = 0;

implementation

uses Main;

var
 ScriptParam: string;

//==============================
// TVar
//

constructor TVar.New(nname: string; nt: TVarType; nvalue: string);
begin
 inherited Create;
 name:= nname;
 t:= nt;
 value:= nvalue;
 id:= varcount;
 inc(varcount);
 SetLength(vars,varcount);
 vars[id]:= Self;
end;

destructor TVar.Destroy;
begin
 inherited Destroy;
 dec(varcount);
 vars[id]:= vars[varcount];
 vars[id].id:= id;
end;

procedure InitNatives(Socket: TServerSocket);
begin
 ServerSocket:= Socket;
end;

//===========================
// Natives
//

procedure CreateNPC(id,x,y: integer; rawid: string);
begin
 SendMsg('npc ('+inttostr(id)+','+inttostr(x)+','+inttostr(y)+','+rawid+')',nil);
end;

procedure NPCSendTarget(id: integer; target: string);
begin
 SendMsg('npct ('+inttostr(id)+','+target+')',nil);
end;

procedure NPCSendTargetXY(id: integer; target: string; x,y: integer);
begin
 SendMsg('npctl ('+inttostr(id)+','+target+','+inttostr(x)+','+inttostr(y)+')',nil);
end;

procedure Msg(msg: string);
begin
 SendMsg('servermsg ('+msg+')',nil);
end;

//===========================
// Compile Syntax
//

function GetParentObj(comname: string; lvl: integer): string;
var i,j: integer;
    pobj: string;
begin
 j:= 1;
 for i:= Length(comname) downto 1 do
 begin
  if comname[i] = '.' then inc(j);
  if (lvl = j) and (comname[i] <> '.') then pobj:= comname[i]+pobj;
 end;
 Result:= pobj;
end;

function FindVarId(name: string):integer;
 var i: integer;
begin
 for i:= 0 to varcount-1 do
  if vars[i].name = name then
  begin
   result:= i;
   exit;
 end;
 result:= -1;
end;

function IsVariable(par: string):boolean;
begin
 Result:= false;
 if Length(par) < 1 then exit;
 Result:= par[1] = '@';
end;

function IsIntVar(str: string): boolean;
 var i,e: integer;
begin
 Val(str,i,e);
 Result:= e = 0;
end;

function SummStr(str: string): string;
 var i: integer;
     count: integer;
     s: array [0..1000] of string;
begin
 count:= 0;
 i:= Pos('+',str);
 while true do
 begin
  if i = 0 then break;
  s[count]:= Copy(str,1,i-1);
  Delete(str,1,i);
  inc(count);
  i:= Pos('+',str);
 end;
 s[count]:= str;
 Result:= '';
 for i:= 0 to count do
  Result:= Result + GetValue(s[i]);
end;
  

function GetValue(parm: string):string;
 var i: integer;
     comname,params: string;
     par: array [1..10] of string;
begin
 if IsVariable(parm) and (Pos('.',parm) = 0) then
 begin
  i:= FindVarId(Copy(parm,2,Length(parm)));
  if i >= 0 then
   Result:= vars[i].value
  else
   Result:= parm;
  exit;
 end;

 comname:= GetComName(parm);
 params:= GetComParams(parm);
 //==========================
 // GetScriptParam(@num): param
 //==========================
 if comname = 'GetScriptParam' then
 begin
  par[1]:= GetValue(GetComParam(params,1));
  if IsIntVar(par[1]) then
   Result:= GetComParam(ScriptParam,StrToInt(par[1]));
  exit;
 end;
 //==========================
 // GetCharByName( @name ): HChar
 //==========================
 if comname = 'GetCharByName' then
 begin
  par[1]:= GetValue(GetComParam(params,1));
  Result:= IntToStr(DataBase.FindChar(par[1]));
  exit;
 end;
 //==========================
 // GetCharByID( @ID ): HChar
 //==========================
 if comname = 'GetCharByID' then
 begin
  par[1]:= GetValue(GetComParam(params,1));
  if IsIntVar(par[1]) then
   Result:= IntToStr(DataBase.FindCharById(StrToInt(par[1])));
  exit;
 end;
 //==========================
 // @charid.Name: charname
 //==========================
 if GetParentObj(comname,1) = 'Name' then
 begin
  par[1]:= GetValue(GetParentObj(comname,2));
  if IsIntVar(par[1]) then
   Result:= DataBase.Chars[StrToInt(par[1])].name;
  exit;
 end;
 //==========================
 // @charid.CharID: charid
 //==========================
 if GetParentObj(comname,1) = 'CharID' then
 begin
  par[1]:= GetValue(GetParentObj(comname,2));
  if IsIntVar(par[1]) then
   Result:= IntToStr(DataBase.Chars[StrToInt(par[1])].charid);
  exit;
 end;
 //==========================
 // @charid.X: charxpos
 //==========================
 if GetParentObj(comname,1) = 'X' then
 begin
  par[1]:= GetValue(GetParentObj(comname,2));
  if IsIntVar(par[1]) then
   Result:= IntToStr(DataBase.Chars[StrToInt(par[1])].x);
  exit;
 end;
 //==========================
 // @charid.Y: charypos
 //==========================
 if GetParentObj(comname,1) = 'Y' then
 begin
  par[1]:= GetValue(GetParentObj(comname,2));
  if IsIntVar(par[1]) then
   Result:= IntToStr(DataBase.Chars[StrToInt(par[1])].y);
  exit;
 end;
 //==========================
 // @charid.Life
 //==========================
 if GetParentObj(comname,1) = 'Life' then
 begin
  par[1]:= GetValue(GetParentObj(comname,2));
  if IsIntVar(par[1]) then
   Result:= IntToStr(DataBase.Chars[StrToInt(par[1])].health);
  exit;
 end;
 //==========================
 // @charid.Mana
 //==========================
 if GetParentObj(comname,1) = 'Mana' then
 begin
  par[1]:= GetValue(GetParentObj(comname,2));
  if IsIntVar(par[1]) then
   Result:= IntToStr(DataBase.Chars[StrToInt(par[1])].mana);
  exit;
 end;

 if Pos('+',parm) > 0 then
  Result:= SummStr(parm)
 else
  Result:= parm;
 //Result:= StringReplace(parm,'+','',[rfReplaceAll]);
end;

function ConvertVarType(s: string): TVarType;
begin
 if s = '[str]' then Result:= vtString;
 if s = '[int]' then Result:= vtInteger;
 if s = '[rea]' then Result:= vtReal;
end;

procedure GetRunPoints(scriptlist: TStrings; var rp: TRunPoints; var count: integer);
 var i: integer;
     buf: string;
begin
 count:= 0;
 for i:= 0 to scriptlist.Count-1 do
 begin
  buf:= TrimLeft(scriptlist[i]);
  if Length(buf) > 0 then
   if buf[1] = ':' then
   begin
    inc(count);
    SetLength(rp,count);
    rp[count-1].name:= Copy(buf,2,Length(buf));
    rp[count-1].pos:= i;
   end;
 end;
end;

function FindRunPoint(rp: TRunPoints; name: string; count: integer): TRunPoint;
 var i: integer;
begin
 for i:= 0 to count-1 do
  if rp[i].name = name then
  begin
   Result:= rp[i];
   exit;
 end;
end;

procedure Compile(scriptlist: TStrings; runpoint: string; scriptparams: string);
 var i,j,z,loopcount: integer;
     looppos,loopcircles: array [1..8192] of integer;
     comname,params,buf: string;
     par: array [1..10] of string;
     localtextfile: TStrings;
     last: TNPC;
     //======
     RunPoints: TRunPoints;
     RPCount: integer;
     userun: boolean;
begin
 i:= 0;
 loopcount:= 0;
 userun:= false;
 ScriptParam:= scriptparams;
 GetRunPoints(scriptlist,RunPoints,RPCount);
 while i <= scriptlist.Count-1 do
 begin
  buf:= scriptlist[i];
  if Pos('//',buf) > 0 then
   Delete(buf,Pos('//',buf),Length(buf));
  comname:= GetComName(buf);
  params:= GetComParams(buf);
  //=====================
  // loop (@circles)
  //=====================
  if comname = 'loop' then
  begin
   inc(loopcount);
   //params:= GetComParams(scriptlist[i]);
   looppos[loopcount]:= i;
   par[1]:= GetValue(GetComParam(params,1));
   loopcircles[loopcount]:= strtoint(par[1]);
  end;
  //=====================
  // endloop
  //=====================
  if comname = 'endloop' then
  begin
   dec(loopcircles[loopcount]);
   if loopcircles[loopcount] > 1 then
    i:= looppos[loopcount]
   else
    dec(loopcount);
   end;
  //=====================
  // InitEnd
  //=====================
  if comname = 'InitEnd' then
   userun:= true;
  //=====================
  // Jump (@name)
  //=====================
  if comname = 'Jump' then
  begin
   par[1]:= GetValue(GetComParam(params,1));
   i:= FindRunPoint(RunPoints,par[1],RPCount).pos;
  end;
  //=====================
  // run point
  //=====================
  if Length(runpoint) > 0 then
   if userun then
   begin
    i:= FindRunPoint(RunPoints,runpoint,RPCount).pos;
    userun:= false;
   end;
  //=====================
  // Stop
  //=====================
  if comname = 'Stop' then
   exit;


   //SCRIPTS FUNCTIONS


  //=====================
  // LoadMGS (@path, @runpoint, @runparams)
  //=====================
  if comname = 'LoadMGS' then
  begin
   par[1]:= GetValue(GetComParam(params,1));
   if FileExists(par[1]) then
   begin
    par[2]:= GetValue(GetComParam(params,2));
    par[3]:= GetValue(GetComParam(params,3));
    localtextfile:= TStringList.Create;
    localtextfile.LoadFromFile(par[1]);
    Compile(localtextfile,par[2],par[3]);
    localtextfile.Destroy;
   end;
  end;
  //=====================
  // CreateNPC (@id,@rawid,@x,@y,@yaw)
  //=====================
  if comname = 'CreateNPC' then
  begin
   //params:= GetComParams(scriptlist[i]);
   par[1]:= GetValue(GetComParam(params,1));
   par[2]:= GetValue(GetComParam(params,2));
   par[3]:= GetValue(GetComParam(params,3));
   par[4]:= GetValue(GetComParam(params,4));
   par[5]:= GetValue(GetComParam(params,5));
   last:= TNPC.Create(par[2],strtoint(par[1]),strtoint(par[3]),strtoint(par[4]),strtoint(par[5]));
   SendMsg(last.FormServerStr,nil);
  end;
  //======================
  // DestOperate (@id,@operateid)
  //======================
  if comname = 'DestOperate' then
  begin
   par[1]:= GetValue(GetComParam(params,1));
   par[2]:= GetValue(GetComParam(params,2));
   if FindDestByID(StrToInt(par[1])) < 0 then
    DefineDest(StrToInt(par[1]),OperateIDToDestState(StrToInt(par[2])))
   else
    Dests[StrToInt(par[1])].state:= OperateIDToDestState(StrToInt(par[2]));
   SendMsg('destoperate ('+par[1]+','+par[2]+')',nil);
  end;
  {//=====================
  // @id.NPCSendTarget(@target)
  //=====================
  if GetParentObj(comname,1) = 'NPCSendTarget' then
  begin
   //params:= GetComParams(scriptlist[i]);
   par[1]:= GetValue(GetParentObj(comname,2));
   par[2]:= GetValue(GetComParam(params,1));
   NPCSendTarget(strtoint(par[1]),par[2]);
  end;
  //=====================
  // @id.NPCSendTargetXY(@target,@x,@y)
  //=====================
  if GetParentObj(comname,1) = 'NPCSendTargetXY' then
  begin
   //params:= GetComParams(scriptlist[i]);
   par[1]:= GetValue(GetParentObj(comname,2));
   par[2]:= GetValue(GetComParam(params,1));
   par[3]:= GetValue(GetComParam(params,2));
   par[4]:= GetValue(GetComParam(params,3));
   NPCSendTargetXY(strtoint(par[1]),par[2],strtoint(par[3]),strtoint(par[4]));
  end;}
  //======================
  // DebugMsg(@msg)
  //======================
  if comname = 'DebugMsg' then
  begin
   //params:= GetComParams(scriptlist[i]);
   par[1]:= GetValue(GetComParam(params,1));
   ShowMessage(par[1]);
  end;
  //======================
  // SendMsg(@msg)
  //======================
  if comname = 'SendMsg' then
  begin
   par[1]:= GetValue(GetComParam(params,1));
   SendMsg('servermsg ('+par[1]+')',nil);
  end;


  // VARIABLES FUNCTIONS


  //======================
  // inc(@intvar)
  // (only Integer Variables)
  //======================
  if comname = 'inc' then
  begin
   par[1]:= GetComParam(params,1);
   par[1]:= Copy(par[1],2,Length(par[2]));
   j:= FindVarId(par[1]);
   if j > 0 then
    if vars[j].t = vtInteger then
     vars[j].value:= IntToStr(StrtoInt(vars[j].value)+1);
  end;
  //======================
  // def [vt] @varname = @value
  //======================
  params:= buf;
  if Copy(params,1,3) = 'def' then
  begin
   par[3]:= Copy(params,5,5);
   params:= Copy(params,11,Length(params));
   j:= Pos('=',params);
   par[1]:= Copy(params,1,j-2);
   par[2]:= GetValue(Copy(params,j+2,Length(params)));
   if FindVarId(par[1]) < 0 then
    TVar.New(par[1],ConvertVarType(par[3]),par[2]);
  end;
  //======================
  // @varname.set(@newvalue)
  //======================
  params:= GetComParams(buf);
  if GetParentObj(comname,1) = 'set' then
  begin
   par[1]:= GetParentObj(comname,2);
   par[2]:= GetValue(GetComParam(GetComParams(scriptlist[i]),1));
   j:= FindVarId(par[1]);
   if j > 0 then
    vars[j].value:= par[2];
  end;
  params:= GetComParams(buf);
  //======================
  // @strvar.add(@addstr)
  // (only for string variables)
  //======================
  if GetParentObj(comname,1) = 'add' then
  begin
   par[1]:= GetValue(GetParentObj(comname,2));
   par[2]:= GetValue(GetComParam(params,1));
   j:= FindVarId(par[1]);
   if j >= 0 then
    if vars[j].t = vtString then
     vars[j].value:= vars[j].value+par[2];
  end;
  //======================
  // @str.Free
  //======================
  if GetParentObj(comname,1) = 'Free' then
  begin
   par[1]:= GetValue(GetParentObj(comname,2));
   j:= FindVarID(par[1]);
   if j >= 0 then
    vars[j].Destroy;
  end; 

  //======================
  // if (@bool) then
  //  @actionthen
  // else
  //  @actionelse
  // endif
  //======================
  inc(i);
 end;
 ScriptParam:= '';
end;



end.
