unit OtherDataBase;

interface

uses SysUtils,Windows,WorldContent,Classes;

const
    effect_TargetFinQuest     = 'war3mapImported\floating questionmark_v2.mdl';
    effect_TargetNotFinQuest  = 'Objects\RandomObject\RandomObject.mdl';
    effect_GiveQuest          = 'Abilities\Spells\Other\TalkToMe\TalkToMe.mdl';

type
 TPercent = 0..100;
 TRawCode = string [4];
 TItemInfo = record
  name: string [32];
  forclass: array [0..9] of boolean;
  gold: cardinal;
  honor: word;
  useable: boolean;
  charges: integer;
  rawcode: TRawCode;
  needlevel: byte;
  //
  Str: word; //
  Agi: word; //
  Int: word; //
  Stamina: word; //
  //----
  Resistance: word; //
  Armor: word; //
  MagicResist: TMagicResist; // 
  Parr: word; //
  Absorb: word; //
  //----
  Damage: word; // 
  DblDamage: byte; // . 
  Punch: word; //
  Accuracy: word; //
  Speed: word; // 
  //----
  MagicDamage: word; // 
  MagicPower: word; // 
  MagicDblDmg: byte; //.   
  MagicCast: word; //
 end;
 TUnitInfo = record
  name: string [16];
  level: byte;
  giveexp: integer;
  spell: array [0..5] of TRawCode;
  missper: TPercent;
  maxhealth: integer;
  maxmana: integer;
  rawcode: TRawCode;
  mingold: integer;
  maxgold: integer;
 end;
 TItemInfoDataBase = class
  count: integer;
  items: array of TItemInfo;
  _path: string;
  constructor Create;
  destructor Destroy(s: boolean);
  private
   function FGetByRaw(rawcode: TRawCode): TItemInfo;
  public
   property ItemsRaw [rawcode: TRawCode]: TItemInfo read FGetByRaw; default;
   function Add(name: string; rawcode: TRawCode): integer;
   function Load(path: string): boolean;
   function RemoveItem(id: integer): boolean;
   procedure Save(path: string);
   procedure QSave;
 end;
 TUnitInfoDataBase = class
  count: integer;
  units: array of TUnitInfo;
  _path: string;
  constructor Create;
  destructor Destroy(s: boolean);
  private
   function FGetByRaw(rawcode: TRawCode): TUnitInfo;
  public
   property UnitsRaw [rawcode: TRawCode]: TUnitInfo read FGetByRaw; default;
   function Add(name: string; rawcode: TRawCode): integer;
   function Load(path: string): boolean;
   procedure Save(path: string);
   procedure QSave;
  end;
 TQuestInfo = record
   iconPath: string [128];
   title: string [128];
   discription: string [255];
   qItem: array [0..15] of string [255];
   statusPath: string [64];   //   [GDT]
   finishedPath: string [64]; //    [GDT]
   completePath: string [64]; //    [GDT]
   effectPath: string [64];   //       [LDT]
   needLvl: integer;
   needClass: array [0..9] of boolean;
   giverUnit: integer; //HUnit
   targetUnit: integer;//HUnit
   needPathList: array [0..4] of string [64]; //      (AND) [GDT]
 end;
 TQuestDataBase = class
   List: array of TQuestInfo;
   Count: integer;
   constructor Create;
   destructor Destroy (s: boolean);
   private
     FPath: string;
   public
     function Add(title: string): integer;
     procedure Delete (id: integer);
     function Load(path: string): boolean;
     procedure Save(path: string);
     procedure QSave;
     function Find(status: string): integer;
     function ToStringNew(id: integer): string;
     function ItemToString(id,pos: integer; un: TUnit): string;
     function CheckStatus (id: integer; un: TUnit): boolean;
     function CheckGiveEffect (id: integer; un: TUnit; need: boolean): boolean;
     function CheckTargetEffect (id: integer; un: TUnit; need: boolean): boolean;
     procedure CheckOtherQuests (id: integer; char: TUnit);
 end;

 var UnDataBase: TUnitInfoDataBase;
     ItmDataBase: TItemInfoDataBase;
     QuestDataBase: TQuestDataBase;

implementation

uses CharDataBase;

{ TUnitInfoDataBase }

function TUnitInfoDataBase.Add(name: string; rawcode: TRawCode): integer;
begin
 inc(count);
 SetLength(units,count);
 Units[count-1].name:= name;
 Units[count-1].rawcode:= rawcode;
end;

constructor TUnitInfoDataBase.Create;
begin
 inherited Create;
 count:= 0;
 _path:= '';
end;

destructor TUnitInfoDataBase.Destroy(s: boolean);
begin
 if s then
  if _path <> '' then
   Save(_path);
 inherited Destroy;
end;

function TUnitInfoDataBase.FGetByRaw(rawcode: TRawCode): TUnitInfo;
 var i: integer;
begin
 for i:= 0 to count-1 do
  if Self.units[i].rawcode = rawcode then
  begin
   Result:= Self.Units[i];
   exit;
  end;
end;

function TUnitInfoDataBase.Load(path: string): boolean;
 var f: file of TUnitInfo;
begin
 _path:= path;
 Result:= false;
 if not FileExists(path) then exit;
 AssignFile(f,path);
 Reset(f);
 while not EoF(f) do
 begin
  inc(count);
  SetLength(units,count);
  Read(f,Units[count-1]);
 end;
 CloseFile(f);
 Result:= true;
end;

procedure TUnitInfoDataBase.QSave;
begin
 if _path <> '' then Save(_path);
end;

procedure TUnitInfoDataBase.Save(path: string);
 var i: integer;
     f: file of TUnitInfo;
begin
 _path:= path;
 AssignFile(f,path);
 Rewrite(f);
 for i:= 0 to count-1 do
  Write(f,Units[i]);
 CloseFile(f);
end;

{ TItemInfoDataBase }

function TItemInfoDataBase.Add(name: string; rawcode: TRawCode): integer;
 var i: integer;
begin
 Result:= count;
 inc(count);
 SetLength(items,count);
 items[Result].name:= name;
 for i:= 0 to 9 do
  items[Result].forclass[i]:= true;
 items[Result].useable:= false;
 items[Result].charges:= 0;
 items[Result].rawcode:= rawcode;
end;

constructor TItemInfoDataBase.Create;
begin
 inherited Create;
 count:= 0;
 _path:= '';
end;

destructor TItemInfoDataBase.Destroy(s: boolean);
begin
 if s and (_path <> '') then Save(_path);
 inherited Destroy;
end;

function TItemInfoDataBase.FGetByRaw(rawcode: TRawCode): TItemInfo;
 var i: integer;
begin
 ZeroMemory(@Result,SizeOf(TItemInfo));
 for i:= 0 to count-1 do
  if items[i].rawcode = rawcode then
  begin
   Result:= items[i];
   exit;
  end;
end;

function TItemInfoDataBase.Load(path: string): boolean;
 var f: file of TItemInfo;
     i: integer;
begin
 Result:= false;
 _path:= path;
 if not FileExists(path) then exit;
 try
  try
   AssignFile(f,path);
   Reset(f);
   while not EoF(f) do
   begin
    i:= count;
    inc(count);
    SetLength(items,count);
    ZeroMemory(@items[i],SizeOf(TItemInfo));
    Read(f,items[i]);
   end;
   CloseFile(f);
  except
   exit;
  end;
 finally
  Result:= true;
 end;
end;

procedure TItemInfoDataBase.QSave;
begin
 if _path <> '' then Save(_path);
end;

function TItemInfoDataBase.RemoveItem(id: integer): boolean;
begin
 Result := false;
 if (id < 0) or (id >= count) then Exit;
 dec(count);
 items[id] := items[count];
 Result := true;
end;

procedure TItemInfoDataBase.Save(path: string);
 var f: file of TItemInfo;
     i: integer;
begin
 try
  AssignFile(f,path);
  Rewrite(f);
  for i:= 0 to count-1 do
   write(f,items[i]);
  CloseFile(f);
 except
 end;
end;

{ TQuestDataBase }

function TQuestDataBase.Add(title: string): integer;
begin
  Result := Count;
  inc(Count);
  SetLength(List,Count);
  ZeroMemory(@List[Result],SizeOf(TQuestInfo));
  List[Result].title := title;
end;

function TQuestDataBase.CheckGiveEffect(id: integer; un: TUnit; need: boolean): boolean;
  var char: TCharA;
      give: TUnit;
      i: integer;
begin
  Result := false;
  if (id < 0) or (id >= Count) then Exit;
  i := FindUnitByHUNIT (List[id].giverUnit);
  if i < 0 then Exit;
  give := Units[i];
  char := un as TCharA;
  if char.DataTable.Values[List[id].completePath] = 'true' then Exit;
  if char.DataTable.Values[List[id].statusPath] = 'true' then Exit;
  if char.locid <> give.locid then Exit;
  if not List[id].needClass[char.classid] then Exit;
  if char.Level < List[id].needLvl then Exit;
  // 
  char.DataTableLoc.Values[List[id].effectPath] := 'giver';
  char.client.Send( 'enewu ("'+effect_GiveQuest+'","overhead",'+IntToStr(List[id].giverUnit)+','+IntToStr(-id-1)+')' );
  Result := true;
end;

procedure TQuestDataBase.CheckOtherQuests(id: integer; char: TUnit);
  var i: integer;
begin
    //  
    for i := 0 to Count - 1 do
    begin
      if (List[i].giverUnit = List[id].giverUnit) or (List[i].targetUnit = List[id].targetUnit) then
        if CheckStatus(i, char) then
          if not CheckGiveEffect(i, char, false) then
          begin
            if CheckTargetEffect(i, char, false) then
              Exit;
          end
          else
            Exit;
    end;
end;

function TQuestDataBase.CheckStatus(id: integer; un: TUnit): boolean;
  var i: integer;
  function SimpleBoolMath(str: string): boolean;
    var p: integer;
        s: string;
  begin
    Result := false;
    p := Pos('==',str);
    if p > 0 then
    begin
      s := Copy(str,1,p-1);
      s := un.DataTable.Values[s];
      System.Delete(str,1,p+1);
      Result := s = str;
      Exit;
    end;
    p := Pos('!=',str);
    if p > 0 then
    begin
      s := Copy(str,1,p-1);
      s := un.DataTable.Values[s];
      System.Delete(str,1,p+1);
      Result := s <> str;
      Exit;
    end;
  end;
begin
  Result := true;
  for i := 0 to 4 do
    if List[id].needPathList[i] <> '' then
      Result := Result and SimpleBoolMath(List[id].needPathList[i]);
  if Result then
    Result := Result and (un.DataTable.Values[List[i].finishedPath] <> 'true');
end;

function TQuestDataBase.CheckTargetEffect(id: integer; un: TUnit; need: boolean): boolean;
  var char: TCharA;
      target: TUnit;
      i: integer;
      s: string;
begin
  Result := false;
  if (id < 0) or (id >= Count) then Exit;
  i := FindUnitByHUNIT (List[id].targetUnit);
  if i < 0 then Exit;
  target := Units[i];
  char := un as TCharA;
  if char.locid <> target.locid then Exit;
  s := char.DataTableLoc.Values[List[id].effectPath];
  if char.DataTable.Values[List[id].completePath] <> 'true' then
  begin
    if char.DataTable.Values[List[id].finishedPath] <> 'true' then
    begin
      if s <> 'notfinished' then
      begin
        if s <> '' then char.client.Send( 'edel('+IntToStr(-id-1)+')' );
        char.DataTableLoc.Values[List[id].effectPath] := 'notfinished';
        char.client.Send( 'enewu ("'+effect_TargetNotFinQuest+'","overhead",'+IntToStr(List[id].targetUnit)+','+IntToStr(-id-1)+')' );
      end;
    end else begin
      if s <> 'finished' then
      begin
        if s <> '' then char.client.Send( 'edel('+IntToStr(-id-1)+')' );
        char.DataTableLoc.Values[List[id].effectPath] := 'finished';
        char.client.Send( 'enewu ("'+effect_TargetFinQuest+'","overhead",'+IntToStr(List[id].targetUnit)+','+IntToStr(-id-1)+')' );
      end;
    end;
  end else begin
    if s <> '' then
    begin
      char.DataTableLoc.Values[List[id].effectPath] := '';
      char.client.Send( 'edel('+IntToStr(-id-1)+')' );
    end;
  end;
  Result := true;
end;

constructor TQuestDataBase.Create;
begin
  inherited Create;
  Count := 0;
end;

procedure TQuestDataBase.Delete(id: integer);
begin
  dec(Count);
  List[id] := List[Count];
end;

destructor TQuestDataBase.Destroy(s: boolean);
begin
  if s then
    Save(FPath);
  inherited Destroy;
end;

  function ReplaceVar(str: string; char: TCharA): string;
    var i,start: integer;
        varname: string;
        get: boolean;
        varlist: TStrings;
  begin
    varlist := TStringList.Create;
    get := false;
    for i:= 1 to Length(str) do
    begin
      if str[i] = '%' then
      begin
        get := not get;
        if get then
        begin
         varname := '';
        end else begin
         if varlist.IndexOf(varname) < 0 then
          varlist.Add(varname);
        end;
      end;
      if get and (str[i] <> '%') then
        varname := varname + str[i];
    end;
    //Replace
    Result := str;
    for i := 0 to varlist.Count-1 do
      Result := StringReplace(Result,'%'+varlist[i]+'%',char.DataTable.Values[varlist[i]],[rfReplaceAll]);
    varlist.Free;
  end;

function TQuestDataBase.Find(status: string): integer;
  var i: integer;
begin
  for i := 0 to count-1 do
    if status = List[i].statusPath then
    begin
      Result := i;
      Exit;
    end;
  Result := -1;
end;

function TQuestDataBase.ItemToString(id, pos: integer; un: TUnit): string;
begin
  Result := IntToStr(id)+','+IntToStr(pos)+',"'+ReplaceVar(List[id].qItem[pos],un as TCharA)+'"';
end;

function TQuestDataBase.Load(path: string): boolean;
  var f: file of TQuestInfo;
      i: integer;
begin
  Fpath := path;
  if not FileExists(path) then
  begin
    Result := false;
    Exit;
  end;
  AssignFile(f, path);
  Reset(f);
  while not EoF(f) do
  begin
    i := Count;
    inc (Count);
    SetLength(List,Count);
    read(f,List[i]);
  end;
  CloseFile(f);
  Result := true;
end;

procedure TQuestDataBase.QSave;
begin
  Save(FPath);
end;

procedure TQuestDataBase.Save(path: string);
  var f: file of TQuestInfo;
      i: integer;
begin
  AssignFile(f,path);
  Rewrite(f);
  for i := 0 to Count-1 do
  begin
    Write(f,List[i]);
  end;
  CloseFile(f);
end;

function TQuestDataBase.ToStringNew(id: integer): string;
begin
  Result := IntToStr(id)+',"'+List[id].title+'","'+List[id].discription+'","'+List[id].iconPath+'"';
end;

end.
