unit WorldContent;

interface

uses SysUtils,Classes,IniFiles;

type
 TDamageType = (dtNil,dtDamgeMelee,dtDamgeDistantion,dtDamageMagic,dtHealMagic);
 TMagicType = (mtNil,mtHoly,mtUnholy,mtNature,mtEarth,mtWater,mtFire);
 TAttackResult = (arNormal,arMissle,arAbsorb,arParry);
 TMagicResist = record
  value: array [mtHoly..mtFire] of byte;
 end;

 TUnitCyclesThread = class(TThread)
  constructor Create;
  protected
   procedure Execute; override;
 end;
 TDestState = (dsOpen,dsClose,dsDestroy);
 TTrigState = (tsOn,tsOff);
 TUnitType = (utPlayer,utEnemy,utNPC);
 TUnitTypeSet = set of TUnitType;
 TOrderType = (otNoTarget,otXY,otUnit,otNil);
 TRawCode = string[4];
 TOrder = record
  _type: TOrderType;
  order: string;
  target: integer;
  x,y: integer;
 end;
 TSpell = record
  rawcode: TRawCode;
  lvl: byte;
 end;
 TTwoByte = 0..65535;
 TItem = record
  rawcode: TRawCode;
  counts: TTwoByte;
  used: boolean;
 end;
 TStockItem = record
  rawcode: TRawCode;
  count: integer;
  max: integer;
 end;
 TItemRemoveThread = class(TThread)
  constructor Create(itm: TObject; time: integer);
  protected
   procedure Execute; override;
  private
   FItem: TObject;
   FTime: integer;
 end;
 TItm = class
  HItem: integer;
  x,y: integer;
  locID: integer;
  counts: integer;
  _id: integer;
  constructor Create(rc: TRawCode; locId, nx,ny: integer);
  constructor Create0(own: integer; rc: TRawCode);
  destructor Destroy;
  private
   FRawCode: TRawCode;
   Fowner: integer;
   FRemoveThread: TItemRemoveThread;
   procedure SetOwner(owner: integer);
  public
   property owner: integer read FOwner write SetOwner;
   property RawCode: TRawCode read FRawCode;
   procedure SetOwnerD(owner: integer);
   function GetMainInfo: string;
   function GetInfo: string;
   procedure RunTimer(time: integer);
   procedure SendRemove;
 end;
 TWarAnim = record
  indexed: boolean;
  name: string [30];
  id: byte;
 end;
 TUnit = class
  HUnit: integer;
  posx,posy,posz,yaw: integer;
  health,mana: integer;
  name: string;
  agro: byte;
  Fexp: cardinal;
  order: TOrder;
  locid: integer;
  spells: array [0..10] of TSpell;
  items:  array [0..5] of integer;
  stockitems: array [0..10] of TStockItem;
  _id: integer;
  _type: TUnitType;
  ispause: boolean;
  isinvul: boolean;
  constructor Create(rc: string; locId, x,y,z,ry,h,m: integer; utype: TUnitType);
  procedure SetExp(exp: cardinal); virtual;
  destructor Destroy; override;
  private
   _rawCode: string;
   //Stats
   function FGetMagicResist: TMagicResist;
   function FGetArmor: integer;
   //Other
   function FGetLevel: integer;
   function FReadItemRawCode(index: integer): TRawCode;
   function GetData(field: string): string;
   procedure SetData(field: string; value: string);
   procedure SetDataLoc(field: string; value: string);
   function GetDataLoc(field: string): string;
  public
   DataTable: TStringList;
   DataTableLoc: TStringList;
   Animation: TWarAnim;
   //Stats
   property MagicResist: TMagicResist read FGetMagicResist;
   property Armor: integer read FGetArmor;
   //Other
   property Level: integer read FGetLevel;
   property Data [field: string]: string read GetData write SetData;
   property DataLoc [field: string]: string read GetDataLoc write SetDataLoc;
   property exp: cardinal read Fexp write SetExp;
   property ItemRC [index: integer]: TRawCode read FReadItemRawCode;
   //Data Table
   procedure FlushData(field: string);
   //Animations
   procedure PlayAnim(id: integer); overload;
   procedure PlayAnim(name: string); overload;
   function GetAnimFunc: string;
   //Orders
   procedure SendOrderTarget(x,y: integer; orders: string);
   procedure SendOrderTargetUnit(HUnit: integer; orders: string);
   procedure SendOrder(orders: string);
   //not sorted
   function AddItem(hitem: integer): boolean;
   function DropItem(hitem: integer): boolean;
   property rawCode: string read _rawCode;
   function GetHM: string;
   function GetXY: string;
   //Ability
   function FindAbility(raw: string): integer;
   function AddAbility(raw: string): boolean;
   //Functions
   function HaveItem(hitem: integer): boolean;
   //Stock Item
   function FindStockItem(raw: string): integer;
   procedure AddStockItem(raw: string; count,max: integer);
   procedure RemoveStockItem(raw: string);
 end;
 TEffect = class
  HEffect: integer;
  locid,posx,posy: integer;
  modelname: string;
  attachpoint: string;
  target: integer; //HUNIT
  toUnit: boolean;
  _id: integer;
  constructor Create;
  destructor Destroy;
  public
    procedure AddEffect(model: string; lid,x,y: integer);
    procedure AddEffectTarget(model,attach: string; un: integer);
    function GetEffectFunc: string;
 end;
 TDest = class
  HDest: integer;
  posx,posy,posz,yaw: integer;
  locid: integer;
  state: TDestState;
  show: boolean;
  _id: integer;
  constructor Create(rc: string; locID, x,y,z,ry: integer);
  destructor Destroy;
  private
   _rawCode: string;
  public
   property rawCode: string read _rawCode;
   function GetMainInfo: string;
 end;
 TGroup = class
  constructor Create;
  destructor Destroy;
  public
   HGroup: integer;
   index: integer;
   owner: integer;
   Funits: array [0..4] of integer;
   function AddUnit(hunit: integer): boolean;
   function GetCount: integer;
   procedure RemUnit(hunit: integer);
   procedure SendAll(msg: string);
 end;
 TUnitGroup = class
  count: integer;
  HUGroup: integer;
  constructor Create;
  destructor Destroy;
  private
   FUnits: array of TUnit;
   function FGetUnit(index: integer): TUnit;
  public
   id: integer;
   procedure AddUnit(u: TUnit);
   procedure RemoveUnit(u: TUnit);
   function IsHaveUnit(u: TUnit): boolean;
   property Units[index: integer]: TUnit read FGetUnit; default;
   function GetUnitFromRange(locId,x,y: integer; radius: real; utypeset: TUnitTypeSet): boolean;
   function GetRandomUnit: TUnit; 
 end;
 TRect = class
  x,y,width,height: integer;
  locid: integer;
  HRect: integer;
  eventscript: string;
  _id: integer;
  constructor Create(locid,px,py,w,h: integer);
  destructor Destroy;
  private
   units: TUnitGroup;
  public
   function IsPointIn(px,py: integer): boolean;
   function IsEnter(u: TUnit): boolean;
   function IsOut(u: TUnit): boolean;
   function AddUnit(u: TUnit): boolean;
   procedure RemUnit(u: TUnit);
   procedure PickUnits(utypes: TUnitTypeSet);
 end;
 TAnswerData = record
  text: string;
  actionId: byte;
  actionData: string;
  handle: integer;
 end;
 TDialogTableData = record
  title: string;
  AnswerList: array of TAnswerData;
  AnswerCount: integer;
 end;
 TDialogTable = class
  DialogTableList: array of TDialogTableData;
  DialogTableCount: integer;
  HDialog: integer;
  _id: integer;
  _path: string;
  scriptevent: string;
  char: TUnit; //TCharA
  currentDT: integer;
  constructor Create;
  destructor Destroy;
  public
   function LoadDialogTable(path: string): boolean;
   procedure SetTitle(title: string);
   procedure ShowDialog(dt: integer);
   procedure onBtnClick(btnid: integer);
   procedure ChangeButtonText(tableid,btnid: integer; newText: string);
   function FindButtonOnBtnID(btnid: integer): integer;
   function ConvertString(str: string): string;
 end;
 TMap = class
    name: string;
    id: byte;
    constructor Create(name: string);
    destructor Destroy;
 end;
 
 function MathDamageValue(attacker,target: TUnit; dtype: TDamageType; mtype: TMagicType; var value: integer): TAttackResult;

function FindUnitByHUNIT(hunit: integer): integer;
function StrToUType(s: string): TUnitType;
function StrToUnitTypeSet(str: string): TUnitTypeSet;
function DestStateToInt(state: TDestState): integer;
function IntToDestState(state: integer): TDestState;
function FindItemByHITEM(hitem: integer): integer;
function ExpToLevel(exp: integer): integer;
function FindGroupByHGROUP(hgroup: integer): integer;
function FindDialogByHDIALOG(hdialog: integer): integer;
function FindDestByHDEST(hdest: integer): integer;
function FindRectByHRECT(hrect: integer): integer;
function FindUGroupByHUGROUP(hugroup: integer): integer;
function FindEffectByHEFFECT(heffect: integer): integer;


var appPath: string;
    {Units Block}
    Units: array of TUnit;
    UnitsCount: integer = 0;
    {Destructable Block}
    Dests: array of TDest;
    DestCount: integer = 0;
    {Items Block}
    Items: array of TItm;
    ItemCount: integer = 0;
    {Groups Block}
    Groups: array of TGroup;
    GroupCount: integer = 0;
    {Dialog Tables Block}
    Dialogs: array of TDialogTable;
    DialogCount: integer = 0;
    {Rect Block}
    RectList: array of TRect;
    RectCount: integer = 0;
    {Map Block}
    MapList: array of TMap;
    MapCount: integer = 0;
    {Unit Group}
    UGroupList: array of TUnitGroup;
    UGroupCount: integer = 0;
    {Effect}
    EffectList: array of TEffect;
    EffectCount: integer = 0;

const
    c_OutRange: real = 1500;
    c_damageradius: real = 500;
    //c_MaxAttackRadius: real = 800;


implementation

uses CharDataBase,Main,MgScriptNatives,OtherDataBase;

var HUnits: array of boolean;
    HUnitsCount: integer = 0;

    HDests: array of boolean;
    HDestsCount: integer = 0;

    HItems: array of boolean;
    HItemsCount: integer = 0;

    HGroups: array of boolean;
    HGroupCount: integer = 0;

    HDialogs: array of boolean;
    HDialogCount: integer = 0;

    HRects: array of boolean;
    HRectCount: integer = 0;

    HUGroups: array of boolean;
    HUGroupCount: integer = 0;

    HEffects: array of boolean;
    HEffectCount: integer = 0;

function StrToUnitTypeSet(str: string): TUnitTypeSet;
 var i: integer;
begin
 Result := [];
 for i:= 1 to Length(str) do
  if str[i] = 'P' then Result := Result + [utPlayer] else
  if str[i] = 'N' then Result := Result + [utNPC] else
  if str[i] = 'E' then Result := Result + [utEnemy];
end;

function MathDamageValue(attacker,target: TUnit; dtype: TDamageType; mtype: TMagicType; var value: integer): TAttackResult;
 var cL,tL: integer;
begin
  Result := arNormal;
  // 
  //   
  if (dtype in [dtDamageMagic,dtHealMagic]) and (mtype <> mtNil) then
  begin
    // 
    if dtype = dtDamageMagic then
    begin
      tL := attacker.Level;
      cL := target.Level;
      if (cL - tL >= 6) then
      begin
       Result := arAbsorb;
       Exit;
      end;
      value := Round(value*(1-target.MagicResist.value[mtype]/(150+(cL-tL)*25)));
      if value <= 0 then
      begin
       Result := arAbsorb;
       Exit;
      end;
    end;
  end;
end;

function FindEffectByHEFFECT(heffect: integer): integer;
  var i: integer;
begin
  for i:= 0 to EffectCount-1 do
    if EffectList[i].HEffect = heffect then
    begin
      Result := i;
      exit;
    end;
  Result := -1;
end;

function FindRectByHRECT(hrect: integer): integer;
  var i: integer;
begin
  for i:= 0 to RectCount-1 do
   if RectList[i].HRect = hrect then
   begin
    Result := i;
    Exit;
   end;
  Result := -1;
end;

function FindUGroupByHUgroup(hugroup: integer): integer;
  var i: integer;
begin
  for i:= 0 to UGroupCount-1 do
   if UGroupList[i].HUGroup = hugroup then
   begin
    Result := i;
    Exit;
   end;
  Result := -1;
end;

function FindDestByHDEST(hdest: integer): integer;
 var i: integer;
begin
 for i:= 0 to DestCount-1 do
  if Dests[i].HDest = hdest then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

function ExpToLevel(exp: integer): integer;
 var lastExp,needExp: integer;
 const NeedHeroXP = 400; //  :  (NeedHeroXP)
       PrevFactor = 1.3; //  :    (NeedHeroXPFormulaA)
       LevelFactor = 50; //  :   (NeedHeroXPFormulaB)
       ConstantFactor = 33; //  :   (NeedHeroXPFormulaC)
begin
 Result:= 1;
 lastExp:= 0;
 needExp:= NeedHeroXP;
 while not ((lastExp <= exp) and (exp < needExp)) do
 begin
  inc(Result);
  lastExp:= needExp;
  needExp:= Round(lastExp*PrevFactor+LevelFactor*(Result+1)+ConstantFactor);
 end;
end;

function StrToUType(s: string): TUnitType;
begin
 if s = 'player' then Result:= utPlayer else
 if s = 'enemy' then Result:= utEnemy else
 if s = 'npc' then Result:= utnpc else
  Result := utNPC;
end;

function GetFreeHEffect: integer;
  var i: integer;
begin
  for i := 0 to HEffectCount-1 do
    if not HEffects[i] then
    begin
      Result := i;
      HEffects[i] := true;
      Exit;
    end;
  SetLength(HEffects,HEffectCount+1);
  Result := HEffectCount;
  inc(HeffectCount);
  Heffects[Result] := true;
end;

function GetFreeHRect: integer;
 var i: integer;
begin
  for i := 0 to HRectCount-1 do
    if not HRects[i] then
    begin
     Result := i;
     HRects[i] := true;
     Exit;
    end;
  SetLength(HRects,HRectCount+1);
  Result := HRectCount;
  HRects[Result] := true;
  inc(HRectCount);
end;

function GetFreeHUGroup: integer;
 var i: integer;
begin
  for i := 0 to HUGroupCount-1 do
    if not HUgroups[i] then
    begin
     Result := i;
     HUGroups[i] := true;
     Exit;
    end;
  SetLength(HUgroups,HUGroupCount+1);
  Result := HUGroupCount;
  HUgroups[Result] := true;
  inc(HUGroupCount);
end;

function GetFreeHdialog: integer;
 var i: integer;
begin
 for i:= 0 to HDialogCount-1 do
  if not HDialogs[i] then
  begin
   Result:= i;
   HDialogs[i]:= true;
   exit;
  end;
 SetLength(HDialogs,HDialogCount+1);
 Result:= HDialogCount;
 HDialogs[HDialogCount]:= true;
 inc(HDialogCount);
end;

function GetFreeHgroup: integer;
 var i: integer;
begin
 for i:= 0 to HGroupCount-1 do
  if not HGroups[i] then
  begin
   Result:= i;
   HGroups[i]:= true;
   exit;
  end;
 SetLength(HGroups,HGroupCount+1);
 Result:= HGroupCount;
 HGroups[HGroupCount]:= true;
 inc(HGroupCount);
end;

function GetFreeHUnit: integer;
 var i: integer;
begin
 for i:= 0 to HUnitsCount-1 do
  if not HUnits[i] then
  begin
   Result:= i;
   HUnits[i]:= true;
   exit;
 end;
 SetLength(HUnits,HUnitsCount+1);
 Result:= HUnitsCount;
 HUnits[HUnitsCount]:= true;
 inc(HUnitsCount);
end;

function GetFreeHDest: integer;
 var i: integer;
begin
 for i:= 0 to HDestsCount-1 do
  if not HDests[i] then
  begin
   Result:= i;
   HDests[i]:= true;
   exit;
 end;
 Result:= HDestsCount;
 SetLength(HDests,HDestsCount+1);
 HDests[HDestsCount]:= true;
 inc(HDestsCount);
end;

function GetFreeHItem: integer;
 var i: integer;
begin
 for i:= 0 to HItemsCount-1 do
  if not HItems[i] then
  begin
   Result:= i;
   HItems[i]:= true;
   exit;
 end;
 Result:= HItemsCount;
 SetLength(HItems,HItemsCount+1);
 HItems[HItemsCount]:= true;
 inc(HItemsCount);
end;

{ TUnit }

constructor TUnit.Create(rc: string; locId, x,y,z, ry, h, m: integer; utype: TUnitType);
 var i: integer;
begin
 inherited Create;
 DataTable:= TStringList.Create;
 DataTableLoc:= TStringList.Create;
 _RawCode:= rc;
 agro:= 10;
 self.locid := locid;
 posx:= x;
 posy:= y;
 posz:= z;
 yaw:= ry;
 health:= h;
 mana:= m;
 name:=  UnDataBase[rc].name;
 order._type:= otNil;
 HUnit:= GetFreeHUnit;
 order.target:= -1;
 for i:= 0 to 5 do
  Items[i]:= -1;
 _type:= utype;
 _id:= UnitsCount;
 inc(UnitsCount);
 SetLength(Units,UnitsCount);
 Units[_id]:= Self;
end;

function TUnit.GetHM: string;
begin
 Result:= IntToStr(HUnit)+','+IntToStr(Health)+','+IntToStr(mana);
end;

function TUnit.GetXY: string;
begin
 Result:= IntToStr(HUnit)+','+IntToStr(posX)+','+IntToStr(posY)+','+IntToStr(Yaw);
end;

destructor TUnit.Destroy;
begin
 inherited Destroy;
 dec(UnitsCount);
 Units[_id]:= Units[UnitsCount];
 Units[_id]._id:= _id;
 HUnits[HUnit]:= false;
end;

function FindUnitByHUNIT(hunit: integer): integer;
 var i: integer;
begin
 Result:= -1;
 for i:= 0 to UnitsCount-1 do
  if Units[i].HUnit = hunit then
  begin
   Result:= i;
   exit;
 end;
end;

function TUnit.FindAbility(raw: string): integer;
 var i: integer;
begin
 for i:= 0 to 10 do
  if spells[i].rawcode = raw then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

function TUnit.AddAbility(raw: string): boolean;
 var i: integer;
begin
 Result:= true;
 for i:= 0 to 10 do
 begin
  if spells[i].rawcode = raw then
  begin
   spells[i].lvl:= 1;
   exit;
  end;
  if spells[i].rawcode = '' then
  begin
   spells[i].rawcode:= raw;
   spells[i].lvl:= 1;
   exit;
  end;
 end;
 Result:= false;
end;

procedure TUnit.AddStockItem(raw: string; count,max: integer);
 var i: integer;
begin
 for i:= 0 to 10 do
  if stockitems[i].rawcode = '' then
  begin
   stockitems[i].rawcode:= raw;
   stockitems[i].count:= count;
   stockitems[i].max:= max;
   exit;
  end;
end;

function TUnit.FindStockItem(raw: string): integer;
 var i: integer;
begin
 Result:= -1;
 for i:= 0 to 10 do
  if stockitems[i].rawcode = raw then
  begin
   Result:= i;
   exit;
  end;
end;

procedure TUnit.RemoveStockItem(raw: string);
 var i: integer;
begin
 i:= FindStockItem(raw);
 if i < 0 then exit;
 stockitems[i].rawcode:= '';
 stockitems[i].count:= 0;
 stockitems[i].max:= 0;
end;

function TUnit.AddItem(hitem: integer): boolean;
 var i: integer;
begin
 Result:= true;
 for i:= 0 to 5 do
  if Items[i] < 0 then
  begin
   Items[i]:= hitem;
   exit;
  end;
 Result:= false;
end;

function TUnit.DropItem(hitem: integer): boolean;
 var i,j: integer;
begin
 Result:= true;
 for i:= 0 to 5 do
  if Items[i] = hitem then
  begin
   j:= FindItemByHITEM(hitem);
   WorldContent.Items[j].RunTimer(60);
   Items[i]:= -1;
   exit;
  end;
 Result:= false;
end;

function TUnit.HaveItem(hitem: integer): boolean;
 var i: integer;
begin
 Result:= true;
 for i:= 0 to 5 do
  if Items[i] = hitem then exit;
 Result:= false;
end;

function TUnit.FReadItemRawCode(index: integer): TRawCode;
begin
 Result:= 'nil';
 if items[index] >= 0 then
 begin
  Result:= WorldContent.Items[items[index]].RawCode; 
 end;
end;

procedure TUnit.SetExp(exp: cardinal);
begin
 FExp:= exp;
end;

function TUnit.GetData(field: string): string;
begin
 Result:= DataTable.Values[field];
end;

procedure TUnit.SetData(field, value: string);
begin
 if DataTable.IndexOfName(field) < 0 then
  DataTable.Add(field+'='+value)
 else
  DataTable.Values[field]:= value;
end;

procedure TUnit.SendOrderTarget(x, y: integer; orders: string);
begin
 order._type:= otXY;
 order.target:= -1;
 order.order:= orderS;
 order.x:= x;
 order.y:= y;
 Main.SendAllPlayersOnLoc('orderxy('+IntToStr(HUNIT)+','+orders+','+IntToStr(x)+','+IntToStr(y)+')',locid,nil);
end;

procedure TUnit.SendOrderTargetUnit(HUnit: integer; orders: string);
begin
 order._type:= otUnit;
 order.order:= orderS;
 order.target:= HUnit;
 Main.SendAllPlayersOnLoc('orderunit('+IntToStr(Self.HUnit)+','+orderS+','+IntToStr(HUNIT)+')',locid,nil);
end;

procedure TUnit.SendOrder(orders: string);
begin
 order._type:= otNoTarget;
 order.order:= orderS;
 order.target:= -1;
 Main.SendAllPlayersOnLoc('order('+IntToStr(HUNIT)+','+orderS+')',locid,nil);
end;

function TUnit.GetAnimFunc: string;
begin
 if Animation.indexed then
  Result:= 'setanimind('+IntToStr(HUNIT)+','+IntToStr(Animation.id)+')'
 else
  Result:= 'setanim('+IntToStr(HUNIT)+',"'+Animation.name+'")'
end;

procedure TUnit.PlayAnim(id: integer);
begin
 Animation.indexed:= true;
 Animation.id:= id;
 Main.SendAllPlayersOnLoc('setanimind('+IntToStr(HUNIT)+','+IntToStr(Animation.id)+')',locid,nil);
end;

procedure TUnit.PlayAnim(name: string);
begin
 Animation.indexed:= false;
 Animation.name:= name;
 Main.SendAllPlayersOnLoc('setanim('+IntToStr(HUNIT)+',"'+Animation.name+'")',locid,nil);
end;

procedure TUnit.FlushData(field: string);
 var i: integer;
begin
 i:= DataTable.IndexOfName(field);
 if i >= 0 then
  DataTable.Delete(i)
 else begin
  i := DataTableLoc.IndexOfName(field);
  if i < 0 then Exit;
  DataTableLoc.Delete(i);
 end;
end;

function TUnit.FGetArmor: integer;
 var i,j: integer;
     itm: TItm;
begin
  Result := 0;
  for i:= 0 to 5 do
  begin
    j := FindItemByHITEM(items[i]);
    if j >= 0 then
    begin
      itm := WorldContent.Items[j];
      Result := Result + ItmDataBase.ItemsRaw[itm.RawCode].Armor; 
    end;
  end;
end;

function TUnit.FGetMagicResist: TMagicResist;
  var i,j: integer;
      itm: TItm;
      k: TMagicType;
begin
  FillChar(Result,SizeOf(TMagicresist),0);
  for i := 0 to 5 do
  begin
    j := FindItemByHITEM(items[i]);
    if j >= 0 then
    begin
      itm := WorldContent.items[j];
      for k := mtHoly to mtFire do
        Result.value[k] := Result.value[k] + ItmDataBase.ItemsRaw[itm.RawCode].MagicResist.value[k];
    end;
  end;
end;

function TUnit.FGetLevel: integer;
begin
 Result := ExpToLevel(exp);
end;

function TUnit.GetDataLoc(field: string): string;
begin
 Result:= DataTableLoc.Values[field]; 
end;

procedure TUnit.SetDataLoc(field, value: string);
begin
 if DataTableLoc.IndexOfName(field) < 0 then
  DataTableLoc.Add(field+'='+value)
 else
  DataTableLoc.Values[field]:= value;
end;

{ TDest }

function DestStateToInt(state: TDestState): integer;
begin
 if state = dsOpen then Result:= 0 else
 if state = dsClose then Result:= 1 else
 if state = dsDestroy then Result:= 2 else
  Result := 0;
end;

function IntToDestState(state: integer): TDestState;
begin
 Result := dsOpen;
   case state of
     0: Result:= dsOpen;
     1: Result:= dsClose;
     2: Result:= dsDestroy;
   end;
end;

constructor TDest.Create(rc: string; locID, x, y, z, ry: integer);
begin
 inherited Create;
 HDest:= GetFreeHDest;
 Self.locid := locid;
 posx:= x;
 posy:= y;
 posz:= z;
 yaw:= ry;
 state:= dsClose;
 _rawCode:= rc;
 _id:= DestCount;
 inc(DestCount);
 SetLength(Dests,DestCount);
 Dests[_id]:= Self;
end;

destructor TDest.Destroy;
begin
 HDests[HDest]:= false;
 dec(DestCount);
 Dests[_id]:= Dests[DestCount];
 Dests[_id]._id:= _id;
 inherited Destroy;
end;

function TDest.GetMainInfo: string;
begin
 Result:= IntToStr(HDest)+','+_rawCode+','+IntToStr(posx)+','+IntToStr(posy)+','+IntToStr(posz)+','+IntToStr(yaw)+',1,1';
end;

{ TUnitGroup }

procedure TUnitGroup.AddUnit(u: TUnit);
 var index: integer;
begin
 index:= count;
 inc(count);
 SetLength(FUnits,count);
 FUnits[index]:= u;
end;

constructor TUnitGroup.Create;
begin
 inherited Create;
 HUGroup := GetFreeHUGroup;
 id := UGroupCount;
 inc(UGroupCount);
 SetLength(UGroupList,UGroupCount);
 UGroupList[id] := Self;
 count:= 0;
end;

destructor TUnitGroup.Destroy;
begin
 HUGroups[HUGroup]:= false;
 dec(UGroupCount);
 UGroupList[id] := UGroupList[UGroupCount];
 UGroupList[id].id := id;
 inherited Destroy;
end;

function TUnitGroup.FGetUnit(index: integer): TUnit;
begin
 Result:= FUnits[index];
end;

function GetUnitPointDist(u: TUnit; x,y: integer): real;
 var dx,dy: integer;
begin
 dx:= u.posx-x;
 dy:= u.posy-y;
 Result:= SqRt(dx*dx+dy*dy);
end;

function TUnitGroup.GetRandomUnit: TUnit;
begin
 Result:= FUnits[Random(count)];
end;

function TUnitGroup.GetUnitFromRange(locId,x, y: integer; radius: real; utypeset: TUnitTypeSet): boolean;
 var i: integer;
     d: real;
begin
 for i:= 0 to UnitsCount-1 do
  if WorldContent.Units[i].locid = locID then
    if WorldContent.Units[i]._type in utypeset then
    begin
     if WorldContent.Units[i] is TCharA then
      if not (WorldContent.Units[i] as TCharA).status then continue;
     d:= GetUnitPointDist(WorldContent.Units[i],x,y);
     if d < radius then
      AddUnit(WorldContent.Units[i]);
    end;
end;

function TUnitGroup.IsHaveUnit(u: TUnit): boolean;
 var i: integer;
begin
 Result:= true;
 for i:= 0 to count-1 do
  if FUnits[i] = u then exit;
 Result:= false;
end;

procedure TUnitGroup.RemoveUnit(u: TUnit);
 var index: integer;
begin
 for index:= 0 to count-1 do
  if FUnits[index] = u then break;
 dec(count);
 FUnits[index]:= FUnits[count];
 FUnits[count]:= nil;
end;

{ TRect }

function TRect.AddUnit(u: TUnit): boolean;
begin
 Result := u.locid = locid;
 if not Result then exit;
 Result:= IsPointIn(u.posx,u.posy);
 if not Result then exit;
 Result:= not Units.IsHaveUnit(u);
 if Result then Units.AddUnit(u);
end;

constructor TRect.Create(locid, px, py, w, h: integer);
begin
 units:= TUnitGroup.Create;
 HRect := GetFreeHRect;
 Self.locid := locid;
 x:= px;
 y:= py;
 width:= w;
 height:= h;
 _id := RectCount;
 inc(RectCount);
 SetLength(RectList,RectCount);
 RectList[_id] := Self;
end;

destructor TRect.Destroy;
begin
 dec(RectCount);
 RectList[_id] := RectList[RectCount];
 RectList[_id]._id := _id;
 units.Destroy;
 inherited Destroy;
end;

function TRect.IsEnter(u: TUnit): boolean;
 var dx,dy: integer;
begin
 Result:= false;
 if u.locid <> locid then Exit;
 if not Units.IsHaveUnit(u) then
 begin
  Result := ((u.posx > x) and (u.posx < x+width)) and ((u.posy < y) and (u.posy > y-height))
 end;
end;

function TRect.IsOut(u: TUnit): boolean;
 var dx,dy: integer;
begin
 Result:= false;
 if Units.IsHaveUnit(u) then
 begin
  Result := (u.locid <> locid) or (not (((u.posx > x) and (u.posx < x+width)) and ((u.posy < y) and (u.posy > y-height))))
 end;
end;

function TRect.IsPointIn(px, py: integer): boolean;
 var dx,dy: integer;
begin
 dx:= px-x;
 dy:= py-y;
 Result:= ((px > x) and (px < x+width)) and ((py < y) and (py > y-height)); 
end;

procedure TRect.PickUnits(utypes: TUnitTypeSet);
 var i: integer;
begin
 for i:= 0 to UnitsCount-1 do
  if Units[i]._type in utypes then
   AddUnit(Units[i]);
end;

procedure TRect.RemUnit(u: TUnit);
begin
 Units.RemoveUnit(u);
end;

{ TItm }

constructor TItm.Create(rc: TRawCode; locID, nx, ny: integer);
begin
 inherited Create;
 FRawCode:= rc;
 HItem:= GetFreeHItem;
 _id:= ItemCount;
 inc(ItemCount);
 SetLength(Items,ItemCount);
 Items[_id]:= Self;
 Fowner:= -1;
 x:= nx;
 y:= ny;
 FRemoveThread := nil;
end;

constructor TItm.Create0(own: integer; rc: TRawCode);
begin
 inherited Create;
 FRawCode:= rc;
 HItem:= GetFreeHItem;
 _id:= ItemCount;
 inc(ItemCount);
 SetLength(Items,ItemCount);
 Items[_id]:= Self;
 Fowner:= own;
 FRemoveThread := nil;
end;

destructor TItm.Destroy;
 var i,j: integer;
begin
  dec(ItemCount);
  Items[_id]:= Items[ItemCount];
  Items[_id]._id:= _id;
  HItems[HItem]:= false;
  j := FindUnitByHUNIT(FOwner);
  if j >= 0 then
   for i:= 0 to 5 do
    if Units[j].items[i] = HItem then
    begin
     Units[j].items[i] := -1;
    end;
  if Assigned(FRemoveThread) then
  begin
   FRemoveThread.Terminate;
   FRemoveThread := nil;
  end;
  inherited;
end;

function TItm.GetInfo: string;
begin
 Result:= IntToStr(HItem)+','+IntToStr(owner);
end;

function TItm.GetMainInfo: string;
begin
 Result:= IntToStr(x)+','+IntToStr(y)+','+rawCode+','+IntToStr(HItem);
end;

function FindItemByHITEM(hitem: integer): integer;
 var i: integer;
begin
 for i:= 0 to ItemCount-1 do
  if Items[i].HItem = hitem then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

procedure TItm.RunTimer(time: integer);
begin
 FRemoveThread:= TItemRemoveThread.Create(Self,time);
end;

procedure TItm.SendRemove;
 var i: integer;
begin
 if Owner >= 0 then
 begin
  i := FindUnitByHUNIT(Owner);
  if i >= 0 then
   Units[i].DropItem(HItem);
 end;
 SendAllPlayersOnLoc('remitm('+IntToStr(HITEM)+')',locId,nil);
end;

procedure TItm.SetOwner(owner: integer);
 var i: integer;
begin
 i:= FindUnitByHUNIT(owner);
 if i < 0 then
 begin
  FOwner:= -1;
  exit;
 end;
 if not Units[i].HaveItem(Hitem) then
  if Units[i].AddItem(Hitem) then
  begin
   FOwner:= owner;
   if Assigned( FRemoveThread ) then
   begin
    FRemoveThread.Terminate;
    FRemoveThread := nil;
   end;
  end;
end;

procedure TItm.SetOwnerD(owner: integer);
begin
 FOwner:= owner;
end;

{ TGroup }

function TGroup.AddUnit(hunit: integer): boolean;
 var i: integer;
begin
 Result:= true;
 for i:= 0 to 4 do
  if FUnits[i] < 0 then
  begin
   FUnits[i]:= hunit;
   exit;
  end;
 Result:= false;
end;

constructor TGroup.Create;
 var i: integer;
begin
 inherited Create;
 HGroup:= GetFreeHGroup;
 index:= GroupCount;
 inc(GroupCount);
 SetLength(Groups,GroupCount);
 Groups[index]:= Self;
 for i:= 0 to 4 do
  FUnits[i]:= -1;
end;

destructor TGroup.Destroy;
begin
 dec(GroupCount);
 HGroups[HGroup]:= false;
 Groups[index]:= Groups[GroupCount];
 Groups[index].index:= index;
 inherited Destroy;
end;

function TGroup.GetCount: integer;
 var i: integer;
begin
 Result:= 0;
 for i:= 0 to 4 do
  if FUnits[i] >= 0 then inc(Result);
end;

procedure TGroup.RemUnit(hunit: integer);
 var i: integer;
begin
 for i:= 0 to 4 do
  if FUnits[i] = hunit then
  begin
   FUnits[i]:= -1;
   exit;
  end;
end;

procedure TGroup.SendAll(msg: string);
 var i,j: integer;
begin
 for i:= 0 to 4 do
 begin
  j:= FindUnitByHUNIT(FUnits[i]);
  if j >= 0 then
   (Units[j] as TCharA).client.Send(msg);
 end;
end;

function FindGroupByHGROUP(hgroup: integer): integer;
 var i: integer;
begin
 for i:= 0 to GroupCount-1 do
  if Groups[i].HGroup = HGroup then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;


{ TUnitCyclesThread }

constructor TUnitCyclesThread.Create;
begin
 inherited Create(true);
 FreeOnTerminate:= true;
 Resume;
end;

function GetUnitsDist(u1,u2: TUnit): real;
 var dx,dy: real;
begin
 dx:= (u1.posx-u2.posx);
 dy:= (u1.posy-u2.posy);
 Result:= Sqrt(dx*dx+dy*dy);
end;

procedure TUnitCyclesThread.Execute;
 var i,j: integer;
     d: real;
     grp: TUnitGroup;
     un: TUnit;
     rct: TRect;
begin
 for i:= 0 to UnitsCount-1 do
 begin
  //=====================================
  //For Rects
  //  Check rect events (onEnter, onLeave)
  //=====================================
  for j := 0 to RectCount-1 do
  begin
    rct := RectList[j];
    if rct.IsEnter(Units[i]) then
    begin
      rct.units.AddUnit(Units[i]);
      MGScriptNatives.RunMGScript(rct.eventscript,'onEnter',IntToStr(rct.HRect)+','+IntToStr(Units[i].HUnit));
    end else
    if rct.IsOut(Units[i]) then
    begin
      rct.units.RemoveUnit(Units[i]);
      MGScriptNatives.RunMGScript(rct.eventscript,'onLeave',IntToStr(rct.HRect)+','+IntToStr(Units[i].HUnit));
    end;
  end;
  //=====================================
  //For Enemy
  //=====================================
  if (Units[i]._type = utEnemy) then
   if (Units[i].order.target >= 0) then
   begin
    j:= FindUnitByHUnit(Units[i].order.target);
    if (j >= 0) and (j < UnitsCount) then
    begin
     d:= GetUnitsDist(Units[i],Units[j]);
     if (d > c_OutRange) or (Units[j].health <= 0) then
     begin
      grp:= TUnitGroup.Create;
      grp.GetUnitFromRange(Units[i].locid,Units[i].posx,Units[i].posy,c_damageradius,[utPlayer]);
      if grp.count > 0 then
      begin
       un:= grp.GetRandomUnit;
       Units[i].SendOrdertargetUnit(un.HUnit,'attack');
      end
      else
      begin
       Units[i].health:= UnDataBase[Units[i].RawCode].maxhealth;
       Main.SendAllPlayers('newHM('+Units[i].GetHM+')',nil);
       Units[i].SendOrderTarget(Units[i].posx,Units[i].posy,'move');
      end;
      grp.Free;
     end;
    end;
   end
   else
   begin //Radius Agro
    grp:= TUnitGroup.Create;
    grp.GetUnitFromRange(Units[i].locid,Units[i].posx,Units[i].posy,c_damageradius,[utPlayer]);
    if grp.count > 0 then
    begin
     un:= grp.GetRandomUnit;
     Units[i].SendOrderTargetUnit(un.HUnit,'attack');
    end;
    grp.Free;
   end;
 end;
end;

{ TItemRemoveThread }

constructor TItemRemoveThread.Create(itm: TObject; time: integer);
begin
 inherited Create(true);
 FreeOnTerminate:= true;
 FItem:= itm;
 FTime:= time;
 Resume;
end;

procedure TItemRemoveThread.Execute;
begin
 Sleep(FTime*1000);
 if (FItem as TItm).owner >= 0 then exit;
 (FItem as TItm).SendRemove;
 (FItem as TItm).Destroy;
end;

{ TDialogTable }

procedure TDialogTable.ChangeButtonText(tableid,btnid: integer; newText: string);
 var i: integer;
begin
 for i:= 0 to DialogTableList[tableid].AnswerCount-1 do
  if DialogTableList[tableid].AnswerList[i].handle = btnid then
  begin
   DialogTableList[tableid].AnswerList[i].text := newText;
   exit;
  end;
end;

function TDialogTable.ConvertString(str: string): string;
begin
 Result := StringReplace(str,'%name%',char.name,[rfReplaceAll]);
 Result := StringReplace(Result,'%class%',ChClassDataBase.classs[(char as TCharA).classid].name,[rfReplaceAll]);
end;

constructor TDialogTable.Create;
begin
 inherited Create;
 _id:= DialogCount;
 inc(DialogCount);
 SetLength(Dialogs,DialogCount);
 Dialogs[_id]:= Self;
 HDialog:= GetFreeHDialog;
end;

destructor TDialogTable.Destroy;
begin
 dec(DialogCount);
 Dialogs[_id]:= Dialogs[DialogCount];
 Dialogs[_id]._id:= _id;
 HDialogs[HDialog]:= false;
 inherited Destroy;
end;

function TDialogTable.FindButtonOnBtnID(btnid: integer): integer;
 var i: integer;
begin
 for i:= 0 to DialogTableList[currentDT].AnswerCount-1 do
  if DialogTableList[currentDT].AnswerList[i].handle = btnid then
  begin
   Result:= i;
   exit;
  end;
 Result:= 0;
end;

function TDialogTable.LoadDialogTable(path: string): boolean;
 var dt: TIniFile;
     i,j: integer;
     buf,section: string;
begin
 Result:= false;
 path:= Main.pathlist.dialogtable+path;
 if not FileExists(path) then exit;
 dt:= TIniFile.Create(apppath+path);
 DialogTableCount:= dt.ReadInteger('main','count',0);
 SetLength(DialogTableList,DialogTableCount);
 for i:= 1 to DialogTableCount do
  with DialogTableList[i-1] do
  begin
   section:= 'table_'+IntToStr(i);
   title:= dt.ReadString(section,'title','<no title>');
   AnswerCount:= dt.ReadInteger(section,'count',0);
   SetLength(AnswerList,AnswerCount);
   for j:= 1 to DialogTableList[i-1].AnswerCount do
   begin
    buf:= 'answer_'+IntToStr(j)+'_';
    AnswerList[j-1].text:= dt.ReadString(section,buf+'text','<no text>');
    AnswerList[j-1].actionId:= dt.ReadInteger(section,buf+'action',0);
    AnswerList[j-1].actionData:= dt.ReadString(section,buf+'actiondata','');
    AnswerList[j-1].handle:= dt.ReadInteger(section,buf+'index',0);
   end;
  end;
 _path:= path;
 dt.Free;
 Result:= true;
end;

function FindDialogByHDIALOG(hdialog: integer): integer;
 var i: integer;
begin
 for i:= 0 to DialogCount-1 do
  if Dialogs[i].HDialog = hdialog then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

procedure TDialogTable.onBtnClick(btnid: integer);
 var btn: integer;
begin
 btn:= FindButtonOnBtnId(btnid);
 if btn < 0 then exit;
 with DialogTableList[currentDT].AnswerList[btn] do
 begin
  case actionId of
   0: Destroy;
   1: ShowDialog(StrToInt(actionData)-1);
   2: RunMGScript(pathlist.script+actionData,'onClick',IntToStr(char.HUnit)+','+IntToStr(HDialog)+','+IntToStr(handle));
   3: begin
        (Char as TCharA).client.Send('d_t('+IntToStr(HDialog)+',"'+actionData+'")');
        (Char as TCharA).client.Send('d_s('+IntToStr(HDialog)+')');
      end;
  end;
 end;
end;

procedure TDialogTable.SetTitle(title: string);
begin
 (char as TCharA).client.Send('d_t('+IntToStr(HDialog)+',"'+ConvertString(title)+'")');
end;

procedure TDialogTable.ShowDialog(dt: integer);
 var handle: string;
     i: integer;
begin
 currentDT:= dt;
 with (char as TCharA).client do
 begin
  handle:= IntToStr(HDialog);
  Send('d_c('+handle+')');
  Send('d_t('+handle+',"'+ConvertString(DialogTableList[dt].title)+'")');
  for i:= 0 to DialogTableList[dt].AnswerCount-1 do
   Send('d_a('+handle+','+IntToStr(DialogTableList[dt].AnswerList[i].handle)+',"'+ConvertString(DialogTableList[dt].AnswerList[i].text)+'")');
  Send('d_s('+handle+')');
 end;
end;

{ TMap }

function FindFreeMapID: integer;
 var i: integer;
begin
 for i:= 0 to MapCount-1 do
  if MapList[i] = nil then
  begin
    Result := i;
    Exit;
  end;
 Result := MapCount;
 inc(MapCount);
 SetLength(MapList,MapCount);
end;

constructor TMap.Create(name: string);
begin
 inherited Create;
 Self.name := name;
 id := FindFreeMapID;
 MapList[id] := Self;
end;

destructor TMap.Destroy;
begin
 MapList[id] := nil;
 inherited Destroy;
end;

{ TEffect }

procedure TEffect.AddEffect(model: string; lid, x, y: integer);
begin
  toUnit := false;
  modelname := model;
  locid := lid;
  posx := x;
  posy := y;
end;

procedure TEffect.AddEffectTarget(model, attach: string; un: integer);
begin
  toUnit := true;
  modelname := model;
  attachpoint := attach;
  target := un;
end;

constructor TEffect.Create;
begin
  inherited Create;
  _id:= EffectCount;
  inc(EffectCount);
  HEffect := GetFreeHEffect;
  SetLength(EffectList,EffectCount);
  EffectList[_id] := Self;
end;

destructor TEffect.Destroy;
begin
  dec(EffectCount);
  EffectList[_id] := EffectList[EffectCount];
  EffectList[_id]._id := _id;
  HEffects[HEffect] := false;
  inherited Destroy;
end;

function TEffect.GetEffectFunc: string;
begin
  if toUnit then
    Result := 'enewu("'+modelname+'","'+attachpoint+'",'+IntToStr(target)+','+IntToStr(HEffect)+')'
  else
    Result := 'enewxy("'+modelname+'",'+IntToStr(posX)+','+IntToStr(posy)+','+IntToStr(HEffect)+')';
end;

end.
