unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Buttons, ExtCtrls, XPMan, OleCtrls, SHDocVw, StdCtrls, Gauges, PngImage,
  IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP, ShellApi,
  SFMPQApi, Registry, Menus;

const
 CWM_downloadend  = $0191;
 CWM_installend   = $0192;
 CWM_setprogress  = $0193;
 CWM_setcaption   = $0194;
 CWM_seterror     = $0195;
 CWM_downloaderr  = $0196;

type
 TImgBtn = class(TImage)
  private
   Ffocused: boolean;
   procedure MouseEnter(var msg: TMessage); message CM_MOUSEENTER;
   procedure MouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
   procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
   procedure MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
   procedure SetCaption(caption: TCaption);
  public
   picDef,picInc,picClick: string;
   property onClick;
   procedure Complete;
  published
   constructor Create(AOwner: TComponent); override;
 end;
 TDownloadThread = class(TThRead)
  constructor Create(url,path: string; Downloader: TIdHTTP);
  protected
   procedure Execute; override;
  private
   FUrl,Fpath: string;
   FDownloader: TIdHTTP;
 end;
  TClientForm = class(TForm)
    XPManifest1: TXPManifest;
    WebBrowser: TWebBrowser;
    ProgressBar: TGauge;
    Downloader: TIdHTTP;
    IdHttp: TIdHTTP;
    StatusLabel: TLabel;
    BcgImg: TImage;
    WD_Timer: TTimer;
    SizeLabel: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure DownloaderWorkBegin(Sender: TObject; AWorkMode: TWorkMode; const AWorkCountMax: Integer);
    procedure DownloaderWork(Sender: TObject; AWorkMode: TWorkMode; const AWorkCount: Integer);
    procedure StatusLabelClick(Sender: TObject);
    procedure BcgImgMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure BcgImgMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure WD_TimerTimer(Sender: TObject);
    procedure DownloaderWorkEnd(Sender: TObject; AWorkMode: TWorkMode);
  private
    { Private declarations }
  public
   WD_Offset: TPoint;
   WD_Drag: boolean;
   btnStartGame: TImgBtn;
   btnExit: TImgBtn;
   InstallError: integer;
   VersionInfoList: TStrings;
   ConnectError: boolean;
   DowloadErrorMsg: string;
   function LoadSettings: boolean;
   procedure StartUpdate;
   procedure InstallUpdate;
   procedure CheckVersion;
   function CorrectVersion: boolean;
   procedure btnStartGameClick(Sender: TObject);
   procedure StartGame;
   procedure EndUpdate(var mes: TMessage); message CWM_downloadend;
   procedure EndInstall(var mes: TMessage); message CWM_installend;
   procedure SetProgress(var mes: TMessage); message CWM_setprogress;
   procedure SetInstallError(var mes: TMessage); message CWM_seterror;
   procedure DownloadError(var mes: TMessage); message CWM_downloaderr;
   procedure BtnExitOnClick(Sender: TObject);
  end;
  TInstallThread = class(TThread)
   constructor Create(WinHandle: cardinal);
   protected
    procedure Execute; override;
   private
    FWinHandle: cardinal; 
  end;

var
  ClientForm: TClientForm;
  NewClientVersion: string;// = '0.83.1';
  LastClientVersion: string;
  ClientVersion: string = '0.83.0';
  launcherpath: string;
  ErrorDetails: string;

implementation

{$R *.dfm}

{ TImgBtn }

function ByteToSize (b: integer): string;
  var r: real;
      m: integer;
      size: string;
begin
  size := ' B';
  m := b div 1024;
  if m <= 0 then
    if b < 512 then
    begin
      Result := IntToStr(b) + size;
      Exit;
    end else begin
      Result := FloatToStrF(b / 1024, ffFixed, 8, 2) + ' kB';
      Exit;
    end;
  size := ' kB';
  r := b / 1024;
  b := b div 1024;
  m := m div 1024;
  if m <= 0 then
  begin
    Result := FloatToStrF(r, ffFixed, 8, 2) + size;
    Exit;
  end;
  size := ' MB';
  r := b / 1024;
  b := b div 1024;
  m := m div 1024;
  if m <= 0 then
  begin
    Result := FloatToStrF(r, ffFixed, 8, 2) + size;
    Exit;
  end;
  size := ' GB';
  r := b / 1024;
  Result := FloatToStrF(r, ffFixed, 0, 2) + size;
end;

procedure TImgBtn.Complete;
begin
 Picture.LoadFromFile(picDef);
end;

constructor TImgBtn.Create(AOwner: TComponent);
begin
  inherited;
 onMouseDown:= MouseDown;
 onMouseUp:= MouseUp;
end;

procedure TImgBtn.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 Picture.LoadFromFile(picClick);
end;

procedure TImgBtn.MouseEnter(var msg: TMessage);
begin
 Ffocused:= true;
  Picture.LoadFromFile(picInc);
end;

procedure TImgBtn.MouseLeave(var msg: TMessage);
begin
 Ffocused:= false;
 Picture.LoadFromFile(picDef);
end;

procedure TImgBtn.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 if FFocused then
  Picture.LoadFromFile(picInc)
 else
  Picture.LoadFromFile(picDef);
end;

{FORM}

function BitmapToRgn(Image: TBitmap): HRGN;
var
  TmpRgn: HRGN;
  x, y: integer;
  ConsecutivePixels: integer;
  CurrentPixel: TColor;
  CreatedRgns: integer;
  CurrentColor: TColor;
begin
  CreatedRgns := 0;
  Result := CreateRectRgn(0, 0, Image.Width, Image.Height);
  inc(CreatedRgns);

  if (Image.Width = 0) or (Image.Height = 0) then
    exit;

  for y := 0 to Image.Height - 1 do
  begin
    CurrentColor := Image.Canvas.Pixels[0,y];
    ConsecutivePixels := 1;
    for x := 0 to Image.Width - 1 do
    begin
      CurrentPixel := Image.Canvas.Pixels[x, y];

      if CurrentColor = CurrentPixel then
        inc(ConsecutivePixels)
      else
      begin
        //    
        if CurrentColor = clWhite then
        begin
          TmpRgn := CreateRectRgn(x - ConsecutivePixels, y, x, y + 1);
          CombineRgn(Result, Result, TmpRgn, RGN_DIFF);
          inc(CreatedRgns);
          DeleteObject(TmpRgn);
        end;
        CurrentColor := CurrentPixel;
        ConsecutivePixels := 1;
      end;
    end;

    if (CurrentColor = clWhite) and (ConsecutivePixels > 0) then
    begin
      TmpRgn := CreateRectRgn(x-ConsecutivePixels, y, x, y+1);
      CombineRgn(Result, Result, TmpRgn, RGN_DIFF);
      inc(CreatedRgns);
      DeleteObject(TmpRgn);
    end;
  end;
end;

procedure TClientForm.FormCreate(Sender: TObject);
 var Mask: TBitmap;
     Png: TPngObject;
     data: TMemoryStream;
     txtF: file;
begin
 if not LoadSettings then
 begin
  ShowMessage('   .      .'+#13#10+'  ');
  Application.Terminate;
  Exit;
 end;
 try
  VersionInfoList:= TStringList.Create;
  data:= TMemoryStream.Create;
  IdHttp.Get('http://war-cry.ucoz.ru/ForgottenLand/info.txt',data);
  data.Position:= 0;
  VersionInfoList.LoadFromStream(data);
  AssignFile(txtF,launcherpath+'/Forgotten Land Client/realmlist.wtf');
  Rewrite(txtF,1);
  BlockWrite(txtF,Pointer(VersionInfoList.Values['server'])^,Length(VersionInfoList.Values['server']));
  CloseFile(txtF);
  data.Destroy;
  data:= nil;
 except
  on E: Exception do
  begin
   ShowMessage('    .'+#13#10+'System information:'+#13#10+'Type: '+E.ClassName+#13#10+'Message: '+E.Message);
   ConnectError:= true;
  end; 
 end; 

 BcgImg.Picture.LoadFromFile('Pictures\bcg.png');
 BcgImg.Transparent:= true;
 Mask:= TBitmap.Create;
 png:= TPngObject.Create;
 try
  png.LoadFromFile('Pictures\bcg.png');
  mask.Assign(png);
  Windows.SetWindowRgn(Handle,BitmapToRgn(Mask),true);
 finally
  Mask.Free;
  png.Free;
 end;
 InstallError:= 0;
 WebBrowser.Navigate('http://war-cry.ucoz.ru/ForgottenLand/News.html');
 StatusLabel.Caption:= '';
 btnStartGame:= TImgBtn.Create(Self);
 with btnStartGame do
 begin
  Parent:= Self;
  Autosize:= true;
  Left:= 660;
  Top:= 62;
  picInc:= 'Pictures\IndNG.png';
  picDef:= 'Pictures\ActNG.png';
  picClick:= 'Pictures\PressNG.png';
  onClick:= btnStartGameClick;
  Hint:= ' ';
  ShowHint:= true;
  Complete;
 end;
 btnExit:= TImgBtn.Create(Self);
 with btnExit do
 begin
  Parent:= Self;
  Autosize:= true;
  Left:= 660;
  Top:= 416;
  picInc:= 'Pictures\IndE.png';
  picDef:= 'Pictures\ActE.png';
  picClick:= 'Pictures\PressE.png';
  onClick:= btnExitOnClick;
  Hint:= '';
  ShowHint:= true;
  Complete;
 end;
 CheckVersion;
end;

procedure TClientForm.DownloaderWorkBegin(Sender: TObject; AWorkMode: TWorkMode; const AWorkCountMax: Integer);
begin
  ProgressBar.Progress:= 0;
  ProgressBar.MaxValue:= AWorkCountMax;
end;

procedure TClientForm.DownloaderWork(Sender: TObject; AWorkMode: TWorkMode; const AWorkCount: Integer);
begin
  ProgressBar.Progress:= AWorkCount;
  SizeLabel.Visible := true;
  SizeLabel.Caption := ByteToSize(AWorkCount)+ ' \ ' + ByteToSize(ProgressBar.MaxValue);
end;

procedure TImgBtn.SetCaption(caption: TCaption);
begin
  
end;

{ TDownloadThread }

constructor TDownloadThread.Create(url, path: string; Downloader: TIdHTTP);
begin
 inherited Create(true);
 FUrl:= url;
 FPath:= path;
 FDownloader:= Downloader;
 FreeOnTerminate:= true;
 Resume;
end;

procedure TDownloadThread.Execute;
 var data: TMemoryStream;
     err: boolean;
begin
  err := false;
  try
    try
      data:= TMemoryStream.Create;
      FDownloader.Get(Furl,Data);
      data.Position:= 0;
      data.SaveToFile(Fpath);
    except
      on E: Exception do
      begin
        err := true;
        ClientForm.DowloadErrorMsg := 'Class: '+E.ClassName+#13#10+'Message: '+E.Message;
        SendMessage(ClientForm.Handle, CWM_downloaderr,0,0);
        if Assigned(data) then
          data.Free;
      end;
    end;
  finally
    if not err then
    begin
      SendMessage(ClientForm.Handle,CWM_downloadend,0,0);
      data.Free;
    end;
  end;
end;

procedure TClientForm.StartUpdate;
 var s: string;
begin
 s:= StringReplace(NewClientVersion,'.','_',[rfReplaceAll]);
 WebBrowser.Navigate('http://war-cry.ucoz.ru/ForgottenLand/update_'+s+'.html');
 if not FileExists('patches\patch_'+s+'.mpq') then
 begin
  TDownloadThread.Create('http://war-cry.ucoz.ru/ForgottenLand/patch_'+s+'.mpq','patches\patch_'+s+'.mpq',Downloader);
  StatusLabel.Caption:= '  ('+NewClientVersion+')';
 end
 else
  InstallUpdate;
end;

procedure TClientForm.CheckVersion;
begin
 NewClientVersion:= VersionInfoList.Values[ClientVersion];
 LastClientVersion:= VersionInfoList.Values['version'];
 if LastClientVersion <> ClientVersion then
 begin
  BtnStartGame.picInc:= 'Pictures\IndU.png';
  BtnStartGame.picDef:= 'Pictures\ActU.png';
  BtnStartGame.picClick:= 'Pictures\PressU.png';
  BtnStartGame.Complete;
  BtnStartGame.Hint:= ' ';
  BtnStartGame.Hint:= ' '
 end
 else
 begin
  BtnStartGame.picInc:= 'Pictures\IndNG.png';
  BtnStartGame.picDef:= 'Pictures\ActNG.png';
  BtnStartGame.picClick:= 'Pictures\PressNG.png';
  BtnStartGame.Complete;
  BtnStartGame.Hint:= ' ';
 end;
end;

function TClientForm.CorrectVersion: boolean;
begin
Result:= ClientVersion = LastClientVersion
end;

procedure TClientForm.btnStartGameClick(Sender: TObject);
begin
 if ConnectError then
 begin
  ShowMessage('      .'+#13#10+'  ');
  exit;
 end;
 if CorrectVersion then
  StartGame
 else
  StartUpdate;
end;

procedure TClientForm.StartGame;
begin
 ShellExecute(Handle,nil,PChar(launcherpath+'\Forgotten Land Client\launcher.exe'),'-runonce',PChar(launcherpath+'\Forgotten Land Client\'),SW_HIDE);
end;

procedure TClientForm.EndUpdate(var mes: TMessage);
begin
 StatusLabel.Caption:= ' ';
 InstallUpdate; 
end;

procedure TClientForm.InstallUpdate;
begin
 TInstallThread.Create(Handle);
end;

procedure TClientForm.StatusLabelClick(Sender: TObject);
 var s: string;
begin
 s:= '';
 case InstallError of
  0: s:= '   (ok)';
  1: s:= '    (file not found: "'+ErrorDetails+'")';
  2: s:= '    (error num: 2; "'+ErrorDetails+'")';
  3: s:= '      (error num: 3; "'+ErrorDetails+'")';
  4: s:= '    (error num: 4; "'+ErrorDetails+'")';
 end;
 ShowMessage(s);
end;

procedure TClientForm.BtnExitOnClick(Sender: TObject);
begin
 Application.Terminate;
end;

function TClientForm.LoadSettings: boolean;
 var reg: TRegistry;
begin
 Result:= false;
 reg:= TRegistry.Create;
 reg.RootKey:= HKEY_CURRENT_USER;
 if not reg.KeyExists('Software\King Ksel corp\Forgotten Land')  then
 begin
  ShowMessage('  .     ');
  reg.Free;
  exit;
 end;
 reg.OpenKey('Software\King Ksel corp\Forgotten Land',false);
 ClientVersion:= reg.ReadString('version');
 launcherpath:= reg.ReadString('clientpath');
 reg.Free;
 Result:= true;
end;

{ TInstallThread }

constructor TInstallThread.Create(WinHandle: cardinal);
begin
 inherited Create(false);
 FWinHandle:= WinHandle;
 FreeOnTerminate:= true;
 Resume;
end;

function GetTableValue(str: string; pos: integer): string;
 var i,p: integer;
begin
 p:= 1;
 Result:= '';
 for i:= 1 to Length(str) do
 begin
  if str[i] = '=' then inc(p) else if p = pos then Result:= Result + str[i];
  if p > pos then exit;
 end;
end;

procedure TInstallThread.Execute;
 var s: string;
     InstallList: TStrings;
     installmode: integer;
     i,k: integer;
     buf: string;
     hMPQ,hUpdMPQ: MPQHANDLE;
     hFile,hUpdFile: THandle;
     MemStr: TMemoryStream;
     buff: Pointer;
     size: LongInt;
     f: file;
     par: array [0..3] of string;
     reg: TRegistry;
 const
  IM_nil      = 0; //No install mode
  IM_file     = 1; //Simpe files
  IM_update   = 2; //Update file in game archive
  IM_location = 3; //Load location in game archive
  IM_reg      = 4; //Write new register values
begin 
 //StatusLabel.Caption:= '  ('+NewClientVersion+')';
 SendMessage(FWinHandle,CWM_setprogress,0,0);
 //ProgressBar.Progress:= 0;
 s:= StringReplace(NewClientVersion,'.','_',[rfReplaceAll]);
 if not FileExists('patches\patch_'+s+'.mpq') then
 begin
  //StatusLabel.Caption:= ' ';
  SendMessage(FWinHandle,CWM_seterror,0,1);
  ErrorDetails := 'File not found: patches\patch_'+s+'.mpq';
  exit;
 end;
 hMPQ:= MpqOpenArchiveForUpdate(PChar('patches\patch_'+s+'.mpq'), MOAU_OPEN_EXISTING, 65535);
 if hMPQ > $85200000 then
 begin
  //StatusLabel.Caption:= ' ';
  SendMessage(FWinHandle,CWM_seterror,0,2);
  ErrorDetails:= 'patches\patch_'+s+'.mpq';
  exit;
 end;
 if not SFileOpenFileEx(hMPQ,'install_info.txt',0,hFile) then
 begin
  //StatusLabel.Caption:= ' ';
  SendMessage(FWinHandle,CWM_seterror,0,3);
  ErrorDetails := 'install_info';
  exit;
 end;
 InstallList:= TStringList.Create;
 size:= SFileGetFileInfo(hFile, SFILE_INFO_SIZE);
 GetMem(buff,size);
 ZeroMemory(buff,size);
 if not SFileReadFile(hFile,buff,size,nil,nil) then
 begin
  //StatusLabel.Caption:= ' ';
  SendMessage(FWinHandle,CWM_seterror,0,4);
  exit;
 end;
 MemStr := TMemoryStream.Create;
 MemStr.WriteBuffer(buff^, size);
 MemStr.Position := 0;
 InstallList.LoadFromStream(MemStr);
 MemStr.Free;
 FreeMem(buff);
 //InstallList.LoadFromFile('install_info.txt');
 {Load Patch File}
 InstallMode:= IM_NIL;
 SFileCloseFile(hFile);
 SendMessage(FWinHandle,CWM_setprogress,1,InstallList.Count-1);
 //ProgressBar.MaxValue:= InstallList.Count-1;
 for i:= 0 to InstallList.Count-1 do
 begin
  SendMessage(FWinHandle,CWM_setprogress,0,i);
  buf:= InstallList[i];
  if buf = '' then continue;
  if buf[1] = '[' then InstallMode:= IM_nil;
  case InstallMode of
   IM_Nil: begin
            if buf = '[file]' then InstallMode:= IM_File;
            if buf = '[update]' then InstallMode:= IM_Update;
            if buf = '[location]' then InstallMode:= IM_Location;
            if buf = '[reg]' then InstallMode:= IM_Reg;
           end; 
   IM_File: begin
              {File
               syntax: archive_file_name=path}
               k:= Pos('=',buf);
               par[1]:= Copy(buf,1,k-1);
               Delete(buf,1,k);
               par[2]:= buf;
               par[2]:= StringReplace(par[2],'%client%',LauncherPath,[rfReplaceAll]);
               //Export File
               if not SFileOpenFileEx(hMPQ,PChar(par[1]),0,hFile) then
               begin
                //StatusLabel.Caption:= ' ';
                SendMessage(FWinHandle,CWM_seterror,0,3);
                ErrorDetails := par[1];
                exit;
               end;
               size:= SFileGetFileInfo(hFile, SFILE_INFO_SIZE);
               GetMem(buff,size);
               if not SFileReadFile(hFile,buff,size,nil,nil) then
               begin
                //StatusLabel.Caption:= ' ';
                SendMessage(FWinHandle,CWM_seterror,0,4);
                ErrorDetails := par[2];
                exit;
               end;
               par[3]:= ExtractFilePath(par[2]);
               if not DirectoryExists(par[3]) then
                ForceDirectories(par[3]);
               AssignFile(f,par[2]);
               Rewrite(f,1);
               BlockWrite(f,buff^,size);
               CloseFile(f);
               FreeMem(buff);
               SFileCloseFile(hFile);
            end;
   IM_Update: begin
               par[1] := '';
               par[2] := '';
               k:= Pos('=',buf);
               if k > 0 then
               begin
                par[1]:= Copy(buf,1,k-1);
                Delete(buf,1,k);
                par[2]:= buf;
                par[2]:= StringReplace(par[2],'%client%',LauncherPath,[rfReplaceAll]);
               end;
               //open archive for update
               if par[1] = 'open' then
               begin
                hUpdMPQ:= MpqOpenArchiveForUpdate(PChar(par[2]),MOAU_OPEN_EXISTING,65535);
                //close archive after update
                if hUpdMPQ > $85200000 then
                begin
                 SendMessage(FWinHandle,CWM_seterror,0,3);
                 ErrorDetails := par[2];
                 exit;
                end;
                continue;
               end;
               if par[1] = 'close' then
               begin
                MpqCompactArchive(hUpdMPQ);
                SFileCloseArchive(hUpdMPQ);
                continue;
               end;
               //Write new file
               if (par[1] <> '') and (par[2] <> '') then
               begin
                if not SFileOpenFileEx(hMPQ,PChar(par[1]),0,hUpdFile) then
                begin
                 SendMessage(FWinHandle,CWM_seterror,0,3);
                 ErrorDetails := par[1];
                 exit;
                end;
                size:= SFileGetFileInfo(hUpdFile, SFILE_INFO_SIZE);
                GetMem(buff,size);
                if not SFileReadFile(hUpdFile,buff,size,nil,nil) then
                begin
                 SendMessage(FWinHandle,CWM_seterror,0,4);
                 ErrorDetails := par[1];
                 exit;
                end;
                MpqAddFileFromBuffer(hUpdMPQ,buff,size,PChar(par[2]),MAFA_REPLACE_EXISTING);
                FreeMem(buff);
                SFileCloseFile(hUpdFile);
               end;
              end;
   IM_Reg: begin
            reg:= TRegistry.Create;
            reg.RootKey:= HKEY_LOCAL_MACHINE;
            par[0]:= GetTableValue(buf,1);//key (directory)
            par[1]:= GetTableValue(buf,2);//value name
            par[2]:= GetTableValue(buf,3);//value type
            par[3]:= GetTableValue(buf,4);//value
            reg.OpenKey(par[0],true);
            k:= StrToIntDef(par[2],0);
            case k of
             0: reg.WriteString(par[1],par[3]);
             1: reg.WriteInteger(par[1],StrToIntDef(par[3],-1));
             2: reg.WriteBool(par[1],StrToBool(par[3]));
             3: reg.WriteFloat(par[1],StrToFloatDef(par[3],0));
             4: reg.WriteBinaryData(par[1],Pointer(par[3])^,Length(par[3]));
            end;
            reg.Free;
           end;
  end;
 end;
 InstallList.Free;
 SFileCloseArchive(hMPQ);
 SendMessage(FWinHandle,CWM_installend,0,0);
end;

procedure TClientForm.EndInstall(var mes: TMessage);
 var reg: TRegistry;
begin
 ClientVersion:= NewClientVersion;
 reg:= Tregistry.Create;
 reg.RootKey:= HKEY_CURRENT_USER;
 reg.OpenKey('Software\King Ksel corp\Forgotten Land',false);
 reg.WriteString('version',ClientVersion);
 reg.Free;
 InstallError:= 0;
 if not CorrectVersion then
  CheckVersion
 else
 begin
  BtnStartGame.picInc:= 'Pictures\IndNG.png';
  BtnStartGame.picDef:= 'Pictures\ActNG.png';
  BtnStartGame.picClick:= 'Pictures\PressNG.png';
  BtnStartGame.Complete;
  BtnStartGame.Hint:= ' ';
  StatusLabel.Caption:= '  ';
 end;
end;

procedure TClientForm.SetProgress(var mes: TMessage);
begin
 case mes.WParam of
  0: ProgressBar.Progress:= mes.LParam;
  1: ProgressBar.MaxValue:= mes.LParam;
 end;
end;

procedure TClientForm.SetInstallError(var mes: TMessage);
begin
 StatusLabel.Caption := ' (  )';
 InstallError:= mes.LParam;
end;

procedure TClientForm.BcgImgMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 WD_Drag:= true;
 WD_Timer.Enabled:= true;
 WD_Offset.X:= X;
 WD_Offset.Y:= Y;
end;

procedure TClientForm.BcgImgMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 WD_Drag:= false;
 WD_Timer.Enabled:= false;
end;

procedure TClientForm.WD_TimerTimer(Sender: TObject);
 var mousepos: TPoint;
begin
 mousepos:= Mouse.CursorPos;
 Left:= mousepos.X-WD_Offset.X;
 Top:= mousepos.Y-WD_Offset.Y; 
end;

procedure TClientForm.DownloaderWorkEnd(Sender: TObject; AWorkMode: TWorkMode);
begin
  SizeLabel.Visible := false;
end;

procedure TClientForm.DownloadError(var mes: TMessage);
  var s, ver: string;
begin
  StatusLabel.Caption := '  ';
  SizeLabel.Visible := false;
  s := '   ,   '+#13#10+'    ,         :'+#13#10;
  ver := StringReplace(NewClientVersion,'.','_',[rfReplaceAll]);  
  s := s + 'http:\\war-cry.ucoz.ru\ForgottenLand\patch_'+ver+'.mpq'+#13#10;
  s := s + '   :'+#13#10;
  s := s + DowloadErrorMsg;
  MessageDlg(s, mtError, [mbOK], 0);
end;

end.
