unit SFmpqapi;

interface

uses
  Windows;

const
  dll = 'sfmpq.dll';

type
  SFMPQVERSION = packed record
    Major,
    Minor,
    Revision,
    Subrevision: Word;
  end;

  // MpqInitialize does nothing.  It is only provided for
  // compatibility with MPQ archivers that use lmpqapi.
  function MpqGetVersionString: PChar; stdcall; external dll;

const
  // General error codes
  MPQ_ERROR_MPQ_INVALID         = $85200065;
  MPQ_ERROR_FILE_NOT_FOUND      = $85200066;
  MPQ_ERROR_DISK_FULL           = $85200068; // Physical write file to MPQ failed. Not sure of exact meaning
  MPQ_ERROR_HASH_TABLE_FULL     = $85200069;
  MPQ_ERROR_ALREADY_EXISTS      = $8520006A;
  MPQ_ERROR_BAD_OPEN_MODE       = $8520006C; // When MOAU_READ_ONLY is used without MOAU_OPEN_EXISTING

  MPQ_ERROR_COMPACT_ERROR       = $85300001;

  // MpqOpenArchiveForUpdate flags
  MOAU_CREATE_NEW               = $00;
  MOAU_CREATE_ALWAYS            = $08; // Was wrongly named MOAU_CREATE_NEW
  MOAU_OPEN_EXISTING            = $04;
  MOAU_OPEN_ALWAYS              = $20;
  MOAU_READ_ONLY                = $10; // Must be used with MOAU_OPEN_EXISTING
  MOAU_MAINTAIN_LISTFILE        = $01;

  // MpqAddFileToArchive flags
  MAFA_EXISTS                   = $80000000; // Will be added if not present
  MAFA_UNKNOWN40000000          = $40000000;
  MAFA_MODCRYPTKEY              = $00020000;
  MAFA_ENCRYPT                  = $00010000;
  MAFA_COMPRESS                 = $00000200;
  MAFA_COMPRESS2                = $00000100;
  MAFA_REPLACE_EXISTING         = $00000001;

  // MpqAddFileToArchiveEx compression flags
  MAFA_COMPRESS_STANDARD        = $08; // Standard PKWare DCL compression
  MAFA_COMPRESS_DEFLATE         = $02; // ZLib's deflate compression
  MAFA_COMPRESS_WAVE            = $81; // Standard wave compression
  MAFA_COMPRESS_WAVE2           = $41; // Unused wave compression

  // Flags for individual compression types used for wave compression
  MAFA_COMPRESS_WAVECOMP1       = $80; // Main compressor for standard wave compression
  MAFA_COMPRESS_WAVECOMP2       = $40; // Main compressor for unused wave compression
  MAFA_COMPRESS_WAVECOMP3       = $01; // Secondary compressor for wave compression

  // ZLib deflate compression level constants (used with MpqAddFileToArchiveEx and MpqAddFileFromBufferEx)
  Z_NO_COMPRESSION              = 0;
  Z_BEST_SPEED                  = 1;
  Z_DEFAULT_COMPRESSION         = -1;
  Z_BEST_COMPRESSION            = 9;

  // MpqAddWaveToArchive quality flags
  MAWA_QUALITY_HIGH             = 1;
  MAWA_QUALITY_MEDIUM           = 0;
  MAWA_QUALITY_LOW              = 2;

  // SFileGetFileInfo flags
  SFILE_INFO_BLOCK_SIZE         = $01; // Block size in MPQ
  SFILE_INFO_HASH_TABLE_SIZE    = $02; // Hash table size in MPQ
  SFILE_INFO_NUM_FILES          = $03; // Number of files in MPQ
  SFILE_INFO_TYPE               = $04; // Is MPQHANDLE a file or an MPQ?
  SFILE_INFO_SIZE               = $05; // Size of MPQ or uncompressed file
  SFILE_INFO_COMPRESSED_SIZE    = $06; // Size of compressed file
  SFILE_INFO_FLAGS              = $07; // File flags (compressed, etc.), file attributes if a file not in an archive
  SFILE_INFO_PARENT             = $08; // Handle of MPQ that file is in
  SFILE_INFO_POSITION           = $09; // Position of file pointer in files
  SFILE_INFO_LOCALEID           = $0A; // Locale ID of file in MPQ
  SFILE_INFO_PRIORITY           = $0B; // Priority of open MPQ
  SFILE_INFO_HASH_INDEX         = $0C; // Hash index of file in MPQ

  // SFileListFiles flags
  SFILE_LIST_MEMORY_LIST        = $01; // Specifies that lpFilelists is a file list from memory, rather than being a list of file lists
  SFILE_LIST_ONLY_KNOWN         = $02; // Only list files that the function finds a name for
  SFILE_LIST_ONLY_UNKNOWN       = $04; // Only list files that the function does not find a name for

  SFILE_TYPE_MPQ                = $01;
  SFILE_TYPE_FILE               = $02;

  SFILE_OPEN_HARD_DISK_FILE     = $0000; // Open archive without regard to the drive type it resides on
  SFILE_OPEN_CD_ROM_FILE        = $0001; // Open the archive only if it is on a CD-ROM
  SFILE_OPEN_ALLOW_WRITE        = $8000; // Open file with write access

  SFILE_SEARCH_CURRENT_ONLY     = $00; // Used with SFileOpenFileEx; only the archive with the handle specified will be searched for the file
  SFILE_SEARCH_ALL_OPEN         = $01; // SFileOpenFileEx will look through all open archives for the file

type
  MPQHANDLE = THandle;

  PFILELISTENTRY = ^FILELISTENTRY;
  FILELISTENTRY = packed record
    dwFileExists: DWORD;        // Nonzero if this entry is used
    lcLocale: LCID;             // Locale ID of file
    dwCompressedSize: DWORD;    // Compressed size of file
    dwFullSize: DWORD;          // Uncompressed size of file
    dwFlags: DWORD;             // Flags for file
    szFileName: array[0..259] of Char;
  end;

  PMPQHEADER = ^MPQHEADER;
  PBLOCKTABLEENTRY = ^BLOCKTABLEENTRY;
  PHASHTABLEENTRY = ^HASHTABLEENTRY;
  PMPQFILE = ^MPQFILE;

  MPQHEADER = packed record
    dwMPQID: DWORD;             // "MPQ\x1A" for mpq's, "BN3\x1A" for bncache.dat
    dwHeaderSize: DWORD;        // Size of this header
    dwMPQSize: DWORD;           // The size of the mpq archive
    wUnused0C: WORD;            // Seems to always be 0
    wBlockSize: WORD;           // Size of blocks in files equals 512 << wBlockSize
    dwHashTableOffset: DWORD;   // Offset to hash table
    dwBlockTableOffset: DWORD;  // Offset to block table
    dwHashTableSize: DWORD;     // Number of entries in hash table
    dwBlockTableSize: DWORD;    // Number of entries in block table
  end;

  // Archive handles may be typecasted to this struct so you can access
  // some of the archive's properties and the decrypted hash table and
  // block table directly.
  PMPQARCHIVE = ^MPQARCHIVE;
  MPQARCHIVE = packed record
    lpNextArc: PMPQARCHIVE;     // Pointer to the next ARCHIVEREC struct. Pointer to addresses of first and last archives if last archive
    lpPrevArc: PMPQARCHIVE;     // Pointer to the previous ARCHIVEREC struct. 0xEAFC5E23 if first archive
    szFileName: array[0..259] of Char; // Filename of the archive
    hFile: THandle;             // The archive's file handle
    dwFlags1: DWORD;            // Some flags, bit 1 (0 based) seems to be set when opening an archive from a CD
    dwPriority: DWORD;          // Priority of the archive set when calling SFileOpenArchive
    lpLastReadFile: PMPQFILE;   // Pointer to the last read file's FILEREC struct. Only used for incomplete reads of blocks
    dwUnk: DWORD;               // Seems to always be 0
    dwBlockSize: DWORD;         // Size of file blocks in bytes
    lpLastReadBlock: PByte;     // Pointer to the read buffer for archive. Only used for incomplete reads of blocks
    dwBufferSize: DWORD;        // Size of the read buffer for archive. Only used for incomplete reads of blocks
    dwMPQStart: DWORD;          // The starting offset of the archive
    lpMPQHeader: PMPQHEADER;    // Pointer to the archive header
    lpBlockTable: PBLOCKTABLEENTRY; // Pointer to the start of the block table
    lpHashTable: PHASHTABLEENTRY; // Pointer to the start of the hash table
    dwFileSize: DWORD;          // The size of the file in which the archive is contained
    dwOpenFiles: DWORD;         // Count of files open in archive + 1
    MpqHeader: MPQHEADER;
    dwFlags: DWORD;             //The only flag that should be changed is MOAU_MAINTAIN_LISTFILE
    lpFileName: PChar;
  end;

  // Handles to files in the archive may be typecasted to this struct
  // so you can access some of the file's properties directly.
  MPQFILE = packed record
    lpNextFile: PMPQFILE;       // Pointer to the next FILEREC struct. Pointer to addresses of first and last files if last file
    lpPrevFile: PMPQFILE;       // Pointer to the previous FILEREC struct. 0xEAFC5E13 if first file
    szFileName: array[0..259] of Char; // Filename of the file
    hPlaceHolder: THandle;      // Always 0xFFFFFFFF
    lpParentArc: PMPQARCHIVE;   // Pointer to the ARCHIVEREC struct of the archive in which the file is contained
    lpBlockEntry: PBLOCKTABLEENTRY; // Pointer to the file's block table entry
    dwCryptKey: DWORD;          // Decryption key for the file
    dwFilePointer: DWORD;       // Position of file pointer in the file
    dwUnk1: DWORD;              // Seems to always be 0
    dwBlockCount: DWORD;        // Number of blocks in file
    lpdwBlockOffsets: PDWORD;   // Offsets to blocks in file. There are 1 more of these than the number of blocks
    dwReadStarted: DWORD;       // Set to 1 after first read
    dwUnk2: DWORD;              // Seems to always be 0
    lpLastReadBlock: PByte;     // Pointer to the read buffer for file. Only used for incomplete reads of blocks
    dwBytesRead: DWORD;         // Total bytes read from open file
    dwBufferSize: DWORD;        // Size of the read buffer for file. Only used for incomplete reads of blocks
    dwConstant: DWORD;          // Seems to always be 1
    lpHashEntry: PHASHTABLEENTRY;
    lpFileName: PChar;
  end;

  BLOCKTABLEENTRY = packed record
    dwFileOffset: DWORD;        // Offset to file
    dwCompressedSize: DWORD;    // Compressed size of file
    dwFullSize: DWORD;          // Uncompressed size of file
    dwFlags: DWORD;             // Flags for file
  end;

  HASHTABLEENTRY = packed record
    dwNameHashA: DWORD;         // First name hash of file
    dwNameHashB: DWORD;         // Second name hash of file
    lcLocale: LCID;             // Locale ID of file
    dwBlockTableIndex: DWORD;   // Index to the block table entry for the file
  end;

  // Storm functions implemented by this library
  function SFileOpenArchive(lpFileName: PChar; dwPriority, dwFlags: DWORD; var hMPQ: MPQHANDLE): Bool; stdcall; external dll;
  function SFileCloseArchive(hMPQ: MPQHANDLE): Bool; stdcall; external dll;
  function SFileGetArchiveName(hMPQ: MPQHANDLE; lpBuffer: PChar; dwBufferLength: DWORD): Bool; stdcall; external dll;
  function SFileOpenFile(lpFileName: PChar; var hFile: MPQHANDLE): Bool; stdcall; external dll;
  function SFileOpenFileEx(hMPQ: MPQHANDLE; lpFileName: PChar; dwSearchScope: DWORD; var hFile: MPQHANDLE): Bool; stdcall; external dll;
  function SFileCloseFile(hFile: MPQHANDLE): Bool; stdcall; external dll;
  function SFileGetFileSize(hFile: MPQHANDLE; lpFileSizeHigh: MPQHANDLE): DWORD; stdcall; external dll;
  function SFileGetFileArchive(hFile: MPQHANDLE; var hMPQ: MPQHANDLE): Bool; stdcall; external dll;
  function SFileGetFileName(hFile: MPQHANDLE; lpBuffer: PChar; dwBufferLength: DWORD): Bool; stdcall; external dll;
  function SFileSetFilePointer(hFile: MPQHANDLE; lDistanceToMove: Integer; lplDistanceToMoveHigh: PInteger; dwMoveMethod: DWORD): DWORD; stdcall; external dll;
  function SFileReadFile(hFile: MPQHANDLE; lpBuffer: Pointer; nNumberOfBytesToRead: DWORD; lpNumberOfBytesRead: LPDWORD; lpOverlapped: POverlapped): Bool; stdcall; external dll;
  //function SFileReadFile(HFILE : integer; var buf; size,p4,p5 : integer) : boolean; stdcall; external dll index $010D;
  function SFileSetLocale(nNewLocale: LCID): LCID; stdcall; external dll;
  function SFileGetBasePath(lpBuffer: PChar; dwBufferLength: DWORD): Bool; stdcall; external dll;
  function SFileSetBasePath(lpNewBasePath: PChar): Bool; stdcall; external dll;

  // Extra storm-related functions
  function SFileGetFileInfo(hFile: MPQHANDLE; dwInfoType: DWORD): DWORD; stdcall; external dll;
  function SFileSetArchivePriority(hMPQ: MPQHANDLE; dwPriority: DWORD): BOOL; stdcall; external dll;
  function SFileFindMpqHeader(hFile: THandle): DWORD; stdcall; external dll;
  function SFileListFiles(hMPQ: MPQHANDLE; lpFileLists: PChar; lpListBuffer: PFILELISTENTRY; dwFlags: DWORD): BOOL; stdcall; external dll;

  // Archive editing functions implemented by this library
  function MpqOpenArchiveForUpdate(lpFileName: PChar; dwFlags, dwMaximumFilesInArchive: DWORD): MPQHANDLE; stdcall; external dll;
  function MpqCloseUpdatedArchive(hMPQ: MPQHANDLE; dwUnknown2: DWORD): DWORD; stdcall; external dll;
  function MpqAddFileToArchive(hMPQ: MPQHANDLE; lpSourceFileName, lpDestFileName: PChar; dwFlags: DWORD): BOOL; stdcall; external dll;
  function MpqAddWaveToArchive(hMPQ: MPQHANDLE; lpSourceFileName, lpDestFileName: PChar; dwFlags, dwQuality: DWORD): BOOL; stdcall; external dll;
  function MpqRenameFile(hMPQ: MPQHANDLE; lpcOldFileName, lpcNewFileName: PChar): BOOL; stdcall; external dll;
  function MpqDeleteFile(hMPQ: MPQHANDLE; lpFileName: PChar): BOOL; stdcall; external dll;
  function MpqCompactArchive(hMPQ: MPQHANDLE): BOOL; stdcall; external dll;

  // Extra archive editing functions
  function MpqAddFileToArchiveEx(hMPQ: MPQHANDLE; lpSourceFileName: PChar; lpDestFileName: PChar; dwFlags, dwCompressionType, dwCompressLevel: DWORD): BOOL; stdcall; external dll;
  function MpqAddFileFromBufferEx(hMPQ: MPQHANDLE; lpBuffer: Pointer; dwLength: DWORD; lpFileName: PChar; dwFlags, dwCompressionType, dwCompressLevel: DWORD): BOOL; stdcall; external dll;
  function MpqAddFileFromBuffer(hMPQ: MPQHANDLE; lpBuffer: Pointer; dwLength: DWORD; lpFileName: PChar; dwFlags: DWORD): BOOL; stdcall; external dll;
  function MpqAddWaveFromBuffer(hMPQ: MPQHANDLE; lpBuffer: Pointer; dwLength: DWORD; lpFileName: PChar; dwFlags, dwQuality: DWORD): BOOL; stdcall; external dll;
  function MpqSetFileLocale(hMPQ: MPQHANDLE; lpFileName: PChar; nOldLocale, nNewLocale: LCID): BOOL; stdcall; external dll;


implementation

end.
