unit MGSApi;

interface

  uses SysUtils, Dialogs;

  type
    mgsVarType = (mgsVT_Int, mgsVT_Str, mgsVT_Real, mgsVT_Boolean);
    mgsFuncType = (mgsFT_Action, mgsFT_Return);
    mgsParameterType = (mgsPT_const, mgsPT_ref);
    IMGSVariable = interface
        function GetHandle: cardinal;
        function GetName: PAnsiChar;
        function GetValue: PAnsiChar;
        procedure SetValue( value: PAnsiChar );
        function GetType: mgsVarType;
        function IsConstant: boolean;
    end;
    PMGSValue = ^TMGSValue;
    TMGSValue = record
      isConst: boolean;
        //For constant value
        _Type: mgsVarType;
        _Value: PAnsiChar;
        //For var value
        _Ref: IMGSVariable;
    end;
    IMGSVariableScope = interface
        function GetName: PAnsiChar;
        function AddNewVariable(Name: PAnsiChar; VType: mgsVarType; Value: PAnsiChar; IsConstant: boolean; out Variable: IMGSVariable): cardinal;
        function GetVariable(Handle: cardinal; out Variable: IMGSVariable): cardinal; overload;
        function GetVariable(Name: PAnsiChar; out Variable: IMGSVariable): cardinal; overload;
        function FreeVariable(Handle: cardinal): cardinal;
    end;
    TMGSParameter = record
      _PType: mgsParameterType;
      _VType: mgsVarType;
      _Name: PAnsiChar;
    end;
    TMGSFuncParameterList = array of TMGSParameter;
    TMGSFuncValueList = array of TMGSValue;
    TMGSScriptFunction = function (Values: TMGSFuncValueList): PAnsiChar;
    TMGSScriptProcedure = procedure (Values: TMGSFuncValueList);

    IMGSFunction = interface
        function GetType: mgsFuncType;
        function GetReturnType: mgsVarType;
        function GetName: PAnsiChar;
        function GetHandle: cardinal;
        function VariableScope: IMGSVariableScope;
        function Execute (Values: TMGSFuncValueList; var returnValue: PAnsiChar): cardinal;
    end;

    IMGSFunctionScope = interface
      ['{BE1AEC61-AB95-46EE-9F40-1D9EB6EBEEE9}']
        function GetName: PAnsiChar;
        function AddNewFunction(Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; ParameterCount: integer; Address: Pointer; out Func: IMGSFunction): cardinal; overload;
        function AddNewFunction(Name: PAnsiChar; FType: mgsFuncType; ParameterList: TMGSFuncParameterList; ParameterCount: integer; FunctionCode: PAnsiChar; ParentFunctionScope: IInterface; out Func: IMGSFunction): cardinal; overload;
        function GetFunction(Handle: cardinal; out Func: IMGSFunction): cardinal; overload;
        function GetFunction(Name: PAnsiChar; out Func:  IMGSFunction): cardinal; overload;
    end;
    
    IMGSScript = interface
        function GetName: PAnsiChar;
        procedure Load(text: PAnsiChar);
        procedure Execute();
    end;


    function MGSValueGetValue(Value: PMGSValue): PAnsiChar;
    function MGSValueGetType(Value: PMGSValue): mgsVarType;
    procedure MGSValueSetValue(Value: PMGSValue; val: PAnsiChar);
    procedure MGSValueMove(dest, source: PMGSValue);

    function MGSErrorCodeToString(eCode: cardinal): string;

    procedure MGSDebug(msg: string; eCode: integer); overload;
    procedure MGSDebug(msg: string); overload;

  const
    MGS_E_OK                    = $0;
    MGS_E_UNCORRECT_TYPE        = $1;
    MGS_E_UNCORRECT_COMPARE     = $2;
    MGS_E_DATA_OUT              = $3;
    MGS_E_DATA_NULL             = $4;
    MGS_E_NAME_EMPLOY           = $5;
    MGS_E_DATA_NOTFOUND         = $6;
    MGS_E_UNCORRECT_OPERATION   = $7;

implementation

  function MGSErrorCodeToString(eCode: cardinal): string;
  begin
    case eCode of
      MGS_E_OK:
        Result := 'MGS_E_OK';
      MGS_E_UNCORRECT_TYPE:
        Result := 'MGS_E_UNCORRECT_TYPE';
      MGS_E_UNCORRECT_COMPARE:
        Result := 'MGS_E_UNCORRECT_COMPARE';
      MGS_E_DATA_OUT:
        Result := 'MGS_E_DATA_OUT';
      MGS_E_DATA_NULL:
        Result := 'MGS_E_DATA_NULL';
      MGS_E_NAME_EMPLOY:
        Result := 'MGS_E_NAME_EMPLOY';
      MGS_E_DATA_NOTFOUND:
        Result := 'MGS_E_DATA_NOTFOUNC';
      MGS_E_UNCORRECT_OPERATION:
        Result := 'MGS_E_UNCORRECT_OPERATION';
      else
        Result := 'Unknow error ('+IntToStr(eCode)+')';
    end;
  end;

  procedure MGSDebug(msg: string);
  begin
    ShowMessage(msg);
  end;

  procedure MGSDebug(msg: string; eCode: integer);
  begin
    ShowMessage('#'+IntToStr(eCode) + ': ' + #13#10 + msg);
  end;

  procedure MGSValueMove(dest, source: PMGSValue);
  begin
    dest.isConst := source.isConst;
    dest._Type := source._Type;
    dest._Value := GetMemory(StrLen(source._Value)+1);
    StrMove(dest._Value, source._Value, StrLen(source._Value)+1);
    dest._Ref := source._Ref;
  end;

  procedure MGSValueSetValue(Value: PMGSValue; val: PAnsiChar);
  begin
    if Value.isConst then
      Value._Value := val
    else
      Value._Ref.SetValue(val);
  end;

  function MGSValueGetValue(Value: PMGSValue): PAnsiChar;
  begin
    if Value.isConst then
      Result := Value._Value
    else
      Result := Value._Ref.GetValue;
  end;

  function MGSValueGetType(Value: PMGSValue): mgsVarType;
  begin
    if Value.isConst then
      Result := Value._Type
    else
      Result := Value._Ref.GetType;
  end;

end.
