unit MGScriptNatives;

interface

uses MGScript,Classes,SysUtils,Utils,CharDataBase,WorldContent,OtherDataBase,
     ScriptLog;

type
 {Script Var}
 TVarType = (vtString,vtInteger,vtReal,vtBool,vtNull);
 TVar = class
  name,value: string;
  id: integer;
  t: TVarType;
  constructor New(nname: string; nt: TVarType; nvalue: string);
  constructor LocNew(nname: string; nt: TVarType; nvalue: string);
  destructor Destroy; override;
  destructor LocDestroy;
  private
   function GetArrayValue(index: integer): string;
   procedure SetArrayValue(index: integer; v: string);
  public
   property ArrayValue[index: integer]: string read GetArrayValue write SetArrayValue;
 end;
 {Script}
 TIfStatment = record
  skip: boolean;
  value: boolean;
  number: integer;
 end;
 TRunPoint = record
  name: string;
  pos: integer;
 end;
 TFuncParam = record
  name: string;
  vtype: TVarType;
 end;
 TFunctionInfo = record
  pos: integer;
  returnvtype: TVarType;
  parcount: integer;
  name: string;
  params: array of TFuncParam;
 end;
 PFunction = ^TFunction;
 PMGScript = ^TMGScript;
 TFunction = class
  pos,calledpos: integer;
  From: PFunction;
  Script: PMGScript;
  destructor Destroy; override;
  private
   locvariable: array of TVar;
   locvarcount: integer;
  public
   function FindLocalVariable(name: string): integer;
   procedure Init(funcinfo: TFunctionInfo; params: string);
   procedure AddLocalVariable(name: string; t: TVarType; value: string);
 end;
 TMGScript = class
  constructor Create(spath,rpoint,rparams: string); overload;
  constructor Create(line: string); overload;
  destructor Destroy; override;
  private
   path: string;
   runpoint: string;
   runparams: string;
   RunPoints: array of TRunPoint;
   _RPcount: integer;
   Fvars: array of TVar;
   Fvarcount: integer;
   CalledFunction: TFunction;
   functionlist: array of TFunctionInfo;
   funccount: integer;
   procedure GetRunPoints;
  public
   scriptlist: TStrings;
   procedure AddNewFunction(line: string; spos: integer);
   function FindFunc(name: string): integer;
   procedure Compile;
   function GetValue(parm: string; vt: TVarType):string;
   function Math(line: string): string;
   function StrMath(str: string): string;
   function BoolMath(line: string): string;
   function FindRunPoint(name: string): integer;
   procedure AddLocalVariable(name: string; t: TVarType; value: string);
   function FindLocalVariable(name: string): integer;
   //Check Values
   function CheckInt(str: string): boolean;
   function CheckReal(str: string): boolean;
   function CheckChar(id: integer): boolean;
   function CheckRange(int,min,max: integer; rangename: string): boolean;
 end;

 TScriptThread = class(TThRead)
  public
   constructor Create(path,point,params: string);
   destructor Destroy; override;
  protected
   procedure Execute; override;
  private
   FScript: TMGScript;
 end;

 procedure RunMGScript(path,runpoint,params: string);

implementation

uses Main;            

var _varcount: integer = 0;
    _vars: array of TVar;

const _operators = ['*','/','-','+','^'];
      _booloperators = [#1,#2,#3,#4,#5,#6,#7,'=','<','>'];
      _breketO: char = '{';
      _breketC: char = '}';

procedure Replace(var line: string; start,oldlen: integer; newvalue: string);
 var buf: string;
begin
 buf:= line;
 Delete(buf,1,start+oldlen);
 line:= Copy(line,1,start-1)+newvalue+buf;
end;

function UnitGetDirective(rawcode: TRawcode; var s: string): boolean;
begin
 Result:= false;
 if Length(s) < 1 then exit;
 if s[1] <> '$' then exit;
 Delete(s,1,1);
 s:= LowerCase(s);
 Result:= true;
 if s = 'maxlife' then
 begin
  s:= IntToStr(UnDataBase.UnitsRaw[rawcode].maxhealth);
  exit;
 end;
 if s = 'maxmana' then
 begin
  s:= IntToStr(UnDataBase.UnitsRaw[rawcode].maxmana);
  exit;
 end;
 ScriptLogForm.AddLog('"'+s+'" is not unit directive',-1);
end;

//***********************
// Script Variables
//***********************
constructor TVar.New(nname: string; nt: TVarType; nvalue: string);
begin
 inherited Create;
 name:= nname;
 t:= nt;
 value:= nvalue;
 id:= _varcount;
 inc(_varcount);
 SetLength(_vars,_varcount);
 _vars[id]:= Self;
end;

function TVar.GetArrayValue(index: integer): string;
 var i,j: integer;
     head: string;
begin
 Result:= '';
 head:= #1+IntToStr(index)+#1;
 i:= Pos(head,value);
 if i <= 0 then exit;
 for j:= i+Length(Head) to Length(value) do
 begin
  if value[j] = #1 then exit;
  Result:= Result + value[j];
 end;
end;

procedure TVar.SetArrayValue(index: integer; v: string);
 var i,j,len: integer;
     head: string;
begin
 head:= #1+IntToStr(index)+#1;
 j:= Pos(head,value);
 if j <= 0 then
 begin
  value:= value+head+v;
  exit;
 end;
 len:= 0;
 for i:= j+Length(head) to Length(value) do
  if value[i] = #1 then break else inc(len);
 Replace(value,j+Length(head),len-1,v);
end;

constructor TVar.LocNew(nname: string; nt: TVarType; nvalue: string);
begin
 inherited Create;
 name:= nname;
 t:= nt;
 value:= nvalue;
end;

destructor TVar.Destroy;
begin
 inherited Destroy;
 dec(_varcount);
 _vars[id]:= _vars[_varcount];
 _vars[id].id:= id;
end;

destructor TVar.LocDestroy;
begin
 inherited Destroy;
end;

//Find Variable by name
function FindVarId(name: string):integer;
 var i: integer;
begin
 for i:= 0 to _varcount-1 do
  if _vars[i].name = name then
  begin
   result:= i;
   exit;
 end;
 result:= -1;
end;

//Is Variable
function IsVariable(par: string):boolean;
begin
 Result:= false;
 if Length(par) < 1 then exit;
 Result:= par[1] = '@';
end;

{function StrToReal(s: string): real;
 var e: integer;
begin
 Result:= StrToFloat(s);
// Val(s,Result,e);
end;}

function ToIntVar(str: string): string;
begin
 Result:= IntToStr(Round(StrToReal(str)));
end;

function RealToStr(r: real): string;
begin
 Str(r:0:3,Result);
end;

//Convert Script Word To Variable Type  
function ConvertVarType(s: string): TVarType;
begin
 Result := vtString;
 if s = '[str]' then Result:= vtString;
 if s = '[int]' then Result:= vtInteger;
 if s = '[rea]' then Result:= vtReal;
 if s = '[bol]' then Result:= vtBool;
end;


{ TFunction }

function TFunction.FindLocalVariable(name: string): integer;
 var i: integer;
begin
 for i:= 0 to locvarcount-1 do
  if locvariable[i].name = name then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

procedure TFunction.AddLocalVariable(name: string; t: TVarType; value: string);
 var i: integer;
begin
 i:= FindLocalVariable(name);
 if i < 0 then
 begin
  i:= locvarcount;
  inc(locvarcount);
  SetLength(locvariable,locvarcount);
  locvariable[i] := TVar.LocNew(name,t,value);
 end;
end;

procedure TFunction.Init(funcinfo: TFunctionInfo; params: string);
 var i: integer;
begin
 pos:= funcinfo.pos;
 for i:= 0 to funcinfo.parcount-1 do
  AddLocalVariable(funcinfo.params[i].name,funcinfo.params[i].vtype,Script.GetValue(GetComParam(params,i+1),funcinfo.params[i].vtype));
 if funcinfo.returnvtype <> vtNull then
  AddLocalVariable('return',funcinfo.returnvtype,'');
end;

destructor TFunction.Destroy;
 var i: integer;
begin
 for i:= 0 to locvarcount-1 do
  locvariable[i].LocDestroy;
 inherited Destroy;
end;

//***********************
// Script Variables
//***********************
constructor TMGScript.Create(spath,rpoint,rparams: string);
begin
 inherited Create;
 ScriptLogForm.OpenScript(spath);
 ScriptList:= TStringList.Create;
 path:= spath;
 if FileExists(path) then
  ScriptList.LoadFromFile(path);
 runpoint:= rpoint;
 runparams:= rparams;
 GetRunPoints;
 CalledFunction:= nil;
end;

constructor TMGScript.Create(line: string);
begin
 inherited Create;
 ScriptList:= TStringList.Create;
 ScriptList.Add('InitEnd');
 ScriptList.Add(line);
 GetRunPoints;
 CalledFunction:= nil;
end;

//Summ Script String
procedure TMGScript.GetRunPoints;
 var i: integer;
     buf: string;
begin
 _RPcount:= 0;
 for i:= 0 to ScriptList.Count-1 do
 begin
  buf:= TrimLeft(scriptlist[i]);
  if Length(buf) > 0 then
   if buf[1] = ':' then
   begin
    inc(_RPcount);
    SetLength(RunPoints,_RPcount);
    RunPoints[_RPcount-1].name:= Copy(buf,2,Length(buf));
    RunPoints[_RPcount-1].pos:= i;
   end;
 end;
end;

function TMGScript.FindRunPoint(name: string): integer;
 var i: integer;
begin
 Result:= -1;
 for i:= 0 to _RPCount-1 do
  if RunPoints[i].name = name then
  begin
   Result:= i;
   exit;
  end;
end;

function GetNextBracket(line: string; start: integer): integer;
 var i: integer;
     bcount: integer;
     isstr: boolean;
begin
 i:= -1;
 isstr:= false;
 bcount:= 0;
 if start = 0 then start:= 1;
 for i:= start to Length(line) do
 begin
  if line[i] = '"' then isstr:= not isstr;
  if not isstr then
  begin
   if line[i] = _breketO then inc(bcount);
   if line[i] = _breketC then dec(bcount);
   if bcount < 0 then
   begin
    Result:= i;
    exit;
   end;
  end;
 end;
{ i:= -1;
 for i:= Length(line) downto 1 do
  if line[i] = _breketC then Result:= i;}
end;

function GetLeftVar(line: string; pos: integer): string;
 var i: integer;
     scount: integer;
begin
 Result:= '';
 scount:= 0;
 for i:= pos-1 downto 1 do
 begin
  if line[i] = '(' then dec(scount);
  if line[i] = ')' then inc(scount);
  if scount < 0 then exit;
  if (line[i] = #10) and (scount = 0) then exit;
  //if (line[i] in _operators) and (scount = 0) then exit;
  Result:= line[i] + Result;
 end;
end;

function GetRigthVar(line: string; pos: integer): string;
 var i: integer;
     scount: integer;
begin
 Result:= '';
 scount:= 0;
 for i:= pos+2 to Length(line) do
 begin
  if line[i] = '(' then inc(scount);
  if line[i] = ')' then dec(scount);
  if scount < 0 then exit;
  if (line[i] in _operators) and (scount = 0) then exit;
  Result:= Result + line[i];
 end;
end;

function GetRigthVarBool(line: string; pos: integer): string;
 var i: integer;
     scount: integer;
begin
 Result:= '';
 scount:= 0;
 for i:= pos+2 to Length(line) do
 begin
  if line[i] = '(' then inc(scount);
  if line[i] = ')' then dec(scount);
  if scount < 0 then exit;
  if (line[i] in _booloperators) and (scount = 0) then exit;
  Result:= Result + line[i];
 end;
end;

procedure Put(var line: string; value: string; pos: integer);
begin
 insert(value,line,pos);
end;

function TMGScript.BoolMath(line: string): string;
 var i,j: integer;
     b: boolean;
     par: array [0..1] of string;
     lv,rv: string;
begin
 Result:= 'false';
 if (line = 'true') or (line = 'false') then
 begin
  Result:= line;
  exit;
 end;
 i:= Pos(_breketO,line);
 if i > 0 then
 repeat
  par[0]:= Copy(line,i+1,GetNextBracket(line,i+1)-1-i);
  j:= Length(par[0])+1;
  //par[0]:= Copy(par[0],2,j-1);
  Replace(line,i,j,BoolMath(par[0]));
  i:= Pos(_breketO,line);
 until i = 0;
 //Prepare
 j:= 0;
 line:= StringReplace(line,'not',#1,[rfReplaceAll]);
 line:= StringReplace(line,'and',#2,[rfReplaceAll]);
 line:= StringReplace(line,'or',#3,[rfReplaceAll]);
 line:= StringReplace(line,'xor',#4,[rfReplaceAll]);
 line:= StringReplace(line,'!=',#5,[rfReplaceAll]);
 line:= StringReplace(line,'>=',#6,[rfReplaceAll]);
 line:= StringReplace(line,'<=',#7,[rfReplaceAll]);

 par[0]:= line;
 for i:= 1 to Length(line) do
 begin
  if line[i] in _booloperators then
  begin
   Put(par[0],#10,i+1+j);
   inc(j);
  end;
  if (line[i] = _breketO) or (line[i] = _breketC) then
   if i+1 < Length(line) then
    if not ((line[i+1] = _breketO) or (line[i+1] = _breketC)) then
    begin
     Put(par[0],#10,i+2+j);
     inc(j);
    end;
 end;
 line:= par[0];
 //NOT
 i:= Pos(#1,line);
 if i > 0 then
  repeat
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtBool);
   j:= Length(rv);
   b:= not StrToBool(par[0]);
   Replace(line,i,j+1,BoolToStr(b));
   i:= Pos(#1,line);
  until i = 0;
 //AND
 i:= Pos(#2,line);
 if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtBool);
   par[1]:= GetValue(lv,vtBool);
   j:= Length(rv)+Length(lv)+1;
   b:= StrToBool(par[1]) and StrToBool(par[0]);
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos(#2,line);
  until i = 0;
 //OR
 i:= Pos(#3,line);
 if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtReal);
   par[1]:= GetValue(lv,vtReal);
   j:= Length(rv)+Length(lv)+1;
   b:= StrToBool(par[1]) or StrToBool(par[0]);
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos(#3,line);
  until i = 0;
 //XOR
  i:= Pos(#4,line);
  if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtReal);
   par[1]:= GetValue(lv,vtReal);
   j:= Length(rv)+Length(lv)+1;
   b:= StrToBool(par[1]) xor StrToBool(par[0]);
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos(#4,line);
  until i = 0;
 // =
 i:= Pos('=',line);
 if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtString);
   par[1]:= GetValue(lv,vtString);
   j:= Length(rv)+Length(lv)+1;
   b:= par[1] = par[0];
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos('=',line);
  until i = 0;
 // <
  i:= Pos('<',line);
 if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtReal);
   par[1]:= GetValue(lv,vtReal);
   j:= Length(rv)+Length(lv)+1;
   b:= StrToReal(par[1]) < StrToReal(par[0]);
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos('<',line);
  until i = 0;
 // >
 i:= Pos('>',line);
 if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtReal);
   par[1]:= GetValue(lv,vtReal);
   j:= Length(rv)+Length(lv)+1;
   b:= StrToReal(par[1]) > StrToReal(par[0]);
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos('>',line);
  until i = 0;
 // <=
  i:= Pos(#7,line);
 if i > 0 then
  repeat
   lv:= GetLeftVar(line,i);
   rv:= GetRigthVarBool(line,i);
   par[0]:= GetValue(rv,vtReal);
   par[1]:= GetValue(lv,vtReal);
   j:= Length(rv)+Length(lv)+1;
   b:= StrToReal(par[1]) <= StrToReal(par[0]);
   Replace(line,i-Length(lv),j,BoolToStr(b));
   i:= Pos(#7,line);
  until i = 0;
 // >=
  i:= Pos(#6,line);
 if i > 0 then
  repeat
   par[0]:= GetValue(GetRigthVarBool(line,i),vtReal);
   par[1]:= GetValue(GetLeftVar(line,i),vtReal);
   j:= Length(par[0])+Length(par[1])+1;
   b:= StrToReal(par[1]) >= StrToReal(par[0]);
   Replace(line,i-Length(par[1]),j,BoolToStr(b));
   i:= Pos(#6,line);
  until i = 0;


 Result:= line;
end;

function TMGScript.StrMath(str: string): string;
 var i: integer;
     count: integer;
     s: array [0..1000] of string;
begin
 count:= 0;
 i:= Pos('+',str);
 while true do
 begin
  if i = 0 then break;
  s[count]:= Copy(str,1,i-1);
  Delete(str,1,i);
  inc(count);
  i:= Pos('+',str);
 end;
 s[count]:= str;
 Result:= '';
 for i:= 0 to count do
  Result:= Result + GetValue(s[i],vtString);
end;

function TMGScript.Math(line: string): string;
 var i,j,scount: integer;
     m: real;
     par: array [0..1] of string;
     isstr: boolean;
begin
 Result:= '0';
 if Length(line) = 0 then exit;
 //Brackets
 i:= Pos(_breketO,line);
 if i > 0 then
 repeat
  par[0]:= Copy(line,i+1,GetNextBracket(line,i+1)-1-i);
  j:= Length(par[0])+1;
  Replace(line,i,j,Math(par[0]));
  i:= Pos(_breketO,line);
 until i = 0;
 //Prepare
 j:= 0;
 scount:= 0;
 par[0]:= line;
 for i:= 1 to Length(line) do
 begin
  if line[i] = '"' then isstr:= not isstr;
  if (line[i] = '(') and (not isstr) then inc(scount);
  if (line[i] = ')') and (not isstr) then dec(scount);
  if (not isstr) and ((line[i] in _operators) and (scount = 0)) then
  begin
   Put(par[0],#10,i+1+j);
   inc(j);
  end;
  {if (not isstr) and ((line[i] = _breketO) or (line[i] = _breketC)) then
   if i+1 < Length(line) then
    if not ((line[i+1] = _breketO) or (line[i+1] = _breketC)) then
     if scount = 0 then
     begin
      Put(par[0],#10,i+2+j);
      inc(j);
    end;}
 end;
 line:= par[0];
 // *
 i:= Pos('*',line);
 if i > 0 then
 repeat
  par[0]:= GetLeftVar(line,i);
  par[1]:= GetRigthVar(line,i);
  j:= Length(par[0])+Length(par[1])+1;
  m:= StrToReal(GetValue(par[0],vtReal))*StrToReal(GetValue(par[1],vtReal));
  Replace(line,i-Length(par[0]),j,RealToStr(m));
  i:= Pos('*',line);
 until i = 0;
 // /
 i:= Pos('/',line);
 if i > 0 then
 repeat
  par[0]:= GetLeftVar(line,i);
  par[1]:= GetRigthVar(line,i);
  if par[1] = '0' then exit;
  j:= Length(par[0])+Length(par[1])+1;
  m:= StrToReal(GetValue(par[0],vtReal))/StrToReal(GetValue(par[1],vtReal));
  Replace(line,i-Length(par[0]),j,RealToStr(m));
  i:= Pos('/',line);
 until i = 0;
 // +
 i:= Pos('+',line);
 if i > 0 then
 repeat
  par[0]:= GetLeftVar(line,i);
  par[1]:= GetRigthVar(line,i);
  j:= Length(par[0])+Length(par[1])+1;
  m:= StrToReal(GetValue(par[0],vtReal))+StrToReal(GetValue(par[1],vtReal));
  Replace(line,i-Length(par[0]),j,RealToStr(m));
  i:= Pos('+',line);
 until i = 0;
 // -
 i:= Pos('-',line);
 if i > 0 then
 repeat
  par[0]:= GetLeftVar(line,i);
  par[1]:= GetRigthVar(line,i);
  j:= Length(par[0])+Length(par[1])+1;
  if i = 1 then
   m:= -1*StrToReal(GetValue(par[1],vtReal))
  else
   m:= StrToReal(GetValue(par[0],vtReal))-StrToReal(GetValue(par[1],vtReal));
  Replace(line,i-Length(par[0]),j,RealToStr(m));
  i:= Pos('-',line);
 until i < 2;
 Result:= line;
end;

function IsHaveMath(line: string; vt: TVarType): boolean;
 var i: integer;
     isstr: boolean;
     funccount: integer;
     str: string;
begin
 Result:= false;
 isstr:= false;
 funccount:= 0;
 for i:= 1 to Length(line) do
 begin
  if line[i] = '"' then isstr:= not isstr;
  if not isstr then
  begin
   if line[i] = '(' then inc(funccount);
   if line[i] = ')' then dec(funccount);
  end;
  if (not isstr) and (funccount = 0) then
  begin
   if vt in [vtInteger,vtReal,vtString] then
   begin
    if line[i] = '*' then Result:= true;
    if line[i] = '/' then Result:= true;
    if line[i] = '-' then Result:= true;
   end;
   if vt in [vtInteger,vtReal,vtString] then
    if line[i] = '+' then Result:= true;
   if vt in [vtBool] then
   begin
    if line[i] = '>' then Result:= true;
    if line[i] = '=' then Result:= true;
    if line[i] = '<' then Result:= true;
    str:= Copy(line,i,3);
    if str <> '' then
    begin
     if str = 'and' then Result:= true;
     if str = 'not' then Result:= true;
     if str = 'xor' then Result:= true;
    end;
    str:= Copy(line,1,2);
    if str <> '' then
    begin
     if str = 'or' then Result:= true;
     if str = '<=' then Result:= true;
     if str = '>=' then Result:= true;
    end;
   end;
  end;
  if Result then exit;
 end;
 //Result:= (Pos('*',line)+Pos('-',line)+Pos('/',line)+Pos('+',line)+Pos('<=',line)+Pos('>=',line)+Pos('=',line)+Pos('not',line)+Pos('and',line)+Pos('or',line)+Pos('>',line)+Pos('<',line)) > 0;
end;

function TMGScript.FindLocalVariable(name: string): integer;
 var i: integer;
begin
 for i:= 0 to Fvarcount-1 do
  if Fvars[i].name = name then
  begin
   Result:= i;
   exit;
  end;
 result:= -1;
end;

procedure TMGScript.AddLocalVariable(name: string; t: TVarType; value: string);
 var i: integer;
begin
 i:= FindLocalVariable(name);
 if i >= 0 then exit;
 i:= Fvarcount;
 inc(Fvarcount);
 SetLength(Fvars,Fvarcount);
 Fvars[i]:= TVar.LocNew(name,t,value);
end;

//Get Value From Script String
function TMGScript.GetValue(parm: string; vt: TVarType):string;
 var i: integer;
     comname,params: string;
     par: array [1..10] of string;
     int: array [1..10] of integer;
     localObj: TObject;
     b: boolean;
 label fail;
begin
 Result:= parm;

 if IsVariable(Result) and (Pos('.',Result) = 0) and (not IsHaveMath(Result,vt)) then
 begin
  par[1]:= Copy(parm,2,Length(parm));
  //if Array Value
  b:= false;
  int[1]:= Length(par[1]);
  if par[1][int[1]] = ']' then
   begin
    b:= true;
    int[2]:= Pos('[',par[1]);
    par[3]:= par[1];
    par[1]:= Copy(par[1],1,int[2]-1);
    par[9]:= Copy(par[3],int[2]+1,Length(par[3])-int[2]-1);
    int[1]:= StrToInt(GetValue(par[9],vtInteger));
   end;
  //Function local find
  if CalledFunction <> nil then
  begin
   i:= CalledFunction.FindLocalVariable(par[1]);
   if i >= 0 then
   begin
    if b then
     Result:= CalledFunction.locvariable[i].GetArrayValue(int[1])
    else
     Result:= CalledFunction.locvariable[i].value;
    exit;
   end;
  end;
  //Locals Find
  i:= FindLocalVariable(par[1]);
  if i >= 0 then
  begin
   if b then
    Result:= FVars[i].GetArrayValue(int[1])
   else
    Result:= Fvars[i].value;
   exit;
  end;
  //Globals Find
  i:= FindVarId(par[1]);
  if i >= 0 then
  begin
   if b then
    Result:= _vars[i].GetArrayValue(int[1])
   else
    Result:= _vars[i].value;
   exit;
  end;
 end;

 //*****************************************************************************
 if IsHaveMath(Result,vt) then
 begin
  // BOOL MATH
  if vt = vtBool then
  begin
   Result:= BoolMath(Result);
   if not ((Result = 'true') or (Result = 'false')) then
    Result:= 'false';
   exit;
  end;
  //MATH
  if (vt = vtInteger) or (vt = vtReal) then
  begin
   Result:= Math(Result);
   if vt = vtInteger then
    Result:= IntToStr(Round(StrToReal(Result)));
   exit;
  end;
  //STRING MATH
  if vt = vtString then
  begin
   if Pos('+',parm) > 0 then
    Result:= StrMath(parm)
   else
    Result:= parm;
   exit;
  end;
 end;
 Result:= parm;
 //*****************************************************************************
 comname:= GetComName(parm);
 params:= GetComParams(parm);
 //=======================
 // Self: [str]
 //=======================
 if comname = 'Self' then
 begin
  Result:= path;
  exit;
 end;
 //=======================
 // GetScriptParam([int]): [str]
 //=======================
 if comname = 'GetScriptParam' then
 begin
  par[1]:= GetValue(GetComParam(params,1),vtInteger);
  Result:= '';
  if not CheckInt(par[1]) then exit;
  Result:= GetComParam(runparams,StrToInt(par[1]));
  exit;
 end;
 //=======================
 // Round([rea]): [int]
 //=======================
 if comname = 'Round' then
 begin
  Result:= '';
  par[1]:= GetValue(GetComParam(params,1),vtInteger);
  if not CheckReal(par[1]) then exit;
  Result:= IntToStr(Round(StrToReal(par[1])));
  exit;
 end;
 //=======================
 // SqRoot([rea]): [rea]
 //=======================
 if comname = 'SqRoot' then
 begin
  par[1]:= GetValue(GetComParam(params,1),vtReal);
  if not CheckReal(par[1]) then exit;
  Result:= RealToStr(SqRt(StrToReal(par[1])));
  exit;
 end;
 //========================
 // ExtToStr([rea]): [str]
 //========================
 if comname = 'ExtToStr' then
 begin
  par[1]:= GetValue(GetComParam(params,1),vtReal);
  if not CheckReal(par[1]) then exit;
  Result:= par[1];
  exit;
 end;
 //========================
 // BoolToStr([bol]): [str]
 //========================
 if comname = 'BoolToStr' then
 begin
  par[1]:= GetValue(GetComParam(params,1),vtBool);
  Result:= par[1]; //BoolMath(par[1]);
  exit;
 end;
 //========================
 // Random(start[int],len[int]): [int]
 //========================
 if comname = 'Random' then
 begin
  par[1]:= GetValue(GetComParam(params,1),vtInteger); //Start
  par[2]:= GetValue(GetComParam(params,2),vtInteger); //Len
  if not CheckInt(par[1]) then exit;
  if not CheckInt(par[2]) then exit;
  i:= StrToInt(par[1])+Random(StrToInt(par[2]));
  Result:= IntToStr(i);
  exit;
 end;  
 {Char Data Base Functions}
 //========================
 // GetCharID_Name([str]): [int]
 // Find Char by NAME
 //========================
 if GetParentObj(comname,2) = 'GetCharID_Name' then
 begin
  par[1]:= GetValue(GetComParam(params,1),vtString);
  i:= ChDataBase.FindCharByName(par[1]);
  if not CheckRange(i,0,ChDataBase.charcount-1,'char array') then exit;
  Result:= IntToStr(ChDataBase.chars[i].charid);
  exit;
 end;
 //************************
 // -- DIALOG --
 //************************
 if GetChildObj(comname,1) = 'Dialog' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //=================================
  // Create: [int]
  //=================================
  if comname = 'Create' then
  begin
   localObj:= TDialogTable.Create;
   Result:= IntToStr((localObj as TDialogTable).HDialog);
   exit;
  end;
  par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HDIALOG
  if not CheckInt(par[1]) then Exit;
  int[1]:=  StrToInt(par[1]);
  if not CheckRange(int[1],0,DialogCount-1,'dialog array') then Exit;
  //=================================
  // HDIALOG[int].Owner: [int]
  //=================================
  if comname = 'Owner' then
  begin
    Result := IntToStr(DialogList[int[1]].char.HUnit);
    Exit;
  end;
  comname := par[10];
 end;
 //************************
 // -- GROUP --
 //************************
 if GetChildObj(comname,1) = 'Group' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //================================
  // Create: [int]
  //================================
  if comname = 'Create' then
  begin
   localObj:= TGroup.Create;
   Result:= IntToStr((localObj as TGroup).HGroup);
   exit;
  end;
  par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HGROUP
  if not CheckInt(par[1]) then exit;
  int[1]:= StrToInt(par[1]);
  if not CheckRange(int[1],0,GroupCount-1,'group array') then exit;
  //================================
  // @HGROUP.Char(id [int]): [int]
  //================================
  if comname = 'Char' then
  begin
   par[2]:= GetValue(GetComParam(params,1),vtInteger);
   if not CheckInt(par[2]) then exit;
   int[2]:= StrToInt(par[2]);
   if not CheckRange(int[2],0,4,'group char list') then exit;
   Result:= IntToStr(GroupList[int[1]].FUnits[int[2]]);
  end;
  //================================
  // @HGROUP.CharCount: [int]
  //================================
  if comname = 'CharCount' then
  begin
   Result:= IntToStr(GroupList[int[1]].GetCount);
   exit;
  end;
  //################################
  comname:= par[10];
 end;
 //************************
 // -- CHAR BLOCK --
 //************************
 if GetChildObj(comname,1) = 'Char' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //========================
  // @CharID[int].HUNIT: [int]
  // Get char hunit
  //========================
  if comname = 'HUNIT' then
  begin
   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger);
   Result:= '';
   if not CheckInt(par[1]) then exit;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,ChDataBase.charcount,'char data base array') then exit;
   Result:= IntToStr(ChDataBase.chars[int[1]].hunit);
   exit;
  end;

   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HUNIT
   Result:= '';
   if not CheckInt(par[1]) then exit;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,UnitCount,'unit array array') then exit;
   if not CheckChar(int[1]) then exit;
  //========================
  // @HUNIT[int].Gold: [int]
  //========================
  if comname = 'Gold' then
  begin
   Result:= IntToStr( (UnitList[int[1]] as TCharA).gold );
   exit;
  end;
  //========================
  // @HUNIT[int].Honor: [int]
  //========================
  if comname = 'Honor' then
  begin
   Result:= IntToStr((UnitList[int[1]] as TCharA).Honor );
   exit;
  end;
  //========================
  // @HUNIT[int].PvPKill: [int]
  //========================
  if comname = 'PvPKill' then
  begin
    Result := IntToStr((UnitList[int[1]] as TCharA).killCount);
    Exit;
  end;
  //========================
  // @HUNIT[int].SkillPoint: [int]
  // Get hero skill points
  //========================
  if comname = 'SkillPoint' then
  begin
   Result:= IntToStr((UnitList[int[1]] as TCharA).skillpoints );
   exit;
  end;
  //========================
  // @HUNIT[int].ClassID: [int]
  // Get unit class id
  //========================
  if comname = 'ClassID' then
  begin
   Result:= IntToStr((UnitList[int[1]] as TCharA).classid );
   exit;
  end;
  //#######################
  exit;
  //comname:= par[10];
 end;
 //*************************
 // -- END CHAR BLOCK --
 //*************************

 //*************************
 // -- ITEM BLOCK --
 //*************************
 if GetChildObj(comname,1) = 'Item' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //=======================
  // Create(@locId, @x,@y,@rawcode): [HITEM]
  // Create Item in game filed
  //=======================
  if comname = 'Create' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtInteger);
   par[2]:= GetValue(GetComParam(params,2),vtInteger);
   par[3]:= GetValue(GetComParam(params,3),vtInteger);
   par[4]:= GetValue(GetComParam(params,4),vtString);
   for i:= 1 to 3 do
    if not CheckInt(par[i]) then exit;
   int[1] := StrToInt(par[1]);
   localObj:= TItm.Create(par[4],int[1],StrToInt(par[2]),StrToInt(par[3]));
   Result:= IntToStr((localObj as TItm).HItem);
   SendAllPlayersOnLoc('newitm('+par[2]+','+par[3]+','+par[4]+','+Result+')',int[1],nil);
   exit;
  end;
  //#######################
  exit;
  //comname:= par[10];
 end;
 //*************************
 // -- END ITEM BLOCK --
 //*************************

 //*************************
 // -- UNIT DATA BASE BLOCK --
 //*************************
 if GetChildObj(comname,1) = 'UnitDataBase' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //=======================
  // RandomGold(@rawcode[str]): [int]
  //=======================
  if comname = 'RandomGold' then
  begin
   Result:= '0';
   par[1]:= GetValue(GetComParam(params,1),vtInteger); //rawcode;
   int[1]:= UnDataBase.UnitsRaw[par[1]].mingold;
   int[2]:= UnDataBase.UnitsRaw[par[1]].maxgold-int[1];
   Result:= IntToStr(int[1]+Random(int[2]));
   exit;
  end;
  //=======================
  // GiveExp(@rawcode [str]): [int]
  //=======================
  if comname = 'GiveExp' then
  begin
   Result:= '0';
   par[1]:= GetValue(GetComParam(params,1),vtString); //rawcode
   Result:= IntToStr( UnDataBase[par[1]].giveexp );
   exit;
  end;
  //#######################
  comname:= par[10];
 end;
 //************************
 // -- END UNIT DATA BASE BLOCK --
 //************************

 //************************
 // -- UNIT BLOCK --
 //************************
 if GetChildObj(comname,1) = 'Unit' then
 begin
   par[10]:= comname;
   comname:= GetParentObj(comname,1);
   //=====================
   // Create (@utype,@rawcode,@locid,@x,@y,@yaw,@life,@mana): [int]
   // Create unit (RAWCODE) from UTYPE in game word to X,Y with LIFE and MANA and fasing to YAW
   // Return HUNIT a new unit
   //=====================
   if comname = 'Create' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtString);//utype
    par[2]:= GetValue(GetComParam(params,2),vtString);//rawcode
    par[8]:= GetValue(GetComParam(params,3),vtInteger);//locID
    par[3]:= GetValue(GetComParam(params,4),vtInteger);//x
    par[4]:= GetValue(GetComParam(params,5),vtInteger);//y
    par[5]:= GetValue(GetComParam(params,6),vtInteger);//yaw
    par[6]:= GetComParam(params,7);//life
    if not UnitGetDirective(par[2],par[6]) then
     par[6]:= GetValue(par[6],vtInteger);
    par[7]:= GetComParam(params,8);//mana
    if not UnitGetDirective(par[2],par[7]) then
     par[7]:= GetValue(par[7],vtInteger);
    for i := 3 to 8 do
     if not CheckInt(par[i]) then exit;
    int[2]:= StrToInt(par[8]);
    localObj:= TUnit.Create(par[2],StrToInt(par[8]),StrToInt(par[3]),StrToInt(par[4]),0,StrToInt(par[5]),StrToInt(par[6]),StrToInt(par[7]),StrToUType(par[1]));
    int[1]:= (localObj as TUnit).HUnit;
    if par[1] = 'npc' then i:= 1;
    if par[1] = 'enemy' then i:= 2;
    SendAllPlayersOnLoc('createunit('+IntToStr(int[1])+','+par[2]+','+IntToStr(i)+')',int[2],nil);
    SendAllPlayersOnLoc('newXY('+(localObj as TUnit).GetXY+')',int[2],nil);
    SendAllPlayersOnLoc('newHM('+(localObj as TUnit).GetHM+')',int[2],nil);
    Result:= IntToStr(int[1]);
    exit;
   end;
   // ++
   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HUNIT
   Result:= '';
   if not CheckInt(par[1]) then exit;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,UnitCount-1,'unit array') then exit;
  //========================
  // @HUNIT[int].OrderTarget: [int]
  // return current target unit
  //========================
  if comname = 'OrderTarget' then
  begin
   Result:= IntToStr(UnitList[int[1]].order.target);
   exit;
  end;
  //========================
  // @HUNIT[int].Name: [str]
  // Get char name
  //========================
  if comname = 'Name' then
  begin
   Result:= UnitList[int[1]].name;
   exit;
  end;
  //========================
  // @HUNIT[int].EXP: [int]
  // Get unit exp
  //========================
  if comname = 'EXP' then
  begin
   Result:= IntToStr(UnitList[int[1]].exp);
   exit;
  end;
  //========================
  // @HUNIT[int].RawCode: [str]
  // Get unit raw code
  //========================
  if comname = 'RawCode' then
  begin
   Result:= UnitList[int[1]].rawCode;
   exit;
  end;
  //========================
  // @HUNIT[int].Agro: [int]
  // Get unit agro
  //========================
  if comname = 'Agro' then
  begin
    Result:= IntToStr(UnitList[int[1]].agro);
   exit;
  end;
  //========================
  // @HUNIT[int].LocationID: [int]
  // Get unit loc ID
  //========================
  if comname = 'LocationID' then
  begin
    Result := IntToStr(UnitList[int[1]].locid);
    Exit;
  end;
  //========================
  // @HUNIT[int].IsHaveAbility(@raw): [bol]
  // Is Unit Have ability with raw code
  //========================
  if comname = 'IsHaveAbility' then
  begin
   par[2]:= GetValue(GetComParam(params,1),vtString);
   Result:= BoolToStr(UnitList[int[1]].FindAbility(par[2]) > 0);
   exit;
  end;
  //========================
  // @HUNIT[int].GetAbilityLevel(@raw[int]): [int]
  // Return ability level from unit
  //=========================
  if comname = 'GetAbilityLevel' then
  begin
   par[2]:= GetValue(GetComParam(params,1),vtString);
   int[2]:= UnitList[int[1]].FindAbility(par[2]);
   if not CheckRange(int[2],0,10,'unit spell list') then exit;
   Result:= IntToStr(UnitList[int[1]].spells[int[2]].lvl);
   exit;
  end;
  //========================
  // @HUNIT[int].ReadValue(field[str],defvalue[str]): [str]
  // Get unit data table value
  //========================
  if comname = 'ReadValue' then
  begin
   par[3]:= GetValue(GetComParam(params,2),vtString); //default value
   Result:= '';
   par[2]:= GetValue(GetComParam(params,1),vtString); //field
   int[2] := UnitList[int[1]].DataTable.IndexOfName(par[2]);
   if int[2] >= 0 then
   begin
     Result := UnitList[int[1]].DataTable.Values[par[2]];
   end else begin
     int[2] := UnitList[int[1]].DataTableLoc.IndexOfName(par[2]);
     if int[2] >= 0 then
       Result := UnitList[int[1]].DataTableLoc.Values[par[2]]
     else
       Result := par[3];
   end;
   if Result = '' then
    Result:= par[3];
   Exit;
  end;
  //========================
  // @HUNIT[int].IsChar: [bol]
  //========================
  if comname = 'IsChar' then
  begin
   Result := BoolToStr(UnitList[int[1]]._type = utPlayer);
   Exit;
  end;
  //========================
  // @HUNIT[int].Health: [int]
  // Get unit health point
  //========================
  if comname = 'Health' then
  begin
   Result:= '-1';
   Result:= IntToStr(UnitList[int[1]].health);
   exit;
  end;
 //=========================
 // @HUNIT[int].Mana: [int]
 // Get unit mana point
 //=========================
  if comname = 'Mana' then
  begin
   Result:= '-1';
   Result:= IntToStr(UnitList[int[1]].mana);
   exit;
  end;
 //=========================
 // @HUNIT[int].IsHasItem(@rawcode [str]): [bol]
 //=========================
 if comname = 'IsHasItem' then
 begin
  par[2] := GetValue(GetComParam(params,1),vtString);
  for i := 0 to 5 do
  begin
   int[2]:= UnitList[int[1]].items[i];
   int[2]:= int[2];
   if int[2] >= 0 then
   begin
    Result:= BoolToStr( ItemList[int[2]].RawCode = par[2] );
    if Result = 'true' then Exit;
   end;
  end;
  Result := 'false';
  Exit;
 end;
 //=========================
 // @HUNIT[int].GetItemByRC(@rawcode [str]): [int]
 //=========================
 if comname = 'GetItemByRC' then
 begin
  Result:= '-1';
  par[2] := GetValue(GetComParam(params,1),vtString);
  for i:= 0 to 5 do
  begin
   int[2]:= UnitList[int[1]].items[i];
   if int[2] >= 0 then
    if ItemList[int[2]].RawCode = par[2] then
    begin
     Result := IntToStr(UnitList[int[1]].items[i]);
     Exit;
    end;
  end;
  Exit;
 end;
 //=========================
 // @HUNIT[int].Level: [int]
 // Get Unit Level
 //=========================
 if comname = 'Level' then
 begin
  Result:= IntToStr( ExpToLevel( UnitList[int[1]].exp ) );
  Exit; 
 end;
 //=========================
 // @HUNIT[int].X: [int]
 // Get unit x coordinate
 //=========================
  if comname = 'X' then
  begin
   Result:= IntToStr(UnitList[int[1]].posx);
   exit;
  end;
 //=========================
 // @HUNIT[int].Y: [int]
 // Get unit y coordinate
 //=========================
  if comname = 'Y' then
  begin
   Result:= IntToStr(UnitList[int[1]].posy);
   exit;
  end;
 //=========================
 // @HUNIT[int].Rotate: [int]
 // Get unit faceing rotation
 //=========================
  if comname = 'Rotate' then
  begin
   Result:= IntToStr(UnitList[int[1]].yaw);
   exit;
  end;
 //############################
  exit;
  //comname:= par[10];
 end;
 //****************************
 // -- END UNIT BLOCK --
 //****************************

 //****************************
 // -- UNIT GROUP BLOCK --
 //****************************
 if GetChildObj(comname,1) = 'UnitGroup' then
 begin
  par[10] := comname;
  comname := GetParentObj(comname,1);
  //============================
  // Create: HUGROUP
  //============================
  if comname = 'Create' then
  begin
    localObj := TUnitGroup.Create;
    Result := IntToStr((localObj as TUnitGroup).HUGroup);
    Exit;
  end;
  par[1] := GetValue(GetParentObj(par[10],2),vtInteger); //HUGroup
  if not CheckInt(par[1]) then Exit;
  int[1] := StrToInt(par[1]);
  if not Checkrange(int[1], 0, UGroupCount-1, ' unit group') then Exit;
  //==============================
  // GetUnit (@num [int]): HUNIT [int]
  //==============================
  if comname = 'GetUnit' then
  begin
    Result := '-1';
    par[2] := GetValue(GetComParam(params,1),vtInteger);
    if not CheckInt(par[2]) then Exit;
    int[2] := StrToInt(par[2]);
    if not CheckRange(int[2],0,UGroupList[int[1]].count-1,' unit group unit list') then Exit;
    Result := IntToStr(UGroupList[int[1]].Units[int[2]].HUnit);
    Exit;
  end;
  //==============================
  // UnitCount: [int]
  //==============================
  if comname = 'UnitCount' then
  begin
    Result := IntToStr(UGroupList[int[1]].count);
    Exit;
  end;
  //==============================
  // GetRandomUnit: HUNIT [int]
  //==============================
  if comname = 'GetRandomUnit' then
  begin
    localObj := UGroupList[int[1]].GetRandomUnit;
    Result := IntToStr((localObj as TUnit).HUnit);
    Exit;
  end;
  //################
  Exit;
 end;
 //****************************
 // -- END UNIT GROUP BLOCK --
 //****************************

 //****************************
 // -- EFFECT BLOCK --
 //****************************
 if GetChildObj(comname,1) = 'Effect' then
 begin
  par[10] := comname;
  comname := GetParentObj(comname,1);
  //=================================
  //  Create
  //=================================
  if comname = 'Create' then
  begin
    localObj := TEffect.Create;
    Result := IntToStr((localobj as TEffect).HEffect);
    Exit;
  end;
  //#################
  Exit;
 end;
 //****************************
 // -- END EFFECT BLOCK --
 //****************************

 //****************************
 // -- RECT BLOCK
 //****************************
 if GetChildObj(comname,1) = 'Rect' then
 begin
  par[10] := comname;
  comname := GetParentObj(comname,1);
  //=============================
  // Create(locid [int],x [int],y [int],w [int],h [int])
  // Create new Rect in LocID with coordinate X,Y and container W,H
  //=============================
  if comname = 'Create' then
  begin
    par[1] := GetValue(GetComParam(params,1),vtInteger); //LocId
    par[2] := GetValue(GetComParam(params,2),vtInteger); //X
    par[3] := GetValue(GetComParam(params,3),vtInteger); //Y
    par[4] := GetValue(GetComParam(params,4),vtInteger); //W
    par[5] := GetValue(GetComParam(params,5),vtInteger); //H
    for i := 1 to 5 do
      if not CheckInt(par[i]) then Exit else int[i] := StrToInt(par[i]);
    localObj := TRect.Create(int[1], int[2],int[3], int[4],int[5]);
    Result := IntToStr((localObj as TRect).HRect);
    Exit;
  end;
  //#################
  Exit;
 end;
 //****************************
 // -- END RECT BLOCK
 //****************************

 //****************************
 // -- QUEST BLOCK --
 //****************************
 if GetChildObj(comname,1) = 'Quest' then
 begin
   par[10] := comname;
   comname := GetParentObj(comname,1);
   //=================================
   // Find(status [str]): [int]
   //=================================
   if comname = 'Find' then
   begin
     par[1] := GetValue(GetComParam(params,1),vtString);
     Result := IntToStr(QuestDataBase.Find(par[1]));
     Exit;
   end;
   //====
   Exit;
 end;

 //****************************
 // -- DEST BLOCK --
 //****************************
 if GetChildObj(comname,1) = 'Dest' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //======================
  // Create(@rawcode,@locid,@x[int],@y[int],@z[int],@yaw[int]): HDest[int]
  // create a Dest (@rawcode) in @X,@Y,@Z coordinate with @yaw
  //======================
   if comname = 'Create' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtString); //RawCode
    par[2]:= GetValue(GetComParam(params,2),vtInteger);//locId
    par[3]:= GetValue(GetComParam(params,3),vtInteger);//X
    par[4]:= GetValue(GetComParam(params,4),vtInteger);//Y
    par[5]:= GetValue(GetComParam(params,5),vtInteger);//Z
    par[6]:= GetValue(GetComParam(params,6),vtInteger);//Yaw
    for i:= 2 to 6 do
     if not CheckInt(par[i]) then Exit;
    int[1] := StrToInt(par[2]);
    localObj:= TDest.Create(par[1],int[1],StrToInt(par[3]),StrToInt(par[4]),StrToInt(par[5]),StrToInt(par[6]));
    Main.SendAllPlayersOnLoc('destcreate('+(localObj As TDest).GetMainInfo+')',int[1],nil);
    Result:= IntToStr((localObj as TDest).HDest);
    localObj:= nil;
    exit;
   end;
 //############################
  exit;
  //comname:= par[10];
 end;
 //****************************
 // -- END DEST BLOCK
 //****************************

 fail:
end;

destructor TMGScript.Destroy;
 var i: integer;
begin
 inherited Destroy;
 ScriptLogForm.CloseScript;
 ScriptList.Destroy;
 for i:= 0 to Fvarcount-1 do
  Fvars[i].LocDestroy;
end;

procedure TMGScript.AddNewFunction(line: string; spos: integer);
 var i,id: integer;
     param: string;
     par: string;
begin
 id:= funccount;
 inc(funccount);
 SetLength(functionlist,funccount);
 par:= GetComName(line);
 Delete(par,1,8);
 functionlist[id].name:= par;
 functionlist[id].pos:= spos;
 i:= Pos(':',line);
 if i > 0 then
  functionlist[id].returnvtype:= ConvertVarType(TrimLeft(Copy(line,i+1,5)))
 else
  functionlist[id].returnvtype:= vtNull;
 param:= GetComParams(line);
 //Get Params
 par:= GetComParam(param,1);
 with functionlist[id] do
  while par <> '' do
  begin
   SetLength(params,parcount+1);
   params[parcount].vtype:= ConvertVarType(Copy(par,1,5));
   Delete(par,1,5);
   params[parcount].name:= par;
   inc(parcount);
   par:= GetComParam(param,parcount+1);
  end;
end;

function TMGScript.FindFunc(name: string): integer;
 var i: integer;
begin
 for i:= 0 to funccount-1 do
  if functionlist[i].name = name then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

procedure TMGScript.Compile;
 var i,j,z: integer;
     //loop variables
     loopcount: integer;
     looppos,loopcircles: array [1..8192] of integer;
     //if varialbes
     ifcount,ifpos: integer;
     ifvalue: array [1..8192] of boolean;
     //other
     comname,blockname,params,buf: string;
     par: array [1..10] of string;
     int: array [1..10] of integer;
     bol: array [1..10] of boolean;
     localObj: TObject;
     endinit: boolean;
 label nextline;
begin
 i:= 0;
 loopcount:= 0;
 ifcount:= 0;
 ifpos:= 0;
 endinit:= false;
 while i <= scriptlist.Count-1 do
 begin
  ScriptLogForm.lastline:= i;
  buf:= scriptlist[i];
  //Delete comment
  if Pos('//',buf) > 0 then
   Delete(buf,Pos('//',buf),Length(buf));
  buf := TrimLeft(buf);
  comname:= GetComName(buf);
  params:= GetComParams(buf);
  //**********************************
  //Init Function List
  if not endinit then
  begin
   //=====================
   // function name(params): returntype
   //=====================
   if Copy(comname,1,8) = 'function' then
   begin
    AddNewFunction(buf,i);
    goto nextline;
   end;
   //=====================
   // Include(@path)
   //=====================
   if comname = 'Include' then
   begin
    par[1]:= GetComParam(params,1);
    if not FileExists(par[1]) then
    begin
     ScriptLogForm.AddLog('file "'+par[1]+'" not found',i);
     goto nextline;
    end;
    localObj:= TStringList.Create;
    scriptlist.Delete(i);
    (localObj as TStrings).LoadFromFile(par[1]);
    for j:= 0 to (localObj as TStrings).Count-1 do
     scriptlist.Insert(i+j,(localObj as TStrings)[j]);
    localObj.Destroy;
    localObj:= nil;
    i:= i - 1;
    goto nextline;
   end;
   //=====================
   // InitEnd
   //=====================
   if comname = 'InitEnd' then
   begin
    endinit:= true;
    if Length(runpoint) > 0 then
    begin
     j:= FindRunPoint(runpoint);
     if j >= 0 then
      i:= RunPoints[j].pos;
     goto nextline;
    end;
   end;
   // ---------
   goto nextline;
  end;
  //******************************
  //Start Standart Functions
  //======================
  // endfunction
  //======================
  if comname = 'endfunction' then
   if CalledFunction <> nil then
   begin
    i:= CalledFunction.calledpos;
    localobj:= nil;
    if CalledFunction.From <> nil then
     localobj:= @CalledFunction.From;
    CalledFunction.Destroy;
    if localobj <> nil then
     CalledFunction:= localobj as TFunction
    else
     CalledFunction:= nil;
    goto nextline;
  end;
  //======================
  // if (@condition) then
  //======================
  if comname = 'if' then
  begin
   inc(ifcount);
   if ifcount-1 > 0 then
   begin
    if ifvalue[ifcount-1] then
    begin
     ifvalue[ifcount]:= GetValue(params,vtBool) = 'true';
     inc(ifpos);
    end else
     ifvalue[ifcount]:= false;
   end else
   begin
    ifvalue[ifcount]:= GetValue(params,vtBool) = 'true';
    inc(ifpos);
   end;
   goto nextline;
  end;
  //=====================
  // endif
  //=====================
  if comname = 'endif' then
  begin
   if ifcount = ifpos then dec(ifpos);
   dec(ifcount);
   goto nextline;
  end;
  //=====================
  // else
  //=====================
  if (ifcount > 0) then
   if (comname = 'else') and (ifpos = ifcount) then
   begin
    ifvalue[ifcount]:= not ifvalue[ifcount];
    goto nextline;
   end;
  if ifcount > 0 then
   if not ifvalue[ifcount] then goto nextline;
  //======================
  // call
  // calling a custom function
  //======================
  if Copy(comname,1,4) = 'call' then
  begin
   Delete(comname,1,4);
   j:= FindFunc(comname);
   if j < 0 then
   begin
    ScriptLogForm.AddLog('function "'+comname+'" not found',i);
    goto nextline;
   end;
   localObj:= nil;
   if CalledFunction <> nil then
    localObj:= CalledFunction;
   CalledFunction:= TFunction.Create;
   CalledFunction.Script:= @Self;
   CalledFunction.Init(functionlist[j],params);
   if localObj <> nil then
    Calledfunction.From:= @localObj;
   CalledFunction.calledpos:= i;
   i:= Calledfunction.pos;
   goto nextline;
  end;
  //=====================
  // loop (@circles)
  //=====================
  if comname = 'loop' then
  begin
   inc(loopcount);
   //params:= GetComParams(scriptlist[i]);
   looppos[loopcount]:= i;
   par[1]:= GetValue(GetComParam(params,1),vtInteger);
   loopcircles[loopcount]:= strtoint(par[1]);
   goto nextline;
  end;
  //=====================
  // endloop
  //=====================
  if comname = 'endloop' then
  begin
   dec(loopcircles[loopcount]);
   if loopcircles[loopcount] > 1 then
    i:= looppos[loopcount]
   else
    dec(loopcount);
   goto nextline;
  end;
  //=====================
  // Jump (@name)
  //=====================
  if comname = 'Jump' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtString);
   j:= FindRunPoint(par[1]);
    if j >= 0 then
     i:= RunPoints[j].pos
    else
     ScriptLogForm.AddLog('not found run point with name "'+par[1]+'"',i);
   goto nextline;
  end;
  //=====================
  // RunScript(@path,@runpoint,@params)
  //=====================
  if comname = 'RunScript' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtString);
   par[2]:= GetValue(GetComParam(params,2),vtString);
   par[3]:= GetValue(GetComParam(params,3),vtString);
   localObj:= nil;
   try
    localObj:= TMGScript.Create(par[1],par[2],par[3]);
    (localObj As TMGScript).Compile;
   finally
    localObj.Destroy;
   end;
   goto nextline;
  end;
  //=====================
  // RunThreadScript(@path,@runpoint,@params)
  //=====================
  if comname = 'RunThreadScript' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtString);
   par[2]:= GetValue(GetComParam(params,1),vtString);
   par[3]:= GetValue(GetComParam(params,1),vtString);
   RunMGScript(par[1],par[2],par[3]);
   goto nextline;
  end;
  //=====================
  // DebugMsg (@msg)
  //=====================
  if comname = 'DebugMsg' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtString);
   MainForm.AddLog('MGScript','Debug msg from script: '+par[1],nil);
   //ShowMessage(par[1]);
   goto nextline;
  end;
  //=====================
  // SendText (@str)
  //=====================
  if comname = 'SendText' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtString); //string
   Main.SendAllPlayers(par[1],nil);
   goto nextline;
  end;
  //=====================
  // SendMsg (@msg)
  //=====================
  if comname = 'SendMsg' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtString); //Message
   Main.SendAllPlayers('servermsg("'+par[1]+'")',nil);
   goto nextline;
  end;
  //=====================
  //Sleep(@ms[int])
  //=====================
  if comname = 'Sleep' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtInteger); //ms
   if not CheckInt(par[1]) then goto nextline;
   Sleep(StrToInt(par[1]));
   goto nextline;
  end;
  //=====================
  // Stop
  //=====================
  if comname = 'Stop' then
   exit;
  //Start Word Content Functions
  //*********************
  // -- Dialog Boclk --
  //*********************
  if GetChildObj(comname,1) = 'Dialog' then
  begin
   par[10]:= comname;
   comname:= GetParentObj(comname,1);
   //=================================
   // Create: [int]
   //=================================
   if comname = 'Create' then
   begin
    localObj:= TDialogTable.Create;
    (localObj as TDialogTable).LoadDialogTable(par[1]);
    exit;
   end;
   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HDIALOG
   if not CheckInt(par[1]) then goto nextline;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,DialogCount-1,'dialog array') then goto nextline;
   //==================================
   // LoadFromFile(path [str])
   // load dialog table from file
   //==================================
   if comname = 'LoadFromFile' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Path
    DialogList[int[1]].LoadDialogTable(par[2]);
    goto nextline;
   end;
   //===================================
   // SetTitle(title [str])
   //===================================
   if comname = 'SetTitle' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString);
    DialogList[int[1]].SetTitle(par[2]);
    goto nextline;
   end;
   //===================================
   // ChangeOwner(char [int])
   // change owner
   //===================================
   if comname = 'ChangeOwner' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtInteger); //char
    if not CheckInt(par[2]) then goto nextline;
    int[2]:= StrToInt(par[2]);
    if not CheckRange(int[2],0,UnitCount-1,'unit array') then goto nextline;
    DialogList[int[1]].char:= UnitList[int[2]];
    with (UnitList[int[2]] as TCharA).client do
    begin
     Send('d_b('+IntToStr(DialogList[int[1]].HDialog)+')');
    end;
    goto nextline;
   end;
   //===================================
   // ChangeBtnTxt(@tableid [int], @id [int], @text [str])
   // Change Dialog Button Text
   //===================================
   if comname = 'ChangeBtnTxt' then
   begin
    par[2] := GetValue(GetComParam(params,1),vtInteger); //tableid
    par[3] := GetValue(GetComParam(params,2),vtInteger); //btnid
    if not CheckInt(par[2]) then goto nextline;
    if not CheckInt(par[3]) then goto nextline;
    int[2] := StrToInt(par[2]);
    int[3] := StrToInt(par[3]);
    par[2] := GetValue(GetComParam(params,3),vtString); //text
    DialogList[int[1]].ChangeButtonText(int[2],int[3],par[2]);
    Exit;
   end;
   //===================================
   // ShowDialog(table [int])
   // show dialog for char
   //===================================
   if comname = 'ShowDialog' then
   begin
    par[3]:= GetValue(GetComParam(params,1),vtInteger); //dialog table number
    if not CheckInt(par[3]) then goto nextline;
    if not CheckChar(int[1]) then goto nextline;
    int[3]:= StrToInt(par[3]);
    DialogList[int[1]].ShowDialog(int[3]);
    goto nextline;
   end;
   //====================================
   // Destroy
   // destroy dialog
   //====================================
   if comname = 'Destroy' then
   begin
    DialogList[int[1]].Destroy;
    exit;
   end;
   //#####################
   goto nextline;
  end;

  //*********************
  // -- Quest Block --
  //*********************
  if GetChildObj(comname,1) = 'Quest' then
  begin
    par[10] := comname;
    comname := GetParentObj(comname,1);
    //=================
    // Create (id [int], owner [hunit])
    //=================
    if comname = 'Create' then
    begin
      par[1] := GetValue(GetComParam(params,1),vtInteger); //questid
      par[2] := GetValue(GetComParam(params,2),vtInteger); //hunit
      for j := 1 to 2 do
       if CheckInt(par[j]) then
         int[j] := StrToInt(par[j])
       else
         goto nextline;
      if not CheckRange(int[2], 0, UnitCount-1, 'unit array') then goto nextline;
      if not CheckRange(int[1], 0, QuestDataBase.Count-1, 'quest array') then goto nextline;
      if not CheckChar(int[2]) then goto nextline;
      localObj := UnitList[int[2]];
      (localObj as TCharA).client.Send('qnew('+QuestDataBase.ToStringNew(int[1])+')');
      QuestDataBase.CheckTargetEffect(int[1],localObj as TUnit, false);
      for j := 0 to 15 do
        if QuestDataBase.List[int[1]].qItem[j] <> '' then
          (localObj as TCharA).client.Send( 'qnitem('+QuestDataBase.ItemToString(int[1],j,localObj as TUnit)+')' );
      (localObj as TCharA).client.Send('qbflush');
      goto nextline;
    end;
    //=================
    // Flush (owner [hunit])
    //=================
    if comname = 'Flush' then
    begin
      par[1] := GetValue(GetComParam(params,1),vtInteger); //hunit
      if CheckInt(par[j]) then int[1] := StrToInt(par[1]) else goto nextline;
      if not CheckRange(int[1], 0, UnitCount-1, 'unit array') then goto nextline;
      if not CheckChar(int[1]) then goto nextline;
      (UnitList[int[1]] as TCharA).client.Send('qbflush');
      goto nextline;
    end;
    //=================
    //  SetGiveUnit (id [int], unit [hunit])
    //  SetTargetUnit (id [int], unit [hunit])
    //=================
    if (comname = 'SetGiveUnit') or (comname = 'SetTargetUnit') then
    begin
      par[1] := GetValue(GetComParam(params,1),vtInteger); //questid
      par[2] := GetValue(GetComParam(params,2),vtInteger); //hunit
      for j := 1 to 2 do
        if CheckInt(par[j]) then int[j] := StrToInt(par[j]) else goto nextline;
      if not CheckRange(int[2], 0, UnitCount-1, 'unit array') then goto nextline;
      if not CheckRange(int[1], 0, QuestDataBase.Count-1, 'quest array') then goto nextline;
      if comname = 'SetGiveUnit' then
        QuestDataBase.List[int[1]].giverUnit := int[2]
      else
        QuestDataBase.List[int[1]].targetUnit := int[2];
      goto nextline;
    end;
    //=================
    // Remove (id [int], owner [hunit])
    //=================
    if comname = 'Remove' then
    begin
      par[1] := GetValue(GetComParam(params,1),vtInteger); //questid
      par[2] := GetValue(GetComParam(params,2),vtInteger); //hunit
      for j := 1 to 2 do
       if CheckInt(par[j]) then
         int[j] := StrToInt(par[j])
       else
         goto nextline;
      if not CheckRange(int[2], 0, UnitCount-1, 'unit array') then goto nextline;
      if not CheckRange(int[1], 0, QuestDataBase.Count-1, 'quest array') then goto nextline;
      if not CheckChar(int[2]) then goto nextline;
      localObj := UnitList[int[2]];
      QuestDataBase.CheckTargetEffect(int[1], localObj as TUnit, false);
      QuestDataBase.CheckOtherQuests(int[1],localObj as TUnit);
      (localObj as TCharA).client.Send('qrem('+par[1]+')');
      goto nextline;
    end;
    //=================
    // Update (id [int], owner [hunit])
    //=================
    if comname = 'Update' then
    begin
      par[1] := GetValue(GetComParam(params,1),vtInteger); //questid
      par[2] := GetValue(GetComParam(params,2),vtInteger); //hunit
      for j := 1 to 2 do
       if CheckInt(par[j]) then
         int[j] := StrToInt(par[j])
       else
         Exit;
      if not CheckRange(int[2], 0, UnitCount-1, 'unit array') then Exit;
      if not CheckRange(int[1], 0, QuestDataBase.Count-1, 'quest array') then Exit;
      if not CheckChar(int[2]) then Exit;
      localObj := UnitList[int[2]];
      QuestDataBase.CheckTargetEffect(int[1], localObj as TUnit, false);
      if (localObj as TCharA).Data[QuestDataBase.List[int[1]].statusPath] = 'true' then
      begin
        for j := 0 to 15 do
          if QuestDataBase.List[int[1]].qItem[j] <> '' then
          begin
            (localObj as TCharA).client.Send('qcitemd('+QuestDataBase.ItemToString(int[1],j,localObj as TUnit)+')'); 
          end;
      end;
      goto nextline;
    end;
    //=============
    comname := par[10];
  end;

  //*********************
  // -- Unit Group Block --
  //*********************
  if GetChildObj(comname,1) = 'UnitGroup' then
  begin
    par[10]:= comname;
    comname := GetParentObj(comname,1);
    //======================
    // Create
    //======================
    if comname = 'Create' then
    begin
      TUnitGroup.Create;
      goto nextline;
    end;
    //----
    par[1] := GetValue(GetParentObj(par[10],2),vtInteger); //HUGroup
    if not CheckInt(par[1]) then goto nextline;
    int[1] := StrToInt(par[1]);
    if not Checkrange(int[1], 0, UGroupCount-1, ' unit group') then goto nextline;
    //======================
    // @HUGROUP.AddUnitsFromRange(@locid [int], @x [int], @y [int], @radius [int], @typeset [str])
    // typeset = (P, N, E)
    // P - players
    // N - npc
    // E - enemy
    //======================
    if comname = 'AddUnitsFromRange' then
    begin
      par[2] := GetValue(GetComParam(params,1),vtInteger); //locid
      par[3] := GetValue(GetComParam(params,2),vtInteger); //x
      par[4] := GetValue(GetComParam(params,3),vtInteger); //y
      par[5] := GetValue(GetComParam(params,4),vtInteger); //radius
      par[6] := GetValue(GetComParam(params,5),vtString); //typeset
      for j:= 2 to 5 do
      begin
       if not CheckInt(par[j]) then goto nextline else int[j] := StrToInt(par[j]);
      end;
      UGroupList[int[1]].GetUnitFromRange(int[2],int[3],int[4],int[5],StrToUnitTypeSet(par[6]));
      goto nextline;
    end;
    //======================
    // Destroy
    //======================
    if comname = 'Destroy' then
    begin
      UGroupList[int[1]].Destroy;
      goto nextline;
    end;
    //###################
    comname := par[10];
  end;
  //*********************
  // -- End Unit Group Block --
  //*********************
    
  //*********************
  // -- Char Block --
  //*********************
  if GetChildObj(comname,1) = 'Char' then
  begin
   par[10]:= comname;
   comname:= GetParentObj(comname,1);
   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HUNIT
   if not CheckInt(par[1]) then goto nextline;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,UnitCount-1,'unit array') then goto nextline;
   if not CheckChar(int[1]) then goto nextline;
   //================
   // @HUNIT[int].TextTagMsg(@msg [str])
   //================
   if comname = 'TextTagMsg' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString);
    if Assigned((UnitList[int[1]] as TCharA).client) then
     (UnitList[int[1]] as TCharA).client.Send('ttmsg('+IntToStr(UnitList[int[1]].HUnit)+',"'+par[2]+'")');
    goto nextline;
   end;
   //================
   // @HUNIT.TextTagMsgEx(@msg [str], @x [int], @y [int], @size [rea])
   //================
   if comname = 'TextTagMsgEx' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //msg
    par[3]:= GetValue(GetComParam(params,2),vtInteger); //x
    par[4]:= GetValue(GetComParam(params,3),vtInteger); //y
    par[5]:= GetValue(GetComParam(params,4),vtReal); //size
    if Assigned((UnitList[int[1]] as TCharA).client) then
      (UnitList[int[1]] as TCharA).client.Send('ttmex("'+par[2]+'",'+par[3]+','+par[4]+','+par[5]+')');
    goto nextline;
   end;
   //================
   // @HUNIT[int].SetEnemy(unit[int],bol[bol])
   // SetUnit is enemy to current unit
   //================
   if comname = 'SetEnemy' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtInteger); //HUNIT - target
    par[3]:= GetValue(GetComParam(params,2),vtBool); //boolean - IsEnemy
    if not CheckInt(par[2]) then goto nextline;
    if Assigned((UnitList[int[1]] as TCharA).client) then
     (UnitList[int[1]] as TCharA).client.Send('setenemy('+par[2]+','+par[3]+')');
    goto nextline;
   end;
   //================
   // @HUNIT[int].SendMsg(@msg[str])
   // Send private msg to char
   //================
   if comname = 'SendMsg' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString);
    if Assigned((UnitList[int[1]] as TCharA).client) then
     (UnitList[int[1]] as TCharA).client.Send('servermsg("'+par[2]+'")')
    else
     ScriptLogForm.AddLog('Units['+par[1]+']: Client socket not assigned',i);
    goto nextline;
   end;
   //#########################
   //goto nextline;
   comname:= par[10];
  end;
  //*********************
  // -- Unit Block --
  //*********************
  if GetChildObj(comname,1) = 'Unit' then
  begin
   par[10]:= comname;
   comname:= GetParentObj(comname,1);
   //=====================
   // Create (@utype,@rawcode,@locid,@x,@y,@yaw,@life,@mana)
   // Create unit (RAWCODE) from UTYPE in game word to X,Y with LIFE and MANA and fasing to YAW
   //=====================
   if comname = 'Create' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtString);//utype
    par[2]:= GetValue(GetComParam(params,2),vtString);//rawcode
    par[8]:= GetValue(GetComParam(params,3),vtInteger);//locID
    par[3]:= GetValue(GetComParam(params,4),vtInteger);//x
    par[4]:= GetValue(GetComParam(params,5),vtInteger);//y
    par[5]:= GetValue(GetComParam(params,6),vtInteger);//yaw
    par[6]:= GetComParam(params,7);//life
    if not UnitGetDirective(par[2],par[6]) then
     par[6]:= GetValue(par[6],vtInteger);
    par[7]:= GetComParam(params,8);//mana
    if not UnitGetDirective(par[2],par[7]) then
     par[7]:= GetValue(par[7],vtInteger);
    for j := 3 to 8 do
     if not CheckInt(par[j]) then goto nextline;
    int[2]:= StrToInt(par[8]);
    localObj:= TUnit.Create(par[2],int[2],StrToInt(par[3]),StrToInt(par[4]),0,StrToInt(par[5]),StrToInt(par[6]),StrToInt(par[7]),StrToUType(par[1]));
    int[1]:= (localObj as TUnit).HUnit;
    if par[1] = 'npc' then j:= 1;
    if par[1] = 'enemy' then j:= 2;
    SendAllPlayersOnLoc('createunit('+IntToStr(int[1])+','+par[2]+','+IntToStr(j)+')',int[2],nil);
    SendAllPlayersOnLoc('newXY('+(localObj as TUnit).GetXY+')',int[2],nil);
    SendAllPlayersOnLoc('newHM('+(localObj as TUnit).GetHM+')',int[2],nil);
    goto nextline;
   end;
   //======================
   // SendOrderTarget(@unit:[int],@target:[int],@order[str])
   // Send ORDER to UNIT with TARGET
   //======================
   if comname = 'SendOrderTarget' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtInteger);//unit
    par[2]:= GetValue(GetComParam(params,2),vtInteger);//target
    par[3]:= GetValue(GetComParam(params,3),vtString);//order
    for j:= 1 to 2 do
     if not CheckInt(par[j]) then goto nextline;
    int[1]:= StrToInt(par[1]);
    if not CheckRange(int[1],0,UnitCount-1,'unit array') then goto nextline;
    int[2]:= StrToInt(par[2]);
    UnitList[int[1]].SendOrderTargetUnit(int[2],par[3]);
    goto nextline;
   end;
   //======================
   // SendOrderXY(@unit:[int],@x:[int],@y:[int],@order:[str])
   // Send ORDER to UNIT to (X,Y) coord
   //======================
   if comname = 'SendOrderXY' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtInteger); //unit
    par[2]:= GetValue(GetComParam(params,2),vtInteger); //X coord
    par[3]:= GetValue(GetComParam(params,3),vtInteger); //Y coord
    par[4]:= GetValue(GetComParam(params,4),vtString); //Order
    for j:= 1 to 3 do
     if not CheckInt(par[j]) then goto nextline;
    int[1]:= StrToInt(par[1]);
    if not CheckRange(int[1],0,UnitCount-1,'unit array') then goto nextline;
    UnitList[int[1]].SendOrderTarget(StrToInt(par[2]),StrToInt(par[3]),par[4]);
    goto nextline;
   end;
   //======================
   // SendOrder(@unit:[int],@order:[str]);
   // Send ORDER to UNIT
   //======================
   if comname = 'SendOrder' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtInteger); //Unit
    par[2]:= GetValue(GetComParam(params,2),vtString); //Order
    if not CheckInt(par[1]) then goto nextline;
    int[1]:= StrToInt(par[1]);
    if not CheckRange(int[1],0,UnitCount-1,'unit array') then goto nextline;
    UnitList[int[1]].SendOrder(par[2]);
    goto nextline;
   end;
   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger);//HUNIT
   if not CheckInt(par[1]) then goto nextline;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,UnitCount-1,'unit array') then goto nextline;
   //======================
   // @HUNIT[int].Talk(@text)
   // HUNIT talk TEXT in the game
   //======================
   if comname = 'Talk' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString);//Text
    SendAllPlayersOnRange('unittalk('+par[1]+',"'+UnitList[int[1]].name+'","'+par[2]+'")',UnitList[int[1]].locid,UnitList[int[1]].posx,UnitList[int[1]].posy,2000,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].Pause(flag[bol])
   //======================
   if comname = 'Pause' then
   begin
    par[2] := GetValue(GetComParam(params,1),vtBool); //flag
    UnitList[int[1]].ispause := par[2] = 'true';
    SendAllPlayersOnLoc('pu('+IntToStr(UnitList[int[1]].HUnit)+','+par[2]+')',UnitList[int[1]].locid,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].SetInvulnerable(flag[bol])
   //======================
   if comname = 'SetInvulnerable' then
   begin
    par[2] := GetValue(GetComParam(params,1),vtBool); //flag
    UnitList[int[1]].isinvul := par[2] = 'true';
    SendAllPlayersOnLoc('uinv('+IntToStr(UnitList[int[1]].HUnit)+','+par[2]+')',UnitList[int[1]].locid,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].Revive(@locid[int],@x[int],@y[int],@HP[int],@MP[int],@ShowEffects[bol])
   // ReviveUnit in (X,Y) coordinate with HP and MP
   //======================
   if comname = 'Revive' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtInteger); //X
    par[3]:= GetValue(GetComParam(params,2),vtInteger); //Y
    par[4]:= GetValue(GetComParam(params,3),vtInteger); //Health
    par[5]:= GetValue(GetComParam(params,4),vtInteger); //Mana
    par[6]:= GetValue(GetComParam(params,5),vtBool); //Show Effects
     for j:= 2 to 5 do
      if not CheckInt(par[j]) then goto nextline;
    UnitList[i].posx := StrToInt(par[2]);
    UnitList[i].posy := StrToInt(par[3]);
    Main.SendAllPlayersOnLoc('newHM('+unitList[i].GetHM+')',UnitList[i].locid,nil);
    Main.SendAllPlayersOnLoc('revive('+par[1]+','+par[2]+','+par[3]+','+par[6]+')',UnitList[i].locid,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].AddBuyItem(@id[str],@count[int],@max[int])
   // Add Item for buy to unit
   //======================
   if comname = 'AddBuyItem' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Item Raw Code
    par[3]:= GetValue(GetComParam(params,2),vtInteger); //Count
    par[4]:= GetValue(GetComParam(params,3),vtInteger); //Max count
    for j:= 3 to 4 do
     if not CheckInt(par[j]) then goto nextline;
    UnitList[int[1]].AddStockItem(par[2],StrToInt(par[3]),StrToInt(par[4]));
    Main.SendAllPlayersOnLoc('addbit('+par[1]+',"'+par[2]+'",'+par[3]+','+par[4]+')',UnitList[int[1]].locid,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].FlushData(filed[str])
   // Flush data in unit data table
   //======================
   if comname = 'FlushData' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Filed
    UnitList[int[1]].FlushData(par[2]);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].WriteValue(field[str],value[str],isLoc[bol])
   // Write data in unit data table
   //======================
   if comname = 'WriteValue' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Field
    par[3]:= GetValue(GetComParam(params,2),vtString); //Value
    par[4]:= GetValue(GetComParam(params,3),vtBool); //isLocal
    if par[4] = 'true' then
      UnitList[int[1]].DataLoc[par[2]]:= par[3]
    else
      UnitList[int[1]].Data[par[2]]:= par[3];
    goto nextline;
   end;
   //======================
   // @HUNIT[int].AddAbility(@id[str])
   // Add unit new spell
   //======================
   if comname = 'AddAbility' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Ability Raw Code
    if UnitList[int[1]].AddAbility(par[2]) then
     Main.SendAllPlayersOnLoc('addability('+par[1]+',"'+par[2]+'")',UnitList[int[1]].locid,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].PlayAnimIndex(@animindex [int])
   // Play unit animation by index
   //======================
   if comname = 'PlayAnimIndex' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtInteger); //AnimIndex
    if not CheckInt(par[2]) then goto nextline;
    UnitList[int[1]].PlayAnim(StrToInt(par[2]));
    goto nextline;
   end;
   //======================
   // @HUNIT[int].PlayAnim(@animname [str])
   // Play unit animation
   //======================
   if comname = 'PlayAnim' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //AnimName
    UnitList[int[1]].PlayAnim(par[2]);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].RemoveAbility(@id[str])
   // Remove unit ability
   //======================
   if comname = 'RemoveAbility' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Ability Raw Code
    int[2]:= UnitList[int[1]].FindAbility(par[2]);
    if CheckRange(int[2],0,5,'unit spell list') then goto nextline;
    UnitList[int[1]].spells[int[2]].rawcode:= '';
    UnitList[int[1]].spells[int[2]].lvl:= 0;
    Main.SendAllPlayersOnLoc('removeability('+par[1]+',"'+par[2]+'")',UnitList[int[1]].locid,nil);
    goto nextline;
   end;
   //======================
   // @HUNIT[int].SetAbilityLevel(@id[str],@lvl[int])
   // Remove unit ability
   //======================
   if comname = 'SetAbilityLevel' then
   begin
    par[2]:= GetValue(GetComParam(params,1),vtString); //Ability Raw Code
    int[2]:= UnitList[int[1]].FindAbility(par[2]);
    if not CheckRange(int[2],0,5,'unit spell list') then goto nextline;
    par[3]:= GetValue(GetComParam(params,2),vtInteger); //New Ability Level
    if not CheckInt(par[3]) then goto nextline;
    int[3]:= UnitList[int[1]].spells[int[2]].lvl;
    int[4]:= StrToInt(par[3]);
    if int[3] = int[4] then goto nextline;
    UnitList[int[1]].spells[int[2]].lvl:= int[4];
    if int[4] = 0 then
     Main.SendAllPlayersOnLoc('removeability('+par[1]+',"'+par[2]+'")',UnitList[int[1]].locid,nil)
    else
    begin
     if int[3] = 0 then
      Main.SendAllPlayersOnLoc('addability('+par[1]+','+par[2]+')',UnitList[int[1]].locid,nil)
     else
      Main.SendAllPlayersOnLoc('setablvl('+par[1]+',"'+par[2]+'",'+par[3]+')',UnitList[int[1]].locid,nil);
    end;
    goto nextline;
   end;
   //======================
   // @HUNIT[int].Remove
   // Remove HUNIT in game word
   //======================
   if comname = 'Remove' then
   begin
    if UnitList[int[1]]._type <> utPlayer then
     Main.SendAllPlayersOnLoc('removeunit('+par[1]+')',UnitList[int[1]].locid,nil);
    UnitList[int[1]].Destroy;
    goto nextline;
   end;
  //######################
   //goto nextline;
   comname:= par[10];
  end;
  //**********************
  // -- END UNIT BLOCK
  //**********************

 //****************************
 // -- EFFECT BLOCK --
 //****************************
 if GetChildObj(comname,1) = 'Effect' then
 begin
  par[10] := comname;
  comname := GetParentObj(comname,1);
  //=================================
  //  Create
  //=================================
  if comname = 'Create' then
  begin
    TEffect.Create;
    goto nextline;
  end;
  par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HEFFECT
  if not CheckInt(par[1]) then goto nextline;
  int[1]:= StrToInt(par[1]);
  if not CheckRange(int[1],0,EffectCount-1,'effect array') then goto nextline;
  //==================================
  //  @HEFFECT.AddEffect(model [str], locid [int], x [int], y [int])
  //==================================
  if comname = 'AddEffect' then
  begin
    par[2] := GetValue(GetComParam(params,1),vtString); //modelpath
    par[3] := GetValue(GetComParam(params,2),vtInteger); //locid
    par[4] := GetValue(GetComParam(params,3),vtInteger); //x
    par[5] := GetValue(GetComParam(params,4),vtInteger); //y
    for j := 3 to 5 do
      if not CheckInt(par[j]) then goto nextline else int[j] := StrToInt(par[j]);
    EffectList[int[1]].AddEffect(par[2],int[3],int[4],int[5]);
    SendAllPlayersOnLoc(EffectList[int[1]].GetEffectFunc,int[3],nil);
    goto nextline;
  end;
  //===================================
  //  @HEFFECT.AddEffectTarget(model [str], attach [str], unit [int])
  //===================================
  if comname = 'AddEffectTarget' then
  begin
    par[2] := GetValue(GetComParam(params,1),vtString); //model
    par[3] := GetValue(GetComParam(params,2),vtString); //attach
    par[4] := GetValue(GetComParam(params,3),vtInteger); //unit (HUNIT)
    if not CheckInt(par[4]) then goto nextline;
    int[2] := StrToInt(par[4]);
    int[3] := int[2];
    if not CheckRange(int[2],0,UnitCount-1,'unit array') then goto nextline;
    EffectList[int[1]].locid := UnitList[int[2]].locid;
    EffectList[int[1]].AddEffectTarget(par[2],par[3],int[3]);
    SendAllPlayersOnLoc(EffectList[int[1]].GetEffectFunc,UnitList[int[2]].locid,nil);
    goto nextline;
  end;
  //===================================
  // @HEFFECT.Destroy
  //===================================
  if comname = 'Destroy' then
  begin
    EffectList[int[1]].Destroy;
    SendAllPlayersOnLoc('edel('+par[1]+')',EffectList[int[1]].locid,nil);
    goto nextline;
  end;
  //#################
  comname := par[10];
 end;
 //****************************
 // -- END EFFECT BLOCK --
 //****************************

  //*************************
 // -- ITEM BLOCK --
 //*************************
 if GetChildObj(comname,1) = 'Item' then
 begin
  par[10]:= comname;
  comname:= GetParentObj(comname,1);
  //=======================
  // Create(@locid,@x,@y,@rawcode)
  // Create Item in game filed
  //=======================
  if comname = 'Create' then
  begin
   par[1]:= GetValue(GetComParam(params,1),vtInteger);
   par[2]:= GetValue(GetComParam(params,2),vtInteger);
   par[3]:= GetValue(GetComParam(params,3),vtInteger);
   par[4]:= GetValue(GetComParam(params,4),vtString);
   for j:= 1 to 3 do
    if not CheckInt(par[j]) then exit;
   int[1] := StrToInt(par[1]);
   localObj:= TItm.Create(par[4],int[1],StrToInt(par[2]),StrToInt(par[3]));
   SendAllPlayersOnLoc('newitm('+par[2]+','+par[3]+','+par[4]+','+IntToStr((localObj as TItm).HItem)+')',int[1],nil);
   goto nextline;
  end;
  par[1]:= GetValue(GetParentObj(par[10],2),vtInteger); //HITEM
  if not CheckInt(par[1]) then goto nextline;
  int[1]:= StrToInt(par[1]);
  if not CheckRange(int[1],0,ItemCount-1,'item array') then goto nextline;
  //=======================
  // @itm.TimerRun(@time [int])
  //=======================
  if comname = 'TimerRun' then
  begin
   par[2]:= GetValue(GetComParam(params,1),vtInteger); //time
   if not CheckInt(par[2]) then goto nextline;
   int[2]:= StrToInt(par[2]);
   ItemList[int[1]].RunTimer(int[2]);
   exit;
  end;
  //=======================
  // @itm.AddItem(@Unit)
  //=======================
  if comname = 'AddItem' then
  begin
   par[2]:= GetValue(GetComParam(params,1),vtInteger);
   if not CheckInt(par[2]) then goto nextline;
   int[2]:= StrToInt(par[2]);
   if not CheckRange(int[2],0,UnitCount-1,'unit array') then goto nextline;
   if ItemList[int[1]].locID <> UnitList[int[2]].locid then
   begin
    ItemList[int[1]].SendRemove;
    ItemList[int[1]].locID := UnitList[int[2]].locid;
    Main.SendAllPlayersOnLoc('newitm(0,0,'+ItemList[int[1]].RawCode+','+IntToStr(ItemList[int[1]].HItem)+')',ItemList[int[1]].locId,nil);
   end;
   if ItemList[int[1]].owner >= 0 then
   begin
    int[3] := ItemList[int[1]].owner;
    if int[3] >= 0 then UnitList[int[3]].DropItem(ItemList[int[1]].HItem);
   end;
   if UnitList[int[2]].AddItem(ItemList[int[1]].HItem) then
   begin
    ItemList[int[1]].SetOwnerD(UnitList[int[2]].HUnit);
    Main.SendAllPlayersOnLoc('additm('+ItemList[int[1]].GetInfo+')',ItemList[int[1]].locID,nil);
   end;
   goto nextline;
  end;
  //=======================
  // @itm.Remove
  //=======================
  if comname = 'Remove' then
  begin
   ItemList[int[1]].SendRemove;
   ItemList[int[1]].Destroy;
   goto nextline;
  end;
  //#######################
  //goto nextline;
  comname:= par[10];
 end;
 //*************************
 // -- END ITEM BLOCK --
 //*************************

 //*************************
 // -- RECT BLOCK --
 //*************************
 if GetChildObj(comname,1) = 'Rect' then
 begin
  par[10] := comname;
  comname := GetParentObj(comname,1);
  //=============================
  // Create(locid [int],x [int],y [int],w [int],h [int])
  // Create new Rect in LocID with coordinate X,Y and container W,H
  //=============================
  if comname = 'Create' then
  begin
    par[1] := GetValue(GetComParam(params,1),vtInteger); //LocId
    par[2] := GetValue(GetComParam(params,2),vtInteger); //X
    par[3] := GetValue(GetComParam(params,3),vtInteger); //Y
    par[4] := GetValue(GetComParam(params,4),vtInteger); //W
    par[5] := GetValue(GetComParam(params,5),vtInteger); //H
    for j := 1 to 5 do
      if not CheckInt(par[j]) then goto nextline else int[j] := StrToInt(par[j]);
    TRect.Create(int[1], int[2],int[3], int[4],int[5]);
    goto nextline;
  end;
  par[1]:= GetValue(GetChildObj(par[10],2),vtInteger);//HRECT
  if not CheckInt(par[1]) then goto nextline;
  int[1]:= StrToInt(par[1]);
  if not CheckRange(int[1],0,RectCount-1,'rect array') then goto nextline;
  //=============================
  // @HRECT [int].Remove
  // Remove rect
  //=============================
  if comname = 'Remove' then
  begin
    RectList[int[1]].Destroy;
    goto nextline;
  end;
  //#####################
  comname := par[10];
 end;
 //***********************
 // -- END RECT BLOCK
 //***********************

  //**********************
  // -- DEST BLOCK --
  //**********************
  if GetChildObj(comname,1) = 'Dest' then
  begin
   par[10]:= comname;
   comname:= GetParentObj(comname,1);
  //======================
  // Create(@rawcode,@locid,@x[int],@y[int],@variation[int],@yaw[int])
  // create a Dest (@rawcode) in @X,@Y,@Z coordinate with @yaw
  //======================
   if comname = 'Create' then
   begin
    par[1]:= GetValue(GetComParam(params,1),vtString); //RawCode
    par[2]:= GetValue(GetComParam(params,2),vtInteger);//locId
    par[3]:= GetValue(GetComParam(params,3),vtInteger);//X
    par[4]:= GetValue(GetComParam(params,4),vtInteger);//Y
    par[5]:= GetValue(GetComParam(params,5),vtInteger);//Z
    par[6]:= GetValue(GetComParam(params,6),vtInteger);//Yaw
    for j:= 2 to 6 do
     if not CheckInt(par[j]) then Exit;
    int[1] := StrToInt(par[2]);
    localObj:= TDest.Create(par[1],int[1],StrToInt(par[3]),StrToInt(par[4]),StrToInt(par[5]),StrToInt(par[6]));
    Main.SendAllPlayersOnLoc('destcreate('+(localObj As TDest).GetMainInfo+')',int[1],nil);
    localObj:= nil;
    goto nextline;
   end;
   par[1]:= GetValue(GetParentObj(par[10],2),vtInteger);//HDEST
   if not CheckInt(par[1]) then goto nextline;
   int[1]:= StrToInt(par[1]);
   if not CheckRange(int[1],0,DestCount-1,'dest array') then goto nextline;
   //======================
   // @HDEST[int].Destroy
   //======================
   if comname = 'Destroy' then
   begin
    Main.SendAllPlayersOnLoc('destd('+par[1]+')',DestList[int[1]].locid,nil);
    DestList[int[1]].Destroy;
    goto nextline;
   end;
  //######################
   //goto nextline;
   comname:= par[10];
  end;
  //**********************
  // -- END DEST BLOCK
  //**********************

  //======================
  // ** VARIABLES
  // def [vt] @varname = @value
  //======================
  params:= buf;
  if Copy(params,1,3) = 'def' then
  begin
   bol[1]:= false;
   if Copy(params,5,5) = 'local' then
   begin
    Delete(params,5,6);
    bol[1]:= true;
   end;
   par[3]:= Copy(params,5,5); //Var Type
   params:= Copy(params,11,Length(params));
   j:= Pos('=',params);
   if j <= 0 then
   begin
    par[1]:= params;
    par[2]:= '';
   end
   else
   begin
    par[1]:= Copy(params,1,j-2); //Name
    par[2]:= RemoveSpace(Copy(params,j+2,Length(params))); //Get Value
   end;
   par[2]:= GetValue(par[2],ConvertVarType(par[3])); //Get Value
   if bol[1] then
   begin
    if CalledFunction <> nil then
     CalledFunction.AddLocalVariable(par[1],ConvertVarType(par[3]),par[2])
    else
     AddLocalVariable(par[1],ConvertVarType(par[3]),par[2]);
   end
   else
   begin
    if FindVarId(par[1]) < 0 then
     TVar.New(par[1],ConvertVarType(par[3]),par[2]);
   end;
   goto nextline;
  end;
  //======================
  // @varname:= @newvalue
  //======================
  if Pos(':=',buf) > 0 then
  begin
   params:= TrimLeft(buf);
   j:= Pos(':=',params);
   par[1]:= Copy(params,0,j-1); //variable name
   par[2]:= RemoveSpace(Copy(params,j+2,Length(params)-j-1)); //value
   bol[5]:= false;
   //if ARRAY
   if par[1][j-1] = ']' then
   begin
    j:= Pos('[',par[1]);
    bol[5]:= true;
    par[3]:= par[1];
    par[1]:= Copy(par[1],1,j-1);
    par[9]:= Copy(par[3],j+1,Length(par[3])-j-1);
    int[1]:= StrToInt(GetValue(par[9],vtInteger));
   end;
   Delete(params,1,j+2);
   j:= -1;
   //Find Function Local Variable
   if CalledFunction <> nil then
   begin
    j:= CalledFunction.FindLocalVariable(par[1]);
    if j >= 0 then
    begin
     if bol[5] then
      CalledFunction.locvariable[j].ArrayValue[int[1]]:= GetValue(par[2],CalledFunction.locvariable[j].t)
     else
      CalledFunction.locvariable[j].value:= GetValue(par[2],CalledFunction.locvariable[j].t);
     goto nextline;
    end;
   end
   else
      //Find Local Variable
      if j < 0 then
      begin
       j:= FindLocalVariable(par[1]);
       if j >= 0 then
       begin
        if bol[5] then
         FVars[j].ArrayValue[int[1]]:= GetValue(par[2],FVars[j].t)
        else
         FVars[j].value:= GetValue(par[2],FVars[j].t);
        goto nextline;
       end;
   end;
   //Find Globals Varible
   if j < 0 then
    j:= FindVarId(par[1]);
   if j < 0 then
   begin
    //Check on DB value

    //**************************************************************************
    // -- UNIT BLOCK
    if GetChildObj(par[1],1) = 'Unit' then
    begin
      par[10]:= comname;
      comname:= GetParentObj(par[1],1);
      par[3]:= GetValue(GetParentObj(par[1],2),vtInteger);
      if not CheckInt(par[3]) then goto nextline;
      j:= StrToInt(par[3]);
      if not CheckRange(j,0,UnitCount-1,'unit array') then goto nextline;
      //@HUNIT.X
      if comname = 'X' then
      begin
       UnitList[j].posx:= StrToInt(GetValue(par[2],vtInteger));
       Main.SendAllPlayersOnLoc('newXY('+UnitList[j].GetXY+')',UnitList[j].locid,nil);
       goto nextline;
      end;
      //@HUNIT.Y
      if comname = 'Y' then
      begin
       UnitList[j].posy:= StrToInt(GetValue(par[2],vtInteger));
       Main.SendAllPlayersOnLoc('newXY('+UnitList[j].GetXY+')',UnitList[j].locid,nil);
       goto nextline;
      end;
      //@HUNIT.Rotate
      if comname = 'Rotate' then
      begin
       UnitList[j].yaw:= StrToInt(GetValue(par[2],vtInteger));
       Main.SendAllPlayersOnLoc('newXY('+par[3]+','+UnitList[j].GetXY+')',UnitList[j].locid,nil);
       goto nextline;
      end;
      //@HUNIT.EXP
      if comname = 'EXP' then
      begin
       par[2]:= GetValue(par[2],vtInteger);
       UnitList[j].exp:= StrToInt(par[2]);
       Main.SendAllPlayersOnLoc('setEXP('+par[3]+','+par[2]+')',UnitList[j].locid,nil);
       goto nextline;
      end;
      //@HUNIT.Health
      if comname = 'Health' then
      begin
       UnitList[j].health:= StrToInt(GetValue(par[2],vtInteger));
       Main.SendAllPlayersOnLoc('newHM('+UnitList[j].GetHM+')',UnitList[j].locid,nil);
       if UnitList[j].health < 1 then
         RunMGScript(pathlist.script+'server_event.mgscript','PlayerDeath',IntToStr(UnitList[j].HUnit));
       goto nextline;
      end;
      //@HUNIT.Mana
      if comname = 'Mana' then
      begin
       UnitList[j].mana:= StrToInt(GetValue(par[2],vtInteger));
       Main.SendAllPlayersOnLoc('newHM('+UnitList[j].GetHM+')',UnitList[j].locid,nil);
       goto nextline;
      end;
      //@HUNIT.Name
      if comname = 'Name' then
      begin
       UnitList[j].name:= GetValue(par[2],vtString);
       goto nextline;
      end;
     //###########
     end;
     // END UNIT BLOCK

     //*************************************************************************
     // -- RECT BLOCK
     if GetChildObj(par[1],1) = 'Rect' then
     begin
      par[10] := comname;
      comname := GetParentObj(par[1],1);
      par[3] := GetValue(GetParentObj(par[1],2),vtInteger);
      if not CheckInt(par[3]) then goto nextline;
      j := StrToInt(par[3]);
      if not CheckRange(j,0,RectCount-1,'rect array') then goto nextline;
      //@HRect.Script
      if comname = 'Script' then
      begin
        RectList[j].eventscript := GetValue(par[2],vtString);
        goto nextline;
      end;
      //#############
     end;

     //*************************************************************************
     // -- CHAR BLOCK
     if GetChildObj(par[1],1) = 'Char' then
     begin
      par[10]:= comname;
      comname:= GetParentObj(par[1],1);
      par[3]:= GetValue(GetParentObj(par[1],2),vtInteger);
      if not CheckInt(par[3]) then goto nextline;
      j:= StrToInt(par[3]);
      if not CheckRange(j,0,UnitCount-1,'unit array') then goto nextline;
      if not CheckChar(j) then goto nextline;
      //@HUNIT.Honor
      if comname = 'Honor' then
      begin
       par[2]:= GetValue(par[2],vtInteger);
       (UnitList[j] as TCharA).Honor:= StrToInt(par[2]);
       if Assigned((UnitList[j] as TCharA).client) then
        (UnitList[j] as TCharA).client.Send('seth('+par[2]+')')
       else
        ScriptLogForm.AddLog('Units['+par[3]+']: Client socket not assigned',i);
       goto nextline;
      end;
      if comname = 'PvPKill' then
      begin
        par[2] := GetValue(par[2],vtInteger);
        (UnitList[j] as TCharA).killCount := StrToInt(par[2]);
        goto nextline;
      end;
      //@HUNIT.Gold
      if comname = 'Gold' then
      begin
       par[2]:= GetValue(par[2],vtInteger);
       (UnitList[j] as TCharA).gold:= StrToInt(par[2]);
       if Assigned((UnitList[j] as TCharA).client) then
        (UnitList[j] as TCharA).client.Send('setg('+par[2]+')')
       else
        ScriptLogForm.AddLog('Units['+par[3]+']: Client socket not assigned',i);
       goto nextline;
      end;
      //@HUNIT.SkillPoint
      if comname = 'SkillPoint' then
      begin
       par[2]:= GetValue(par[2],vtInteger);
       (UnitList[j] as TCharA).skillpoints:= StrToInt(par[2]);
       if Assigned((UnitList[j] as TCharA).client) then
        (UnitList[j] as TCharA).client.Send('setsp('+par[2]+')')
       else
        ScriptLogForm.AddLog('Units['+par[3]+']: Client socket not assigned',i);
       goto nextline;
      end;
     //################
     end;
     // END CHAR BLOCK

     //*************************************************************************
     // -- DEST BLOCK
     if GetChildObj(par[1],1) = 'Dest' then
     begin
      par[10]:= comname;
      comname:= GetParentObj(par[1],1);
      par[3]:= GetValue(GetParentObj(par[1],2),vtInteger);
      if not CheckInt(par[3]) then goto nextline;
      j:= StrToInt(par[3]);
      if not CheckRange(j,0,DestCount-1,'dest array') then goto nextline;
      //@HDEST.State
      if comname = 'State' then
      begin
       DestList[j].state:= IntToDestState(StrToInt(GetValue(par[2],vtInteger)));
       Main.SendAllPlayersOnLoc('destst('+par[3]+','+par[2]+')',DestList[j].locid,nil);
       goto nextline;
     end;
     //###########
     end;
     // END DEST BLOCK

   //End check
   end
   else//Change variable value
   begin
    if bol[5] then
     _vars[j].ArrayValue[int[1]]:= GetValue(par[2],_vars[j].t)
    else
     _vars[j].value:= GetValue(par[2],_vars[j].t);
   end;
   goto nextline;
  end;
  params:= GetComParams(buf);
  //======================
  // @varname.Free
  //======================
  if GetParentObj(comname,1) = 'Free' then
  begin
   par[1]:= GetParentObj(comname,2);
   j:= FindVarId(par[1]);
   if j < 0 then goto nextline;
   _vars[j].Destroy;
   goto nextline;
  end;
  //*****End*****
  nextline:
  inc(i);
 end;
end;

{ TScriptThread }

constructor TScriptThread.Create(path, point, params: string);
begin
 inherited Create(true);
 FScript:= TMGScript.Create(path,point,params);
 FreeOnTerminate:= true;
 Resume;
end;

destructor TScriptThread.Destroy;
begin
 FScript.Destroy;
 inherited Destroy;
end;

procedure TScriptThread.Execute;
begin
  //MainForm.AddLog('script start',nil);
  FScript.Compile;
  //MainForm.AddLog('script end',nil);
end;

procedure RunMGScript(path,runpoint,params: string);
begin
 if FileExists(path) then
  TScriptThread.Create(path,runpoint,params)
 else
  MainForm.AddLog('MGScript','Script "'+path+'" not found',nil);
end;


function TMGScript.CheckChar(id: integer): boolean;
begin
 Result:= UnitList[id] is TCharA;
 if not Result then
  ScriptLogForm.AddLog('Unit['+IntToStr(id)+'] is not a Char',-1);
end;

function TMGScript.CheckInt(str: string): boolean;
begin
 Result:= IsIntVar(str);
 if not Result then
  ScriptLogForm.AddLog('"'+str+'" is not integer value',-1);
end;

function TMGScript.CheckRange(int, min, max: integer; rangename: string): boolean;
begin
 Result:= (int >= min) and (int <= max);
 if not Result then
  ScriptLogForm.AddLog('out of range ['+IntToStr(min)+'..'+IntToStr(max)+'] ('+IntToStr(int)+') in '+rangename,-1);
end;

function TMGScript.CheckReal(str: string): boolean;
begin
 Result:= IsRealVar(str);
 if not Result then
  ScriptLogForm.AddLog('"'+str+'" is not real value',-1);
end;

end.
