unit StatDB;

interface

type
  TYear   = integer;
  TMounth = 0..11;
  TDay    = 0..30;
  TStatisticTime = record
    year: TYear;
    month: TMounth;
    day: TDay;
  end;
  TStatistic = record
    Date: TStatisticTime;
    maxUptime: integer;
    maxOnline: integer;
  end;
  TStatisticList = array of TStatistic;
  TStatisticDataBase = class
    private
      FStatCount: integer;
      FStatList: TStatisticList;
      function FStatRead(index: integer): TStatistic;
    public
      path: string;
      property Count: integer read FStatCount;
      property Items[index: integer]: TStatistic read FStatRead;
      procedure Save;
      function Load: boolean;
      procedure UpdateData(uptime,online: integer);
      function GetMaxStatInfo(uptime: boolean): string;
      function GetTimeInfo(Date: string): string;
      function GetLastStatInfo: string;
  end;

var
 StatisticDB: TStatisticDataBase;

implementation

uses Windows,SysUtils;

function StatToString(Stat: TStatistic): string;
begin
 Result := IntToStr(stat.Date.day)+'.'+IntToStr(stat.Date.month)+'.'+IntToStr(stat.Date.year)+','+IntToStr(stat.maxUptime)+','+IntToStr(stat.maxOnline);
end;

function SysToStatTime(st: SystemTime): TStatisticTime;
begin
 Result.year := st.wYear;
 Result.month := st.wMonth;
 Result.day := st.wDay;
end;

function CompareStatTime(t1,t2: TStatisticTime): boolean;
begin
 Result := (t1.year = t2.year) and (t1.month = t2.month) and (t1.day = t2.day);
end;

{ TStatisticDataBase }

function TStatisticDataBase.FStatRead(index: integer): TStatistic;
begin
 Result := FStatList[index];
end;

function TStatisticDataBase.GetLastStatInfo: string;
begin
 if FStatCount > 2 then
  Result := StatToString(FStatList[FStatCount-1]);
end;

function TStatisticDataBase.GetMaxStatInfo(uptime: boolean): string;
begin
 if uptime then
  Result := StatToString(FStatList[0])
 else
  Result := StatToString(FStatList[1]);
end;

function TStatisticDataBase.GetTimeInfo(Date: string): string;
 var p: integer;
     QueryDate: TStatisticTime;
     i: integer;
begin
 p := Pos('.',Date);
 QueryDate.day := StrToInt(Copy(Date,0,p));
 Delete(Date,0,p+1);
 p := Pos('.',Date);
 QueryDate.month := StrToInt(Copy(Date,0,p));
 Delete(Date,0,p+1);
 QueryDate.year := StrToInt(Date);
 for i:= 2 to FStatCount-1 do
  if CompareStatTime(FStatList[i].Date,QueryDate) then
  begin
   Result := StatToString(FStatList[i]);
   Exit;
  end;
end;

function TStatisticDataBase.Load: boolean;
 var f: file of TStatistic;
     st: SystemTime;
begin
 Result:= false;
 if not FileExists(path) then Exit;
 Assign(f,path);
 Reset(f);
 while not EoF(f) do
 begin
  inc(FStatCount);
  SetLength(FStatList,FStatCount);
  Read(f,FStatList[FStatCount-1]);
 end;
 if FStatCount = 0 then
 begin
  FStatCount := 2;
  SetLength(FStatList,FStatCount);
  GetSystemTime(st);
  FStatList[0].Date := SysToStatTime(st);
  FStatList[0].maxUptime := 0;
  FStatList[1].Date := FStatList[0].Date;
  FStatList[1].maxOnline := 0;
 end;
 CloseFile(f);
 Result := true;
end;

procedure TStatisticDataBase.Save;
 var i: integer;
     f: file of TStatistic;
begin
 AssignFile(f,path);
 Rewrite(f);
 for i:= 0 to FStatCount-1 do
  Write(f,FStatList[i]);
 CloseFile(f);
end;

procedure TStatisticDataBase.UpdateData(uptime, online: integer);
 var ct: SystemTime;
     cTime: TStatisticTime;
     i: integer;
begin
 GetSystemTime(ct);
 cTime := SysToStatTime(ct);
 i := FStatCount-1;
 if (i < 2) or (not CompareStatTime(FStatList[i].Date,cTime)) then
 begin
  i := FStatCount;
  inc(FStatCount);
  SetLength(FStatList,FStatCount);
  ZeroMemory(@FStatList[i],SizeOf(TStatistic));
  FStatList[i].Date := cTime;
 end;
 if FStatList[i].maxUptime < uptime then FStatList[i].maxUptime := uptime;
 if FStatList[i].maxOnline < online then FStatList[i].maxOnline := online;
 if FStatList[0].maxUptime < uptime then
 begin
  FStatList[0].maxUptime := uptime;
  FStatList[0].Date := cTime;
 end;
 if FStatList[1].maxOnline < online then
 begin
  FStatList[1].maxOnline := online;
  FStatList[1].Date := cTime;
 end;
end;

end.
