unit WorldContent;

interface

uses SysUtils,Classes,IniFiles,Utils;

type
 TDamageType = (dtNil,dtDamgeMelee,dtDamgeDistantion,dtDamageMagic,dtHealMagic);
 TMagicType = (mtNil,mtHoly,mtUnholy,mtNature,mtEarth,mtWater,mtFire);
 TAttackResult = (arNormal,arMissle,arAbsorb,arParry);
 TMagicResist = record
  value: array [mtHoly..mtFire] of byte;
 end;

 TUnitCyclesThread = class(TThread)
  constructor Create;
  protected
   procedure Execute; override;
 end;
 TDestState = (dsOpen,dsClose,dsDestroy);
 TTrigState = (tsOn,tsOff);
 TUnitType = (utPlayer,utEnemy,utNPC);
 TUnitTypeSet = set of TUnitType;
 TOrderType = (otNoTarget,otXY,otUnit,otNil);
 TRawCode = string[4];
 TOrder = record
  _type: TOrderType;
  order: string;
  target: integer;
  x,y: integer;
 end;
 TSpell = record
  rawcode: TRawCode;
  lvl: byte;
 end;
 TTwoByte = 0..65535;
 TItem = record
  rawcode: TRawCode;
  counts: TTwoByte;
  used: boolean;
 end;
 TStockItem = record
  rawcode: TRawCode;
  count: integer;
  max: integer;
 end;
 TItemRemoveThread = class(TThread)
  constructor Create(itm: TObject; time: integer);
  protected
   procedure Execute; override;
  private
   FItem: TObject;
   FTime: integer;
 end;
 TItm = class
  HItem: integer;
  x,y: integer;
  locID: integer;
  counts: integer;
  _id: integer;
  constructor Create(rc: TRawCode; locId, nx,ny: integer);
  constructor Create0(own: integer; rc: TRawCode);
  destructor Destroy; override;
  private
   FRawCode: TRawCode;
   Fowner: integer;
   FRemoveThread: TItemRemoveThread;
   procedure SetOwner(owner: integer);
  public
   property owner: integer read FOwner write SetOwner;
   property RawCode: TRawCode read FRawCode;
   procedure SetOwnerD(owner: integer);
   function GetMainInfo: string;
   function GetInfo: string;
   procedure RunTimer(time: integer);
   procedure SendRemove;
 end;
 TWarAnim = record
  indexed: boolean;
  name: string [30];
  id: byte;
 end;
 TUnit = class
  HUnit: integer;
  posx,posy,posz,yaw: integer;
  spawnX,spawnY,spawnZ: integer;
  health,mana: integer;
  name: string;
  agro: byte;
  Fexp: cardinal;
  order: TOrder;
  locid: integer;
  spells: array [0..10] of TSpell;
  items:  array [0..5] of integer;
  stockitems: array [0..10] of TStockItem;
  _id: integer;
  _type: TUnitType;
  ispause: boolean;
  isinvul: boolean;
  constructor Create(rc: string; locId, x,y,z,ry,h,m: integer; utype: TUnitType);
  procedure SetExp(exp: cardinal); virtual;
  destructor Destroy; override;
  private
   _rawCode: string;
   //Stats
   function FGetMagicResist: TMagicResist;
   function FGetArmor: integer;
   //Other
   function FGetLevel: integer;
   function FReadItemRawCode(index: integer): TRawCode;
   function GetData(field: string): string;
   procedure SetData(field: string; value: string);
   procedure SetDataLoc(field: string; value: string);
   function GetDataLoc(field: string): string;
  public
   DataTable: TStringList;
   DataTableLoc: TStringList;
   Animation: TWarAnim;
   //Stats
   property MagicResist: TMagicResist read FGetMagicResist;
   property Armor: integer read FGetArmor;
   //Other
   property Level: integer read FGetLevel;
   property Data [field: string]: string read GetData write SetData;
   property DataLoc [field: string]: string read GetDataLoc write SetDataLoc;
   property exp: cardinal read Fexp write SetExp;
   property ItemRC [index: integer]: TRawCode read FReadItemRawCode;
   //Data Table
   procedure FlushData(field: string);
   //Animations
   procedure PlayAnim(id: integer); overload;
   procedure PlayAnim(name: string); overload;
   function GetAnimFunc: string;
   //Orders
   procedure SendOrderTarget(x,y: integer; orders: string);
   procedure SendOrderTargetUnit(HUnit: integer; orders: string);
   procedure SendOrder(orders: string);
   //not sorted
   function AddItem(hitem: integer): boolean;
   function DropItem(hitem: integer): boolean;
   property rawCode: string read _rawCode;
   function GetHM: string;
   function GetXY: string;
   //Ability
   function FindAbility(raw: string): integer;
   function AddAbility(raw: string): boolean;
   //Functions
   function HaveItem(hitem: integer): boolean;
   //Stock Item
   function FindStockItem(raw: string): integer;
   procedure AddStockItem(raw: string; count,max: integer);
   procedure RemoveStockItem(raw: string);
 end;
 TEffect = class
  HEffect: integer;
  locid,posx,posy: integer;
  modelname: string;
  attachpoint: string;
  target: integer; //HUNIT
  toUnit: boolean;
  _id: integer;
  constructor Create;
  destructor Destroy; override;
  public
    procedure AddEffect(model: string; lid,x,y: integer);
    procedure AddEffectTarget(model,attach: string; un: integer);
    function GetEffectFunc: string;
 end;
 TDest = class
  HDest: integer;
  posx,posy,posz,yaw: integer;
  locid: integer;
  state: TDestState;
  show: boolean;
  _id: integer;
  constructor Create(rc: string; locID, x,y,z,ry: integer);
  destructor Destroy; override;
  private
   _rawCode: string;
  public
   property rawCode: string read _rawCode;
   function GetMainInfo: string;
 end;
 TGroup = class
  constructor Create;
  destructor Destroy; override;
  public
   HGroup: integer;
   index: integer;
   owner: integer;
   Funits: array [0..4] of integer;
   function AddUnit(hunit: integer): boolean;
   function GetCount: integer;
   procedure RemUnit(hunit: integer);
   procedure SendAll(msg: string);
 end;
 TUnitGroup = class
  count: integer;
  HUGroup: integer;
  constructor Create;
  destructor Destroy; override;
  private
   FUnits: array of TUnit;
   function FGetUnit(index: integer): TUnit;
  public
   id: integer;
   procedure AddUnit(u: TUnit);
   procedure RemoveUnit(u: TUnit);
   function IsHaveUnit(u: TUnit): boolean;
   property Units[index: integer]: TUnit read FGetUnit; default;
   function GetUnitFromRange(locId,x,y: integer; radius: real; utypeset: TUnitTypeSet): boolean;
   function GetRandomUnit: TUnit; 
 end;
 TRect = class
  x,y,width,height: integer;
  locid: integer;
  HRect: integer;
  eventscript: string;
  _id: integer;
  constructor Create(locid,px,py,w,h: integer);
  destructor Destroy; override;
  private
   units: TUnitGroup;
  public
   function IsPointIn(px,py: integer): boolean;
   function IsEnter(u: TUnit): boolean;
   function IsOut(u: TUnit): boolean;
   function AddUnit(u: TUnit): boolean;
   procedure RemUnit(u: TUnit);
   procedure PickUnits(utypes: TUnitTypeSet);
 end;
 TAnswerData = record
  text: string;
  actionId: byte;
  actionData: string;
  handle: integer;
 end;
 TDialogTableData = record
  title: string;
  AnswerList: array of TAnswerData;
  AnswerCount: integer;
 end;
 TDialogTable = class
  DialogTableList: array of TDialogTableData;
  DialogTableCount: integer;
  HDialog: integer;
  _id: integer;
  _path: string;
  scriptevent: string;
  char: TUnit; //TCharA
  currentDT: integer;
  constructor Create;
  destructor Destroy; override;
  public
   function LoadDialogTable(path: string): boolean;
   procedure SetTitle(title: string);
   procedure ShowDialog(dt: integer);
   procedure onBtnClick(btnid: integer);
   procedure ChangeButtonText(tableid,btnid: integer; newText: string);
   function FindButtonOnBtnID(btnid: integer): integer;
   function ConvertString(str: string): string;
 end;
 TMap = class
    name: string;
    id: byte;
    constructor Create(name: string);
    destructor Destroy; override;
 end;
 
 function MathDamageValue(attacker,target: TUnit; dtype: TDamageType; mtype: TMagicType; var value: integer): TAttackResult;

function StrToUType(s: string): TUnitType;
function StrToUnitTypeSet(str: string): TUnitTypeSet;
function DestStateToInt(state: TDestState): integer;
function IntToDestState(state: integer): TDestState;
function ExpToLevel(exp: integer): integer;

function IsUnitHave(HUnit: integer): boolean;
function IsDestHave(HDest: integer): boolean;
function IsItemHave(HItem: integer): boolean;
function IsGroupHave(HGroup: integer): boolean;
function IsDialogHave(HDialog: integer): boolean;
function IsRectHave(HRect: integer): boolean;
function IsUGroupHave(HUGroup: integer): boolean;
function IsEffectHave(HEffect: integer): boolean;
function IsMapHave(HMap: integer): boolean;
//===
procedure Init;

var appPath: string;
    {Units Block}
    UnitList: array of TUnit;                                      
    UnitCount: integer = 0;
    {Destructable Block}
    DestList: array of TDest;
    DestCount: integer = 0;
    {Items Block}
    ItemList: array of TItm;
    ItemCount: integer = 0;
    {Groups Block}
    GroupList: array of TGroup;
    GroupCount: integer = 0;
    {Dialog Tables Block}
    DialogList: array of TDialogTable;
    DialogCount: integer = 0;
    {Rect Block}
    RectList: array of TRect;
    RectCount: integer = 0;
    {Map Block}
    MapList: array of TMap;
    MapCount: integer = 0;
    {Unit Group}
    UGroupList: array of TUnitGroup;
    UGroupCount: integer = 0;
    {Effect}
    EffectList: array of TEffect;
    EffectCount: integer = 0;

const
    c_OutRange: real = 1500;
    c_DamageRadius: real = 500;
    //c_MaxAttackRadius: real = 800;


implementation

uses CharDataBase,Main,MgScriptNatives,OtherDataBase;

var
  FreeHUnits: TIntStack;
  FreeHDests: TIntStack;
  FreeHItems: TIntStack;
  FreeHGroups: TIntStack;
  FreeHDialogs: TIntStack;
  FreeHRects: TIntStack;
  FreeHUGroups: TIntStack;
  FreeHEffects: TIntStack;
  FreeHMaps: TIntStack;


procedure Init;
begin
  FreeHMaps := TIntStack.Create;
  FreeHEffects := TIntStack.Create;
  FreeHUGroups := TIntStack.Create;
  FreeHRects := TIntStack.Create;
  FreeHDialogs := TIntStack.Create;
  FreeHGroups := TIntStack.Create;
  FreeHItems := TIntStack.Create;
  FreeHDests := TIntStack.Create;
  FreeHUnits := TIntStack.Create;
end;

function IsUnitHave(HUnit: integer): boolean;
begin
  Result := (HUnit >= 0) and (HUnit < UnitCount) and (not FreeHUnits.IsHave(HUnit));
end;

function IsDestHave(HDest: integer): boolean;
begin
  Result := (HDest >= 0) and (HDest < DestCount) and (not FreeHDests.IsHave(HDest));
end;

function IsItemHave(HItem: integer): boolean;
begin
  Result := (HItem >= 0) and (HItem < ItemCount) and (not FreeHItems.IsHave(HItem));
end;

function IsGroupHave(HGroup: integer): boolean;
begin
  Result := (HGroup >= 0) and (HGroup < GroupCount) and (not FreeHGroups.IsHave(HGroup));
end;

function IsDialogHave(HDialog: integer): boolean;
begin
  Result := (HDialog >= 0) and (HDialog < DialogCount) and (not FreeHDialogs.IsHave(HDialog));
end;

function IsRectHave(HRect: integer): boolean;
begin
  Result := (HRect >= 0) and (HRect < RectCount) and (not FreeHRects.IsHave(HRect));
end;

function IsUGroupHave(HUGroup: integer): boolean;
begin
  Result := (HUGroup >= 0) and (HUGroup < UGroupCount) and (not FreeHUGroups.IsHave(HUGroup));
end;

function IsEffectHave(HEffect: integer): boolean;
begin
  Result := (HEffect >= 0) and (HEffect < EffectCount) and (not FreeHEffects.IsHave(HEffect));
end;

function IsMapHave(HMap: integer): boolean;
begin
  Result := (HMap >= 0) and (HMap < MapCount) and (not FreeHMaps.IsHave(HMap));
end;

function StrToUnitTypeSet(str: string): TUnitTypeSet;
 var i: integer;
begin
 Result := [];
 for i:= 1 to Length(str) do
  if str[i] = 'P' then Result := Result + [utPlayer] else
  if str[i] = 'N' then Result := Result + [utNPC] else
  if str[i] = 'E' then Result := Result + [utEnemy];
end;

function MathDamageValue(attacker,target: TUnit; dtype: TDamageType; mtype: TMagicType; var value: integer): TAttackResult;
 var cL,tL: integer;
begin
  Result := arNormal;
  // 
  //   
  if (dtype in [dtDamageMagic,dtHealMagic]) and (mtype <> mtNil) then
  begin
    // 
    if dtype = dtDamageMagic then
    begin
      tL := attacker.Level;
      cL := target.Level;
      if (cL - tL >= 6) then
      begin
       Result := arAbsorb;
       Exit;
      end;
      value := Round(value*(1-target.MagicResist.value[mtype]/(150+(cL-tL)*25)));
      if value <= 0 then
      begin
       Result := arAbsorb;
       Exit;
      end;
    end;
  end;
end;

function ExpToLevel(exp: integer): integer;
 var lastExp,needExp: integer;
 const NeedHeroXP = 400; //  :  (NeedHeroXP)
       PrevFactor = 1.3; //  :    (NeedHeroXPFormulaA)
       LevelFactor = 50; //  :   (NeedHeroXPFormulaB)
       ConstantFactor = 33; //  :   (NeedHeroXPFormulaC)
begin
 Result:= 1;
 lastExp:= 0;
 needExp:= NeedHeroXP;
 while not ((lastExp <= exp) and (exp < needExp)) do
 begin
  inc(Result);
  lastExp:= needExp;
  needExp:= Round(lastExp*PrevFactor+LevelFactor*(Result+1)+ConstantFactor);
 end;
end;

function StrToUType(s: string): TUnitType;
begin
 if s = 'player' then Result:= utPlayer else
 if s = 'enemy' then Result:= utEnemy else
 if s = 'npc' then Result:= utnpc else
  Result := utNPC;
end;

{ TUnit }

constructor TUnit.Create(rc: string; locId, x,y,z, ry, h, m: integer; utype: TUnitType);
 var i: integer;
begin
  inherited Create;
  if not FreeHUnits.Pop(HUnit) then
  begin
    HUnit := UnitCount;
    inc(UnitCount);
    SetLength(UnitList, UnitCount);
  end;
  UnitList[HUnit] := Self;
  DataTable:= TStringList.Create;
  DataTableLoc:= TStringList.Create;
  _RawCode:= rc;
  agro:= 10;
  self.locid := locid;
  posx:= x;
  posy:= y;
  posz:= z;
  spawnX := x;
  spawnY := y;
  spawnZ := z;
  yaw:= ry;
  health:= h;
  mana:= m;
  name:= UnDataBase[rc].name;
  order._type:= otNil;
  order.target:= -1;
  for i:= 0 to 5 do
    Items[i]:= -1;
  _type:= utype;
end;

function TUnit.GetHM: string;
begin
  Result:= IntToStr(HUnit)+','+IntToStr(Health)+','+IntToStr(mana);
end;

function TUnit.GetXY: string;
begin
  Result:= IntToStr(HUnit)+','+IntToStr(posX)+','+IntToStr(posY)+','+IntToStr(Yaw);
end;

destructor TUnit.Destroy;
begin
  inherited Destroy;
  if HUnit = UnitCount - 1 then
    dec(UnitCount)
  else
    FreeHUnits.Push(HUnit);
  UnitList[HUnit] := nil;
end;

function TUnit.FindAbility(raw: string): integer;
  var i: integer;
begin
  for i:= 0 to 10 do
    if spells[i].rawcode = raw then
    begin
      Result:= i;
      exit;
    end;
  Result:= -1;
end;

function TUnit.AddAbility(raw: string): boolean;
  var i: integer;
begin
  Result:= true;
  for i:= 0 to 10 do
  begin
    if spells[i].rawcode = raw then
    begin
      spells[i].lvl:= 1;
      exit;
    end;
    if spells[i].rawcode = '' then
    begin
      spells[i].rawcode:= raw;
      spells[i].lvl:= 1;
      exit;
    end;
  end;
  Result:= false;
end;

procedure TUnit.AddStockItem(raw: string; count,max: integer);
  var i: integer;
begin
  for i:= 0 to 10 do
    if stockitems[i].rawcode = '' then
    begin
      stockitems[i].rawcode:= raw;
      stockitems[i].count:= count;
      stockitems[i].max:= max;
      exit;
    end;
end;

function TUnit.FindStockItem(raw: string): integer;
  var i: integer;
begin
  Result:= -1;
  for i:= 0 to 10 do
    if stockitems[i].rawcode = raw then
    begin
      Result:= i;
      exit;
    end;
end;

procedure TUnit.RemoveStockItem(raw: string);
 var i: integer;
begin
  i:= FindStockItem(raw);
  if (i < 0) or (i > 10) then exit;
  stockitems[i].rawcode:= '';
  stockitems[i].count:= 0;
  stockitems[i].max:= 0;
end;

function TUnit.AddItem(hitem: integer): boolean;
  var i: integer;
begin
  Result:= true;
  for i:= 0 to 5 do
    if not IsItemHave(Items[i]) then
    begin
      Items[i]:= hitem;
      exit;
    end;
  Result:= false;
end;

function TUnit.DropItem(hitem: integer): boolean;
  var i,j: integer;
begin
  Result:= true;
  for i:= 0 to 5 do
    if Items[i] = hitem then
    begin
      WorldContent.ItemList[j].RunTimer(60);
      Items[i]:= -1;
      exit;
    end;
 Result:= false;
end;

function TUnit.HaveItem(hitem: integer): boolean;
 var i: integer;
begin
  Result:= true;
  for i:= 0 to 5 do
    if Items[i] = hitem then exit;
  Result:= false;
end;

function TUnit.FReadItemRawCode(index: integer): TRawCode;
begin
 Result:= 'nil';
 if IsItemHave(items[index]) then
 begin
  Result:= WorldContent.ItemList[items[index]].RawCode; 
 end;
end;

procedure TUnit.SetExp(exp: cardinal);
begin
 FExp:= exp;
end;

function TUnit.GetData(field: string): string;
begin
 Result:= DataTable.Values[field];
end;

procedure TUnit.SetData(field, value: string);
begin
 if DataTable.IndexOfName(field) < 0 then
  DataTable.Add(field+'='+value)
 else
  DataTable.Values[field]:= value;
end;

procedure TUnit.SendOrderTarget(x, y: integer; orders: string);
begin
  order._type:= otXY;
  order.target:= -1;
  order.order:= orderS;
  order.x:= x;
  order.y:= y;
  Main.SendAllPlayersOnLoc('orderxy('+IntToStr(HUNIT)+','+orders+','+IntToStr(x)+','+IntToStr(y)+')',locid,nil);
end;

procedure TUnit.SendOrderTargetUnit(HUnit: integer; orders: string);
begin
  order._type:= otUnit;
  order.order:= orderS;
  order.target:= HUnit;
  Main.SendAllPlayersOnLoc('orderunit('+IntToStr(Self.HUnit)+','+orderS+','+IntToStr(HUNIT)+')',locid,nil);
end;

procedure TUnit.SendOrder(orders: string);
begin
  order._type:= otNoTarget;
  order.order:= orderS;
  order.target:= -1;
  Main.SendAllPlayersOnLoc('order('+IntToStr(HUNIT)+','+orderS+')',locid,nil);
end;

function TUnit.GetAnimFunc: string;
begin
  if Animation.indexed then
    Result:= 'setanimind('+IntToStr(HUNIT)+','+IntToStr(Animation.id)+')'
  else
    Result:= 'setanim('+IntToStr(HUNIT)+',"'+Animation.name+'")'
end;

procedure TUnit.PlayAnim(id: integer);
begin
  Animation.indexed:= true;
  Animation.id:= id;
  Main.SendAllPlayersOnLoc('setanimind('+IntToStr(HUNIT)+','+IntToStr(Animation.id)+')',locid,nil);
end;

procedure TUnit.PlayAnim(name: string);
begin
  Animation.indexed:= false;
  Animation.name:= name;
  Main.SendAllPlayersOnLoc('setanim('+IntToStr(HUNIT)+',"'+Animation.name+'")',locid,nil);
end;

procedure TUnit.FlushData(field: string);
 var i: integer;
begin
  i:= DataTable.IndexOfName(field);
  if i >= 0 then
    DataTable.Delete(i)
  else begin
    i := DataTableLoc.IndexOfName(field);
    if i < 0 then Exit;
    DataTableLoc.Delete(i);
  end;
end;

function TUnit.FGetArmor: integer;
 var i,j: integer;
     itm: TItm;
begin
  Result := 0;
  for i:= 0 to 5 do
  begin
    if IsItemHave(items[i]) then
    begin
      itm := WorldContent.ItemList[items[i]];
      Result := Result + ItmDataBase.ItemsRaw[itm.RawCode].Armor; 
    end;
  end;
end;

function TUnit.FGetMagicResist: TMagicResist;
  var i,j: integer;
      itm: TItm;
      k: TMagicType;
begin
  FillChar(Result,SizeOf(TMagicresist),0);
  for i := 0 to 5 do
  begin
    if IsItemHave(items[i]) then
    begin
      itm := WorldContent.ItemList[items[i]];
      for k := mtHoly to mtFire do
        Result.value[k] := Result.value[k] + ItmDataBase.ItemsRaw[itm.RawCode].MagicResist.value[k];
    end;
  end;
end;

function TUnit.FGetLevel: integer;
begin
 Result := ExpToLevel(exp);
end;

function TUnit.GetDataLoc(field: string): string;
begin
 Result:= DataTableLoc.Values[field]; 
end;

procedure TUnit.SetDataLoc(field, value: string);
begin
 if DataTableLoc.IndexOfName(field) < 0 then
  DataTableLoc.Add(field+'='+value)
 else
  DataTableLoc.Values[field]:= value;
end;

{ TDest }

function DestStateToInt(state: TDestState): integer;
begin
 if state = dsOpen then Result:= 0 else
 if state = dsClose then Result:= 1 else
 if state = dsDestroy then Result:= 2 else
  Result := 0;
end;

function IntToDestState(state: integer): TDestState;
begin
 Result := dsOpen;
   case state of
     0: Result:= dsOpen;
     1: Result:= dsClose;
     2: Result:= dsDestroy;
   end;
end;

constructor TDest.Create(rc: string; locID, x, y, z, ry: integer);
begin
  inherited Create;
  if not FreeHDests.Pop(HDest) then
  begin
    HDest := DestCount;
    inc(DestCount);
    SetLength(DestList,DestCount);
  end;
  DestList[HDest] := Self;
  Self.locid := locid;
  posx:= x;
  posy:= y;
  posz:= z;
  yaw:= ry;
  state:= dsClose;
  _rawCode:= rc;
end;

destructor TDest.Destroy;
begin
  if HDest >= DestCount - 1 then
    dec(DestCount)
  else
    FreeHDests.Push(HDest);
  DestList[HDest] := nil;
  inherited Destroy;
end;

function TDest.GetMainInfo: string;
begin
  Result:= IntToStr(HDest)+','+_rawCode+','+IntToStr(posx)+','+IntToStr(posy)+','+IntToStr(posz)+','+IntToStr(yaw)+',1,1';
end;

{ TUnitGroup }

procedure TUnitGroup.AddUnit(u: TUnit);
 var index: integer;
begin
 index:= count;
 inc(count);
 SetLength(FUnits,count);
 FUnits[index]:= u;
end;

constructor TUnitGroup.Create;
begin
  inherited Create;
  if not FreeHUGroups.Pop(HUGroup) then
  begin
    HUGroup := UGroupCount;
    inc(UGroupCount);
    SetLength(UGroupList, UGroupCount);
  end;
  UGRoupList[HUGroup] := Self;
  count:= 0;
end;

destructor TUnitGroup.Destroy;
begin
  if HUGroup = UGroupCount - 1 then
    dec(UGroupCount)
  else
    FreeHUGroups.Push(HUGroup);
  UGroupList[HUGroup] := nil;
  inherited Destroy;
end;

function TUnitGroup.FGetUnit(index: integer): TUnit;
begin
 Result:= FUnits[index];
end;

function GetUnitPointDist(u: TUnit; x,y: integer): real;
 var dx,dy: integer;
begin
 dx:= u.posx-x;
 dy:= u.posy-y;
 Result:= SqRt(dx*dx+dy*dy);
end;

function TUnitGroup.GetRandomUnit: TUnit;
begin
 Result:= FUnits[Random(count)];
end;

function TUnitGroup.GetUnitFromRange(locId,x, y: integer; radius: real; utypeset: TUnitTypeSet): boolean;
 var i: integer;
     d: real;
begin
  for i:= 0 to UnitCount-1 do
    if UnitList[i] <> nil then
      if WorldContent.UnitList[i].locid = locID then
        if WorldContent.UnitList[i]._type in utypeset then
        begin
          if WorldContent.UnitList[i] is TCharA then
            if not (WorldContent.UnitList[i] as TCharA).status then continue;
          d:= GetUnitPointDist(WorldContent.UnitList[i],x,y);
          if d < radius then
            AddUnit(WorldContent.UnitList[i]);
        end;
end;

function TUnitGroup.IsHaveUnit(u: TUnit): boolean;
 var i: integer;
begin
  Result:= true;
  for i:= 0 to count-1 do
    if FUnits[i] = u then exit;
  Result:= false;
end;

procedure TUnitGroup.RemoveUnit(u: TUnit);
 var index: integer;
begin
  for index:= 0 to count-1 do
    if FUnits[index] = u then break;
  dec(count);
  FUnits[index]:= FUnits[count];
  FUnits[count]:= nil;
end;

{ TRect }

function TRect.AddUnit(u: TUnit): boolean;
begin
  Result := u.locid = locid;
  if not Result then
    exit;
  Result:= IsPointIn(u.posx,u.posy);
  if not Result then
    exit;
  Result:= not Units.IsHaveUnit(u);
  if Result then
    Units.AddUnit(u);
end;

constructor TRect.Create(locid, px, py, w, h: integer);
begin
  if not FreeHRects.Pop(HRect) then
  begin
    HRect := RectCount;
    inc(RectCount);
    SetLength(RectList, RectCount);
  end;
  RectList[HRect] := Self;
  units:= TUnitGroup.Create;
  Self.locid := locid;
  x:= px;
  y:= py;
  width:= w;
  height:= h;
end;

destructor TRect.Destroy;
begin
  if HRect = RectCount - 1 then
    dec(RectCount)
  else
    FreeHRects.Push(HRect);
  RectList[HRect] := nil;
  units.Destroy;
  inherited Destroy;
end;

function TRect.IsEnter(u: TUnit): boolean;
 var dx,dy: integer;
begin
 Result:= false;
 if u.locid <> locid then Exit;
 if not Units.IsHaveUnit(u) then
 begin
  Result := ((u.posx > x) and (u.posx < x+width)) and ((u.posy < y) and (u.posy > y-height))
 end;
end;

function TRect.IsOut(u: TUnit): boolean;
 var dx,dy: integer;
begin
 Result:= false;
 if Units.IsHaveUnit(u) then
 begin
  Result := (u.locid <> locid) or (not (((u.posx > x) and (u.posx < x+width)) and ((u.posy < y) and (u.posy > y-height))))
 end;
end;

function TRect.IsPointIn(px, py: integer): boolean;
 var dx,dy: integer;
begin
 dx:= px-x;
 dy:= py-y;
 Result:= ((px > x) and (px < x+width)) and ((py < y) and (py > y-height)); 
end;

procedure TRect.PickUnits(utypes: TUnitTypeSet);
 var i: integer;
begin
 for i:= 0 to UnitCount-1 do
  if UnitList[i]._type in utypes then
    AddUnit(UnitList[i]);
end;

procedure TRect.RemUnit(u: TUnit);
begin
 Units.RemoveUnit(u);
end;

{ TItm }

constructor TItm.Create(rc: TRawCode; locID, nx, ny: integer);
begin
  inherited Create;
  if not FreeHItems.Pop(HItem) then
  begin
    HItem := ItemCount;
    inc(ItemCount);
    SetLength(ItemList, ItemCount);
  end;
  ItemList[HItem] := Self;
  FRawCode:= rc;
  Fowner:= -1;
  x:= nx;
  y:= ny;
  FRemoveThread := nil;
end;

constructor TItm.Create0(own: integer; rc: TRawCode);
begin
  inherited Create;
  if not FreeHItems.Pop(HItem) then
  begin
    HItem := ItemCount;
    inc(ItemCount);
    SetLength(ItemList, ItemCount);
  end;
  ItemList[HItem] := Self;
  FRawCode:= rc;
  Fowner:= own;
  FRemoveThread := nil;
end;

destructor TItm.Destroy;
  var
    i: integer;
begin
  if HItem = ItemCount - 1 then
    dec(ItemCount)
  else
    FreeHItems.Push(HItem);
  ItemList[HItem] := nil;

  for i := 0 to 5 do
    if UnitList[FOwner].items[i] = HItem then
    begin
      UnitList[FOwner].items[i] := -1;
    end;
  if Assigned(FRemoveThread) then
  begin
    FRemoveThread.Terminate;
    FRemoveThread := nil;
  end;
  inherited;
end;

function TItm.GetInfo: string;
begin
 Result:= IntToStr(HItem)+','+IntToStr(owner);
end;

function TItm.GetMainInfo: string;
begin
 Result:= IntToStr(x)+','+IntToStr(y)+','+rawCode+','+IntToStr(HItem);
end;

procedure TItm.RunTimer(time: integer);
begin
 FRemoveThread:= TItemRemoveThread.Create(Self,time);
end;

procedure TItm.SendRemove;
 var i: integer;
begin
 if Owner >= 0 then
 begin
  i := Owner;
  if i >= 0 then
   UnitList[i].DropItem(HItem);
 end;
 SendAllPlayersOnLoc('remitm('+IntToStr(HITEM)+')',locId,nil);
end;

procedure TItm.SetOwner(owner: integer);
 var i: integer;
begin
  if not IsUnitHave(i) then
  begin
    FOwner := -1;
    Exit;
  end;
  if not UnitList[i].HaveItem(Hitem) then
    if UnitList[i].AddItem(Hitem) then
    begin
      FOwner:= owner;
      if Assigned( FRemoveThread ) then
      begin
        FRemoveThread.Terminate;
        FRemoveThread := nil;
      end;
    end;
end;

procedure TItm.SetOwnerD(owner: integer);
begin
 FOwner:= owner;
end;

{ TGroup }

function TGroup.AddUnit(hunit: integer): boolean;
 var i: integer;
begin
 Result:= true;
 for i:= 0 to 4 do
  if FUnits[i] < 0 then
  begin
   FUnits[i]:= hunit;
   exit;
  end;
 Result:= false;
end;

constructor TGroup.Create;
 var i: integer;
begin
  inherited Create;
  if not FreeHGroups.Pop(HGroup) then
  begin
    HGroup := GroupCount;
    inc(GroupCount);
    SetLength(GroupList, GroupCount);
  end;
  GroupList[HGroup] := Self;
  for i:= 0 to 4 do
    FUnits[i]:= -1;
end;

destructor TGroup.Destroy;
begin
  if HGroup = GroupCount - 1 then
    dec(GroupCount)
  else
    FreeHGroups.Push(HGroup);
  GroupList[HGroup] := nil;
  inherited Destroy;
end;

function TGroup.GetCount: integer;
 var i: integer;
begin
 Result:= 0;
 for i:= 0 to 4 do
  if FUnits[i] >= 0 then inc(Result);
end;

procedure TGroup.RemUnit(hunit: integer);
 var i: integer;
begin
 for i:= 0 to 4 do
  if FUnits[i] = hunit then
  begin
   FUnits[i]:= -1;
   exit;
  end;
end;

procedure TGroup.SendAll(msg: string);
 var i,j: integer;
begin
  for i:= 0 to 4 do
  begin
    if IsUnitHave(FUnits[i]) then
      (UnitList[FUnits[i]] as TCharA).client.Send(msg);
 end;
end;

{ TUnitCyclesThread }

constructor TUnitCyclesThread.Create;
begin
 inherited Create(true);
 FreeOnTerminate:= true;
 Resume;
end;

function GetUnitsDist(u1,u2: TUnit): real;
 var dx,dy: real;
begin
 dx:= (u1.posx-u2.posx);
 dy:= (u1.posy-u2.posy);
 Result:= Sqrt(dx*dx+dy*dy);
end;

procedure TUnitCyclesThread.Execute;
 var i,j: integer;
     d: real;
     grp: TUnitGroup;
     un: TUnit;
     rct: TRect;
begin
  for i:= 0 to UnitCount-1 do
  begin
    if UnitList[i] = nil then
      Continue;
  //=====================================
  //For Rects
  //  Check rect events (onEnter, onLeave)
  //=====================================
    for j := 0 to RectCount-1 do
    begin
      if RectList[j] = nil then
        continue;
      rct := RectList[j];
      if rct.IsEnter(UnitList[i]) then
      begin
        rct.units.AddUnit(UnitList[i]);
        MGScriptNatives.RunMGScript(rct.eventscript,'onEnter',IntToStr(rct.HRect)+','+IntToStr(UnitList[i].HUnit));
      end else
        if rct.IsOut(UnitList[i]) then
        begin
          rct.units.RemoveUnit(UnitList[i]);
          MGScriptNatives.RunMGScript(rct.eventscript,'onLeave',IntToStr(rct.HRect)+','+IntToStr(UnitList[i].HUnit));
        end;
    end;
  //=====================================
  //For Enemy
  //=====================================
    if (UnitList[i]._type = utEnemy) then
      if IsUnitHave(UnitList[i].order.target) then
      begin
        j := UnitList[i].order.target;
        d:= GetUnitsDist(UnitList[i],UnitList[j]);
        if (d > c_OutRange) or (UnitList[j].health <= 0) then
        begin
          grp:= TUnitGroup.Create;
          grp.GetUnitFromRange(UnitList[i].locid,UnitList[i].posx,UnitList[i].posy,c_damageradius,[utPlayer]);
          if grp.count > 0 then
          begin
            un:= grp.GetRandomUnit;
            UnitList[i].SendOrdertargetUnit(un.HUnit,'attack');
          end else begin
            UnitList[i].health:= UnDataBase[UnitList[i].RawCode].maxhealth;
            Main.SendAllPlayers('newHM('+UnitList[i].GetHM+')',nil);
            UnitList[i].SendOrderTarget(UnitList[i].spawnX,UnitList[i].spawnY,'move');
          end;
          grp.Free;
        end;
      end else begin
      //Radius Agro
        grp:= TUnitGroup.Create;
        grp.GetUnitFromRange(UnitList[i].locid,UnitList[i].posx,UnitList[i].posy,c_damageradius,[utPlayer]);
        if grp.count > 0 then
        begin
          un:= grp.GetRandomUnit;
          UnitList[i].SendOrderTargetUnit(un.HUnit,'attack');
        end;
        grp.Free;
      end;
  end;
end;

{ TItemRemoveThread }

constructor TItemRemoveThread.Create(itm: TObject; time: integer);
begin
 inherited Create(true);
 FreeOnTerminate:= true;
 FItem:= itm;
 FTime:= time;
 Resume;
end;

procedure TItemRemoveThread.Execute;
begin
 Sleep(FTime*1000);
 if (FItem as TItm).owner >= 0 then exit;
 (FItem as TItm).SendRemove;
 (FItem as TItm).Destroy;
end;

{ TDialogTable }

procedure TDialogTable.ChangeButtonText(tableid,btnid: integer; newText: string);
 var i: integer;
begin
 for i:= 0 to DialogTableList[tableid].AnswerCount-1 do
  if DialogTableList[tableid].AnswerList[i].handle = btnid then
  begin
   DialogTableList[tableid].AnswerList[i].text := newText;
   exit;
  end;
end;

function TDialogTable.ConvertString(str: string): string;
begin
 Result := StringReplace(str,'%name%',char.name,[rfReplaceAll]);
 Result := StringReplace(Result,'%class%',ChClassDataBase.classs[(char as TCharA).classid].name,[rfReplaceAll]);
end;

constructor TDialogTable.Create;
begin
  inherited Create;
  if not FreeHDialogs.Pop(HDialog) then
  begin
    HDialog := DialogCount;
    inc(DialogCount);
    SetLength(DialogList,DialogCount);
  end;
  DialogList[HDialog] := Self;
end;

destructor TDialogTable.Destroy;
begin
  if HDialog = DialogCount - 1 then
    dec(DialogCount)
  else
    FreeHDialogs.Push(HDialog);
  DialogList[HDialog] := nil;
 inherited Destroy;
end;

function TDialogTable.FindButtonOnBtnID(btnid: integer): integer;
 var i: integer;
begin
 for i:= 0 to DialogTableList[currentDT].AnswerCount-1 do
  if DialogTableList[currentDT].AnswerList[i].handle = btnid then
  begin
   Result:= i;
   exit;
  end;
 Result:= 0;
end;

function TDialogTable.LoadDialogTable(path: string): boolean;
 var dt: TIniFile;
     i,j: integer;
     buf,section: string;
begin
 Result:= false;
 path:= Main.pathlist.dialogtable+path;
 if not FileExists(path) then exit;
 dt:= TIniFile.Create(apppath+path);
 DialogTableCount:= dt.ReadInteger('main','count',0);
 SetLength(DialogTableList,DialogTableCount);
 for i:= 1 to DialogTableCount do
  with DialogTableList[i-1] do
  begin
   section:= 'table_'+IntToStr(i);
   title:= dt.ReadString(section,'title','<no title>');
   AnswerCount:= dt.ReadInteger(section,'count',0);
   SetLength(AnswerList,AnswerCount);
   for j:= 1 to DialogTableList[i-1].AnswerCount do
   begin
    buf:= 'answer_'+IntToStr(j)+'_';
    AnswerList[j-1].text:= dt.ReadString(section,buf+'text','<no text>');
    AnswerList[j-1].actionId:= dt.ReadInteger(section,buf+'action',0);
    AnswerList[j-1].actionData:= dt.ReadString(section,buf+'actiondata','');
    AnswerList[j-1].handle:= dt.ReadInteger(section,buf+'index',0);
   end;
  end;
 _path:= path;
 dt.Free;
 Result:= true;
end;

procedure TDialogTable.onBtnClick(btnid: integer);
 var btn: integer;
begin
 btn:= FindButtonOnBtnId(btnid);
 if btn < 0 then exit;
 with DialogTableList[currentDT].AnswerList[btn] do
 begin
  case actionId of
   0: Destroy;
   1: ShowDialog(StrToInt(actionData)-1);
   2: RunMGScript(pathlist.script+actionData,'onClick',IntToStr(char.HUnit)+','+IntToStr(HDialog)+','+IntToStr(handle));
   3: begin
        (Char as TCharA).client.Send('d_t('+IntToStr(HDialog)+',"'+actionData+'")');
        (Char as TCharA).client.Send('d_s('+IntToStr(HDialog)+')');
      end;
  end;
 end;
end;

procedure TDialogTable.SetTitle(title: string);
begin
 (char as TCharA).client.Send('d_t('+IntToStr(HDialog)+',"'+ConvertString(title)+'")');
end;

procedure TDialogTable.ShowDialog(dt: integer);
 var handle: string;
     i: integer;
begin
 currentDT:= dt;
 with (char as TCharA).client do
 begin
  handle:= IntToStr(HDialog);
  Send('d_c('+handle+')');
  Send('d_t('+handle+',"'+ConvertString(DialogTableList[dt].title)+'")');
  for i:= 0 to DialogTableList[dt].AnswerCount-1 do
   Send('d_a('+handle+','+IntToStr(DialogTableList[dt].AnswerList[i].handle)+',"'+ConvertString(DialogTableList[dt].AnswerList[i].text)+'")');
  Send('d_s('+handle+')');
 end;
end;

{ TMap }

function FindFreeMapID: integer;
 var i: integer;
begin
 for i:= 0 to MapCount-1 do
  if MapList[i] = nil then
  begin
    Result := i;
    Exit;
  end;
 Result := MapCount;
 inc(MapCount);
 SetLength(MapList,MapCount);
end;

constructor TMap.Create(name: string);
begin
 inherited Create;
end;

destructor TMap.Destroy;
begin
 inherited Destroy;
end;

{ TEffect }

procedure TEffect.AddEffect(model: string; lid, x, y: integer);
begin
  toUnit := false;
  modelname := model;
  locid := lid;
  posx := x;
  posy := y;
end;

procedure TEffect.AddEffectTarget(model, attach: string; un: integer);
begin
  toUnit := true;
  modelname := model;
  attachpoint := attach;
  target := un;
end;

constructor TEffect.Create;
begin
  inherited Create;
  if not FreeHEffects.Pop(HEffect) then
  begin
    HEffect := EffectCount;
    inc(EffectCount);
    SetLength(EffectList, EffectCount);
  end;
  EffectList[HEffect] := Self;
end;

destructor TEffect.Destroy;
begin
  if HEffect = EffectCount - 1 then
    dec(EffectCount)
  else
    FreeHEffects.Push(HEffect);
  EffectList[HEffect] := nil;
  inherited Destroy;
end;

function TEffect.GetEffectFunc: string;
begin
  if toUnit then
    Result := 'enewu("'+modelname+'","'+attachpoint+'",'+IntToStr(target)+','+IntToStr(HEffect)+')'
  else
    Result := 'enewxy("'+modelname+'",'+IntToStr(posX)+','+IntToStr(posy)+','+IntToStr(HEffect)+')';
end;

end.
