library MMOClient;

uses
  SysUtils,
  ScktComp,
  Dialogs,
  Windows,
  jass_types in 'jass_types.pas',
  nAPI in 'nAPI.pas',
  mgscript in '..\Source - actual\MGScript.pas';

type
 TWar3Sckt = class
    Client: TClientSocket;
    constructor Create;
    destructor Destroy; override;
    public
      lastRecMsg: string;
      lastError: integer;
    private
      msgBuffer: string;
      procedure onSocketRead(Sender: TObject; Socket: TCustomWinSocket);
      procedure onSocketError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
      procedure onSocketConnect(Sender: TObject; Socket: TCustomWinSocket);
      procedure onSocketDisconnect(Sender: TObject; Socket: TCustomWinSocket);
      procedure MessageProc(msg: string);
  end;
 TServerUnit = class
  _unit: jass_unit;
  _id,HUNIT: integer;
  //DataTable: TStringList;
  constructor Create(u: jass_unit; hu: integer);
  destructor Destroy;  override;
 end;
 TServerDest = class
  _dest: jass_destructable;
  _id,HDEST: integer;
  constructor Create(d: jass_destructable; hd: integer);
  destructor Destroy;  override;
 end;
 TServerItem = class
  _itm: jass_item;
  _id,HITEM: integer;
  constructor Create(i: jass_item; hi: integer);
  destructor Destroy;  override;
 end;
 TServerEffect = class
  _effect: jass_effect;
  _id,HEFFECT: integer;
  constructor Create(i: jass_effect; he: integer);
  destructor Destroy; override;
 end;

var
 War3Client: TWar3Sckt = nil;
 SocketEvent: array of TTriggerGroup;
 SocketEventCount: integer = 0;
 //==========
 ServerUnits: array of TServerUnit;
 ServerUnitCount: integer = 0;

 ServerDests: array of TServerDest;
 ServerDestCount: integer = 0;

 ServerItems: array of TServerItem;
 ServerItemCount: integer = 0;

 ServerEffects: array of TServerEffect;
 ServerEffectCount: integer = 0;

const
 NewPack = #10#12;
 sce_Connect = 0;
 sce_Disconnect = 1;
 sce_RecMsg = 2;
 sce_Error = 3;
 liFalse = LongInt(false);
 liTrue = LongInt(true);

var
  isFirstRun : boolean = true;

{$R *.res}

procedure InitLog;
  var f: text;
begin
  Assign(f,'Forgotten Land Client\logs\MMOClientLog.txt');
  Rewrite(f);
  WriteLn(f,'Log -> Init complete');
  CloseFile(f);
end;

procedure AddToLog(msg: string);
  var f: text;
begin
  Assign(f,'Forgotten Land Client\logs\MMOClientLog.txt');
  Append(f);
  WriteLn(f,msg);
  CloseFile(f);
end;

procedure ResetAllData; cdecl;
 var i: integer;
begin
  try
    while ServerUnitCount > 0 do
    begin
      if ServerUnits[0] <> nil then
        ServerUnits[0].Destroy;
      ServerUnits[0] := nil;
    end;

    while ServerDestCount > 0 do
    begin
      if ServerDests[0] <> nil then
        ServerDests[0].Destroy;
      ServerDests[0] := nil;
    end;

    while ServerItemCount > 0 do
    begin
      if ServerItems[0] <> nil then
        ServerItems[0].Destroy;
      ServerItems[0] := nil;
    end;

    while ServerEffectCount > 0 do
    begin
      if ServerEffects[0] <> nil then
        ServerEffects[0].Destroy;
      ServerEffects[0] := nil;
    end;

    ServerUnitCount := 0;
    ServerItemCount := 0;
    ServerDestCount := 0;
    ServerEffectCount := 0;
    if War3Client <> nil then
      War3Client.Destroy;
    War3Client := nil;
    for i := 0 to SocketEventCount-1 do
    begin
      if SocketEvent[i] <> nil then
        DestroyTriggerGroup(SocketEvent[i]);
      SocketEvent[i] := nil;
    end;
    
  except
    on E: Exception do AddToLog('Core -> Cannot reset data ' + #13#10 + 'Class: '+E.ClassName + #13#10 + 'Message: ' + E.Message);
  end;
  
end;

function BoolToJB(b: boolean): jass_boolean;
begin
 Result:= liFalse;
 if b then
  Result:= liTrue;
end;

function ReadIP: LongInt; cdecl;
 var f: text;
     s: string;
 const
    path: string = 'Forgotten Land Client\realmlist.wtf';
begin
 AssignFile(f,path);
 Reset(f);
 ReadLn(f,s);
 CloseFile(f);
 AddToLog('Core -> IP readed: '+s);
 Result:= StringMap(PChar(s));
end;

function JDiv(a,b: jass_integer): jass_integer;
begin
 Result:= a div b;
end;

function JMod(a,b: jass_integer): jass_integer;
begin
 Result:= a mod b;
end;

{ TWar3Sckt }

constructor TWar3Sckt.Create;
begin
 inherited Create;
 try
   Client:= TClientSocket.Create(nil);
   Client.OnConnect:= onSocketConnect;
   Client.OnDisconnect:= onSocketDisconnect;
   Client.OnRead:= onSocketRead;
   Client.OnError:= onSocketError;
   AddToLog('Web socket created');
 except
   on E: Exception do AddToLog('Socket -> Can not create' + #13#10 + 'Class: ' + E.ClassName + #13#10 + 'Message: '+E.Message);
 end;
end;

destructor TWar3Sckt.Destroy;
begin
  try
   if Client <> nil then
    Client.Destroy;
   Client := nil;
   inherited Destroy;
  except
    on E: Exception do AddToLog('Socket -> can not destroyed' + #13#10 + 'Class: ' + E.ClassName + #13#10 + 'Message: '+E.Message);
  end;
end;

procedure TWar3Sckt.onSocketConnect(Sender: TObject; Socket: TCustomWinSocket);
begin
  AddToLog('Socket -> Connect');
  if SocketEvent[sce_connect] <> nil then
   TriggerGroupRun(SocketEvent[sce_connect]);
end;

procedure TWar3Sckt.onSocketDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
  AddToLog('Socket -> Disconnect');
  if SocketEvent[sce_disconnect] <> nil then
   TriggerGroupRun(SocketEvent[sce_disconnect]);
end;

procedure TWar3Sckt.onSocketError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
  AddToLog('Socket -> Error: '+IntToStr(ErrorCode));
 lastError:= ErrorCode;
 ErrorCode:= 0;
 if SocketEvent[sce_error] <> nil then
   TriggerGroupRun(SocketEvent[sce_error]);
end;

procedure TWar3Sckt.MessageProc(msg: string);
begin
 lastRecMsg:= msg;
 if SocketEvent[sce_recmsg] <> nil then
   TriggerGroupRun(SocketEvent[sce_recmsg]);
end;

procedure TWar3Sckt.onSocketRead(Sender: TObject; Socket: TCustomWinSocket);
 var i: integer;
     buf_msg,msg: string;
begin
 buf_msg:= msgBuffer + Socket.ReceiveText;
 AddToLog('Msg -> read: '+buf_msg);
 repeat
  i:= Pos(NewPack,buf_msg);
  if i <> 0 then
   begin
    msg:= Copy(buf_msg,1,i-1);
    Delete(buf_msg,1,i+Length(NewPack)-1);
    AddToLog('Msg -> proc: '+msg);
    MessageProc(msg);
  end;
 until i = 0;
 msgBuffer := buf_msg;
 AddToLog('Msg -> lost: '+msgBuffer);
end;

{ TServerItem }

constructor TServerItem.Create(i: jass_item; hi: integer);
begin
 inherited Create;
 _itm:= i;
 HITEM:= hi;
 _id:= ServerItemCount;
 inc(ServerItemCount);
 SetLength(ServerItems,ServerItemCount);
 ServerItems[_id]:= Self;
end;

destructor TServerItem.Destroy;
begin
 inherited Destroy;
 dec(ServerUnitCount);
 ServerUnits[_id]:= ServerUnits[ServerUnitCount];
 ServerUnits[_id]._id:= _id;
end;

{ TServerUnit }

constructor TServerUnit.Create(u: jass_unit; hu: integer);
begin
 inherited Create;
 _unit:= u;
 HUNIT:= hu;
 _id:= ServerUnitCount;
 inc(ServerUnitCount);
 SetLength(ServerUnits,ServerUnitCount);
 ServerUnits[_id]:= Self;
end;

destructor TServerUnit.Destroy;
begin
 inherited Destroy;
 dec(ServerUnitCount);
 ServerUnits[_id]:= ServerUnits[ServerUnitCount];
 ServerUnits[_id]._id:= _id;
end;

function FindUnitByHUNIT(hunit: integer): integer;
 var i: integer;
begin
 for i:= 0 to ServerUnitCount-1 do
  if ServerUnits[i].HUNIT = hunit then
  begin
   Result:= i;
   exit;
 end;
 Result:= -1;
end;

function FindUnitByUnit(un: jass_unit): integer;
 var i: integer;
begin
 for i:= 0 to ServerUnitCount-1 do
  if ServerUnits[i]._unit.handleID = un.handleID then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

{ TServerEffect }

constructor TServerEffect.Create(i: jass_effect; he: integer);
begin
 inherited Create;
 HEFFECT := he;
 _effect := i;
 _id := ServerEffectCount;
 inc(ServerEffectCount);
 SetLength(ServerEffects,ServerEffectCount);
 ServerEffects[_id] := Self;
end;

destructor TServerEffect.Destroy;
begin
  dec(ServereffectCount);
  ServerEffects[_id] := ServerEffects[ServerEffectCount];
  ServerEffects[_id]._id := _id;
  inherited Destroy;
end;

function FindEffectByHEFFECT(he: integer): integer;
 var i: integer;
begin
 for i:= 0 to ServerEffectCount-1 do
  if ServerEffects[i].HEFFECT = he then
  begin
    Result := i;
    Exit;
  end;
  Result := -1;
end;

{ TServerDest }

constructor TServerDest.Create(d: jass_destructable; hd: integer);
begin
 inherited Create;
 _dest:= d;
 HDEST:= hd;
 _id:= ServerDestCount;
 inc(ServerDestCount);
 SetLength(ServerDests,ServerDestCount);
 ServerDests[_id]:= Self;
end;

destructor TServerDest.Destroy;
begin
 dec(ServerDestCount);
 ServerDests[_id]:= ServerDests[ServerDestCount];
 ServerDests[_id]._id:= _id;
 inherited Destroy;
end;

function FindDestByHDEST(hdest: integer): integer;
 var i: integer;
begin
 for i:= 0 to ServerDestCount-1 do
  if ServerDests[i].HDEST = hdest then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

function FindDestByDest(dest: jass_destructable): integer;
 var i: integer;
begin
 for i:= 0 to ServerDestCount-1 do
  if ServerDests[i]._dest.handleID = dest.handleID then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

//============================\\
// Native Functions           \\
//============================\\

 {Socket Functions}
 //Get Last Receving Message
function SocketGetReciveMsg: LongInt; cdecl;
begin
  Result := StringMap(nil);
  if War3Client <> nil then
   Result:= StringMap(PChar(War3Client.lastRecMsg));
end;
 //Get Last Error Code
function SocketGetErrorCode: jass_integer; cdecl;
begin
  Result := 0;
  if War3Client <> nil then
   Result:= War3Client.lastError;
end;
 //Open Connection on Socket for IP on PORT
function SocketOpen(host: LongInt; port: jass_integer): jass_boolean; cdecl;
begin
 Result := liFalse;
 if War3Client <> nil then
 begin
   Result:= liTrue;
   War3Client.Client.Host:= StringGet(host);
   War3Client.Client.Port:= port;
   try
    War3Client.Client.Active:= true;
   except
    Result:= liFalse;
    AddToLog('Socket -> can not open socket' + #13#10 + 'Host: '+ StringGet(host) + ':'+IntToStr(port));
   end;
 end;
end;
 //Close Connection on current Socket
procedure SocketClose; cdecl;
begin
  if War3Client <> nil then
   War3Client.Client.Active:= false;
end;
 //Send Msg to server
function SocketSendMsg(msg: LongInt): jass_boolean; cdecl;
 var i: integer;
begin
  i:= War3Client.Client.Socket.SendText(StringGet(msg)+NewPack);
  Result:= BoolToJB(i > 0);
end;

procedure TriggerRegisterSocketEvent(t: jass_trigger; event_type: jass_integer); cdecl;
begin
// ShowMessage('load sock event '+IntToStr(event_type));
 if Socketevent[event_type] <> nil then
   TriggerGroupAdd(SocketEvent[event_type],t);
end;

 {MGScript Functions}

// Get function name
function MGSGetComName(com: LongInt): LongInt; cdecl;
 var buf: string;
begin
 buf:= StringGet(com);
 buf:= mgscript.GetComName(buf);
 Result:= StringMap(PChar(buf));
end;
// Get parametrs line
function MGSGetComParams(com: LongInt): LongInt; cdecl;
 var buf: string;
begin
 buf:= StringGet(com);
 buf:= mgscript.GetComParams(buf);
 Result:= StringMap(PChar(buf));
end;
 //Get Param in function by NUMBER
function MGSGetComParam(com: LongInt; pos: jass_integer): LongInt; cdecl;
 var buf: string;
begin
 buf:= StringGet(com);
 buf:= mgscript.GetComParam(buf,pos);
 Result:= StringMap(PChar(buf));
end;
 //GET PARENT obj in functionname
function MGSGetParentObj(comname: LongInt; lvl: jass_integer): LongInt; cdecl;
 var buf: string;
begin
 buf:= StringGet(comname);
 buf:= mgscript.GetParentObj(buf,lvl);
 Result:= StringMap(PChar(buf));
end;

function S2RawCode(str: LongInt): jass_integer; cdecl;
 var buf: string;
begin
 buf:= StringGet(str);
 Result:= RawCodeToInt(buf);
end;

function RawCode2S(int: jass_integer): LongInt; cdecl;
begin
 Result:= StringMap(PChar(IntToRawCode(int)));
end;

function SIAddNewItem(i: jass_item; hitem: jass_integer): jass_integer; cdecl;
begin
 TServerItem.Create(i,hitem);
 Result:= ServerItemCount-1;
end;

function SIFindItemByHITEM(hitem: jass_integer): jass_integer; cdecl;
 var i: integer;
begin
 for i:= 0 to ServerItemCount-1 do
  if ServerItems[i].HITEM = hitem then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

function SIFindItemByItem(itm: jass_item): jass_integer; cdecl;
 var i: integer;
begin
 for i:= 0 to ServerItemCount - 1 do
  if ServerItems[i]._itm.handleID = itm.handleID then
  begin
   Result:= i;
   exit;
  end;
 Result:= -1;
end;

function SIGetItem(id: jass_integer): jass_item; cdecl;
begin
 if (id >= 0) and (id < ServerItemCount) then
  if ServerItems[id] <> nil then
    Result:= ServerItems[id]._itm;
end;

function SIGetHITEM(id: jass_integer): jass_integer; cdecl;
begin
 Result:= -1;
 if (id >= 0) and (id < ServerItemCount) then
  if ServerItems[id] <> nil then
    Result:= ServerItems[id].HITEM;
end;

procedure SIDelete(id: jass_integer); cdecl;
begin
 if (id >= 0) and (id < ServerItemCount) then
  if ServerItems[id] <> nil then
    ServerItems[id].Destroy;
end;


 {Server Effects Functions}

function SEAddNewEffect(e: jass_effect; heffect: jass_integer): jass_integer; cdecl;
begin
  TServerEffect.Create(e, heffect);
  Result := ServerEffectCount-1;
end;

function SEFindEffectByHEFFECT(heffect: jass_integer): jass_integer; cdecl;
begin
  Result := FindEffectByHEFFECT(heffect);
end;

function SEGetEffect(id: jass_integer): jass_effect; cdecl;
begin
  if (id < 0) or (id >= ServerEffectCount) then Exit;
  if ServerEffects[id] <> nil then
    Result := ServerEffects[id]._effect;
end;

procedure SEDelete(id: jass_integer); cdecl;
begin
  if (id < 0) or (id >= ServerEffectCount) then exit;
  if ServerEffects[id] <> nil then
    ServerEffects[id].Destroy;
end;

 {Server Units Functions}

 //Add New Unit in Units Array
function SUAddNewUnit(u: jass_unit; hunit: jass_integer):jass_integer; cdecl;
begin
 TServerUnit.Create(u,hunit);
 Result:= ServerUnitCount-1;
end;
 //Find Unit in Units Array by HUNIT
function SUFindUnitByHUNIT(hunit: jass_integer): jass_integer; cdecl;
begin
 Result:= FindUnitByHUnit(hunit);
end;
 //Find Unit in Units Array by UNIT
function SUFindUnitByUnit(u: jass_unit): jass_integer; cdecl;
begin
 Result:= FindUnitByUnit(u);
end;
 //Get Unit on Units array by ID
function SUGetUnit(id: jass_integer): jass_unit; cdecl;
begin
 //Result:= LongInt(nil);
 if (id >= 0) and (id < ServerUnitCount) then
  Result:= ServerUnits[id]._unit;
end;
 //Get HUNIT on units array by ID
function SUGetHUNIT(id: jass_integer): jass_integer; cdecl;
begin
 Result:= -1;
 if (id >= 0) and (id < ServerUnitCount) then
  Result:= ServerUnits[id].HUNIT;
end;

procedure SUDelete(id: jass_integer); cdecl;
begin
 if (id >= 0) and (id < ServerUnitCount) then
  ServerUnits[id].Destroy;
end;

 {Server Dest Functions}

 //Add New Dest in Dest Array
function SDAddNewDest(d: jass_destructable; hdest: jass_integer): jass_integer; cdecl
begin
 TServerDest.Create(d,hdest);
 Result:= ServerDestCount-1;
end;

 //Find Dest in Dest Array by HDEST
function SDFindDestByHDEST(hdest: jass_integer): jass_integer; cdecl;
begin
 Result:= FindDestByHDEST(hdest);
end;
 //Find Dest in Dest Array by DEST
function SDFindDestByDest(dest: jass_destructable): jass_integer; cdecl;
begin
 Result:= FindDestByDest(dest);
end;
 //Get Dest
function SDGetDest(id: jass_integer): jass_destructable; cdecl;
begin
 if (id >= 0) and (id < ServerDestCount) then
  if ServerDests[id] <> nil then
    Result:= ServerDests[id]._dest;
end;
 //Get HDEST
function SDGetHDEST(id: jass_integer): jass_integer; cdecl;
begin
 Result:= -1;
 if (id >= 0) and (id < ServerDestCount) then
  if ServerDests[id] <> nil then
    Result:= ServerDests[id].HDEST;
end;

procedure SDDelete(id: jass_integer); cdecl;
begin
 if (id >= 0) and (id < ServerDestCount) then
  if ServerDests[id] <> nil then
    ServerDests[id].Destroy;
end;

//=======================
// INIT NATIVE PACK
//=======================

procedure MapEnd_proc; cdecl; export;
begin
  AddToLog('Map -> end game');
  ResetAllData;
end;

procedure MapStart_proc; cdecl; export;
 var i: integer;
begin
  InitLog();
  AddToLog('Map -> Start game');
  if isFirstRun then
  begin
    AddToLog('Game -> First run');
    isFirstRun := false;
    ResetAllData;
    SetLength(SocketEvent, 4);
    SocketEventCount := 4;
    for i:= 0 to 3 do
      SocketEvent[i]:= CreateTriggerGroup;
    War3Client:= TWar3Sckt.Create;
  end else
    AddToLog('Game -> Not first run');
end;

function LoadNativePack(m: HModule): boolean; cdecl; export;
begin
 Result:= false;
 if not InitJAPI(m) then
 begin
   AddToLog('Native -> Can not init JAPI');
   Exit;
 end;
 //Add New Natives
  //MGScript native functions
  AddNative(@MGSGetComName,'MGSGetComName','(S)S');
  AddNative(@MGSGetComParams,'MGSGetComParams','(S)S');
  AddNative(@MGSGetComParam,'MGSGetComParam','(SI)S');
  AddNative(@MGSGetParentObj,'MGSGetParentObj','(SI)S');
  //Socket native functions
  AddNative(@SocketClose,'SocketClose','()');
  AddNative(@SocketGetErrorCode,'SocketGetErrorCode','()I');
  AddNative(@SocketGetReciveMsg,'SocketGetReciveMsg','()S');
  AddNative(@SocketOpen,'SocketOpen','(SI)B');
  AddNative(@SocketSendMsg,'SocketSendMsg','(S)B');
  AddNative(@TriggerRegisterSocketEvent,'TriggerRegisterSocketEvent','(Htrigger;I)');
  //Server Units native functions
  AddNative(@SUAddNewUnit,'SUAddNewUnit','(Hunit;I)I');
  AddNative(@SUFindUnitByHUNIT,'SUFindUnitByHUNIT','(I)I');
  AddNative(@SUFindUnitByUnit,'SUFindUnitByUnit','(Hunit;)I');
  AddNative(@SUGetHUNIT,'SUGetHUNIT','(I)I');
  AddNative(@SUGetUnit,'SUGetUnit','(I)Hunit;');
  AddNative(@SUDelete,'SUDelete','(I)');
  //Server Dest native functions
  AddNative(@SDAddNewDest,'SDAddNewDest','(Hdestructable;I)I');
  AddNative(@SDFindDestByHDEST,'SDFindDestByHDEST','(I)I');
  AddNative(@SDFindDestByDest,'SDFindDestByDest','(Hdestructable;)I');
  AddNative(@SDGetDest,'SDGetDest','(I)Hdestructable;');
  AddNative(@SDGetHDEST,'SDGetHDEST','(I)I');
  AddNative(@SDDelete,'SDDelete','(I)');
  //Server Items native functions
  AddNative(@SIAddNewItem,'SIAddNewItem','(Hitem;I)I');
  AddNative(@SIFindItemByHITEM,'SIFindItemByHITEM','(I)I');
  AddNative(@SIFindItemByItem,'SIFindItemByItem','(Hitem;)I');
  AddNative(@SIGetItem,'SIGetItem','(I)Hitem;');
  AddNative(@SIGetHITEM,'SIGetHITEM','(I)I');
  AddNative(@SIDelete,'SIDelete','(I)');
  //Server Effects native functions
  AddNative(@SEAddNewEffect,'SEAddNewEffect','(Heffect;I)I');
  AddNative(@SEFindEffectByHEFFECT,'SEFindEffectByHEFFECT','(I)I');
  AddNative(@SEGetEffect,'SEGetEffect','(I)Heffect;');
  AddNative(@SEDelete,'SEDelete','(I)');
  //Other functions
  AddNative(@S2RawCode,'S2RawCode','(S)I');
  AddNative(@RawCode2S,'RawCode2S','(I)S');
  AddNative(@JDiv,'div','(II)I');
  AddNative(@JMod,'mod','(II)I');
  AddNative(@ReadIP,'ReadIp','()S');
  AddNative(@ResetAllData,'ResetAllData','()');
 AddToLog('Native -> Loaded');
 Result:= true;
end;

exports
 MapStart_proc name 'MapStart_proc',
 MapEnd_proc name 'MapEnd_proc',
 LoadNativePack name 'LoadNativePack';

begin
end.
