library WarSock;

uses
  SysUtils,
  Classes,
  ScktComp,
  Dialogs,
  //=====
  jass_types,
  nAPI;

type
 TWarClient = class
  ClientSocket: TClientSocket;
  constructor Create;
  destructor Destroy; override;
  //===============
  procedure ClientSocketConnect(Sender: TObject; Socket: TCustomWinSocket);
  procedure ClientSocketRead(Sender: TObject; Socket: TCustomWinSocket);
  procedure ClientSocketError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
  procedure ClientSocketDisconnect(Sender: TObject; Socket: TCustomWinSocket);
 end;

var WarClient: TWarClient;
    SocketEvent: array [0..3] of TTriggerGroup;
    lastrecivemsg: string;
    lasterrorcode: integer;
    const delitel = #10#11;


{$R *.res}

procedure TWarClient.ClientSocketConnect(Sender: TObject; Socket: TCustomWinSocket);
begin
 TriggerGroupRun(SocketEvent[0]);
end;

procedure MessageProc(msg: string; Socket: TCustomWinSocket);
begin
 lastrecivemsg:= msg;
 TriggerGroupRun(SocketEvent[1]);
end;

procedure TWarClient.ClientSocketRead(Sender: TObject; Socket: TCustomWinSocket);
 var buf_msg,msg: string;
     i: integer;
begin
 buf_msg:= Socket.ReceiveText;
 i:= 0;
 repeat
  i:= Pos(delitel,buf_msg);
  if i <> 0 then
   begin
    msg:= Copy(buf_msg,1,i-1);
    Delete(buf_msg,1,i+Length(delitel)-1);
    MessageProc(msg,Socket);
  end;
 until i = 0;
end;

procedure TWarClient.ClientSocketError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
 lasterrorcode:= ErrorCode;
 ErrorCode:= 0;
 TriggerGroupRun(SocketEvent[2]);
end;

procedure TWarClient.ClientSocketDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
 TriggerGroupRun(SocketEvent[3]);
end;

constructor TWarClient.Create;
begin
 inherited Create;
 ClientSocket:= TClientSocket.Create(nil);
 ClientSocket.OnError:= ClientSocketError;
 ClientSocket.OnConnect:= ClientSocketConnect;
 ClientSocket.OnDisconnect:= ClientSocketDisconnect;
 ClientSocket.OnRead:= ClientSocketRead;
end;

destructor TWarClient.Destroy;
begin
 inherited Destroy;
 ClientSocket.Active:= false;
end;

procedure MapStart_proc; cdecl; export;
 var i: integer;
begin
 for i:= 0 to 3 do SocketEvent[i]:= CreateTriggerGroup;
 WarClient:= TWarClient.Create;
end;

procedure MapEnd_proc; cdecl; export;
 var i: integer;
begin
 for i:= 0 to 3 do DestroyTriggerGroup(SocketEvent[i]);
 WarClient.Destroy;
 //WarClient.ClientSocket.Active:= false;
{ WarClient.ClientSocket.Active:= false;
 WarClient.ClientSocket.Destroy;
 WarClient:= nil;}
end;

//================
//NativeFunctions
//================

function GetRecivingMessage: LongInt; cdecl;
begin
 Result:= StringMap(PChar(lastrecivemsg));
end;

function GetErrorCode: Jass_integer; cdecl;
begin
 Result:= lasterrorcode;
end;

procedure TriggerRegisterSocketEvent(t: jass_trigger; event_type: jass_integer); cdecl;
begin
 TriggerGroupAdd(SocketEvent[event_type],t);
end;

procedure SocketConnect(ip: LongInt; port: jass_integer); cdecl;
begin
 WarClient.ClientSocket.Host:= StringGet(ip);
 WarClient.ClientSocket.Port:= port;
 WarClient.ClientSocket.Active:= true;
end;

procedure SocketDisconnect; cdecl;
begin
 WarClient.ClientSocket.Active:= false;
end;

procedure SocketSendString(str: LongInt); cdecl;
begin
 WarClient.ClientSocket.Socket.SendText(StringGet(str)+delitel);
end;

//===========
// MGS
//===========

function GetComName(jass_str: LongInt):LongInt; cdecl;
 var comname,icom: string;
     i: integer;
     get: boolean;
begin
 comname:= '';
 get:= true;
 icom:= StringGet(jass_str);
 //ShowMessage(icom);
 for i:= 1 to Length(icom) do
 begin
  if icom[i] = '(' then get:= false;
  if get and (icom[i] <> ' ') and (icom[i] <> '(') then comname:= comname + icom[i];
 end;
 //ShowMessage(comname);
 Result:= StringMap(PChar(comname));
end;

function GetComParams(com: LongInt):LongInt; cdecl;
 var params,icom: string;
     i: integer;
     get,isstr: boolean;
     scobcount: integer;
begin
 scobcount:= 0;
 params:= '';
 get:= false;
 icom:= StringGet(com);
 isstr:= false;
 for i:= 1 to Length(icom) do
 begin
  if icom[i] = '"' then isstr:= not isstr;
  if icom[i] = ')' then
  begin
   if not isstr then dec(scobcount);
   if scobcount = 0 then get:= false;
  end;
  if get then params:= params + icom[i];
  if icom[i] = '(' then
  begin
   if scobcount = 0 then get:= true;
   if not isstr then inc(scobcount);
  end;
 end;
 result:= StringMap(PChar(params));
end;

function GetComParam(params: LongInt; pos: jass_integer):LongInt; cdecl;
 var iparams,par: string;
     i,n: integer;
     isstr: boolean;
begin
 iparams:= StringGet(params);
 par:= '';
 isstr:= false;
 n:= 1;
 for i:= 1 to Length(iparams) do
 begin
  if (iparams[i] = ',') and (not isstr) then inc(n);
  if iparams[i] = '"' then isstr:= not isstr;
  if n = pos then
   if (iparams[i] <> '"') then
    if isstr then
     par:= par + iparams[i]
    else
     if (iparams[i] <> ',') and (iparams[i] <> ' ') then
      par:= par + iparams[i];
 end;
 Result:= StringMap(PChar(par));
end;

//===========
//Init Nativs functions
//===========

function LoadNativePack(m: HModule): boolean; cdecl; export;
begin
 Result:= false;
 if not InitJAPI(m) then exit;
 //add natives
  AddNative(@GetRecivingMessage,'GetRecivingMessage','()S');
  AddNative(@GetErrorCode,'GetErrorCode','()I');
  AddNative(@TriggerRegisterSocketEvent,'TriggerRegisterSocketEvent','(Htrigger;I)V');
  AddNative(@SocketConnect,'SocketConnect','(SI)');
  AddNative(@SocketDisconnect,'SocketDisconnet','()');
  AddNative(@SocketSendString,'SocketSendString','(S)');
  AddNative(@GetComName,'GetComName','(S)S');
  AddNative(@GetComParams,'GetComParams','(S)S');
  AddNative(@GetComParam,'GetComParam','(SI)S');
 //============
 Result:= true;
end;

exports
  MapStart_proc name 'MapStart_proc',
  MapEnd_proc name 'MapEnd_proc',
  LoadNativePack name 'LoadNativePack';

begin
end.
 